------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------


with Text_Io;
use  Text_Io;

with Xml_Tag;
use Xml_Tag;
with Unbounded_Strings;
use Unbounded_Strings;

with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;
with Buffer_Set;
use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;

with Ada.Exceptions;
use Ada.Exceptions;

with Processors;
use Processors;
with Processor_Set;
use Processor_Set;
use Processor_Set.Generic_Processor_Set;

with Call_Scheduling_Framework;
use Call_Scheduling_Framework;
with Call_Buffer_Framework;
use Call_Buffer_Framework;
use Call_Buffer_Framework.Buffer_Result_Package;
with Call_Resource_Framework;
use Call_Resource_Framework;
with Call_Dependency_Framework;
use Call_Dependency_Framework;
with Call_Random_Framework;
use Call_Random_Framework;

with Scheduler;
use Scheduler;

with Multiprocessor_Services;
use Multiprocessor_Services;
use Multiprocessor_Services.Scheduling_Result_Per_Processor;



with Config;
use Config;
with Translate;
use Translate;


package body Call_Framework  is



   type Sub_Framework_Request is 
      record 
         Statement : Framework_Statement_Type;  
         Param     : Parameters_Table;  
         Target    : Processor_Ptr;  
      end record; 





   procedure Sequential_Framework_Request (
         Sys      : in out System;                  
         Request  : in     Framework_Request_Table; 
         Response : in out Framework_Response_Table ) is 
   begin
      Sequential_Framework_Request(Sys, Request, Response, Total_Order,
         String_Output);
   end   Sequential_Framework_Request;


   procedure Put (
         F : in     Framework_Request ) is 
   begin
      Put(F.Statement'Img & "/" & To_String(F.Target));
   end Put;


   procedure Initialize (
         F : in out Framework_Request ) is 
   begin
      F.Target:=Empty_String;
      Initialize(F.Param);
   end Initialize;


   procedure Initialize (
         F : in out Framework_Response ) is 
   begin
      F.Title:=Empty_String;
      F.Text:=Empty_String;
   end Initialize;


   procedure Put (
         F : in     Framework_Response ) is 
   begin
      Put_Line(To_String(F.Title));
      Put_Line(To_String(F.Text));
   end Put;




   procedure Call_A_Request (
         Sys        : in out System;                                
         Request    : in     Sub_Framework_Request;                 
         A_Response : in out Unbounded_String;                      
         Output     : in     Output_Format         := String_Output ) is 



   begin



      case Request.Statement is

         when Scheduling_Simulation_Time_Line =>


            -- Check parameters
            --
            if (Request.Param.Nb_Entries/=5)
                  then
               Raise_Exception(Invalid_Number_Of_Parameter'Identity,
                  "5 expected parameters");
            end if;

            if (Request.Param.Entries(0).Name/=To_Unbounded_String(
                     "period")) then
               Raise_Exception(Unexpected_Parameter'Identity, "Period");
            end if;
            if (Request.Param.Entries(1).Name/=To_Unbounded_String(
                     "schedule_with_offsets")) then
               Raise_Exception(Unexpected_Parameter'Identity,
                  "schedule_with_offsets");
            end if;
            if (Request.Param.Entries(2).Name/=To_Unbounded_String(
                     "schedule_with_precedencies")) then
               Raise_Exception(Unexpected_Parameter'Identity,
                  "schedule_with_precedencies");
            end if;
            if (Request.Param.Entries(3).Name/=To_Unbounded_String(
                     "schedule_with_resources")) then
               Raise_Exception(Unexpected_Parameter'Identity,
                  "schedule_with_resources");
            end if;
            if (Request.Param.Entries(4).Name/=To_Unbounded_String(
                     "seed_value")) then
               Raise_Exception(Unexpected_Parameter'Identity,
                  "seed_value");
            end if;



            if (Request.Param.Entries(0).Ptype/=Integer_Parameter) then
               Raise_Exception(Parameter_Type_Error'Identity, "Period");
            end if;
            if (Request.Param.Entries(1).Ptype/=Boolean_Parameter) then
               Raise_Exception(Parameter_Type_Error'Identity,
                  "schedule_with_offsets");
            end if;
            if (Request.Param.Entries(2).Ptype/=Boolean_Parameter) then
               Raise_Exception(Parameter_Type_Error'Identity,
                  "schedule_with_precedencies");
            end if;
            if (Request.Param.Entries(3).Ptype/=Boolean_Parameter) then
               Raise_Exception(Parameter_Type_Error'Identity,
                  "schedule_with_resources");
            end if;
            if (Request.Param.Entries(4).Ptype/=Integer_Parameter) then
               Raise_Exception(Parameter_Type_Error'Identity,
                  "seed_value");
            end if;



            -- Call framework
            --
            Compute_Simulation_Time_Line (
               Sys,  A_Response,
               Request.Param.Entries(0).Integer_Value,
               Request.Param.Entries(1).Boolean_Value,
               Request.Param.Entries(2).Boolean_Value,
               Request.Param.Entries(3).Boolean_Value,
               Request.Param.Entries(4).Integer_Value,
               Output);


         when Scheduling_Simulation_Basics =>

            Compute_Simulation_Basics (
               Sys, A_Response, Request.Target,Output);


         when  Scheduling_Simulation_Preemption_Number =>
            null;
         when Scheduling_Simulation_Response_Time =>
            null;

         when Scheduling_Feasibility_Response_Time   =>
            null;

         when Scheduling_Feasibility_Basics =>

            if (Get_Number_Of_Task_From_Processor(Sys.Tasks, Request.
                     Target.Name)/=0)
                  then
               Compute_Feasibility_Period (
                  Sys, A_Response, Request.Target,Output);

               Compute_Feasibility_Processor_Utilization_Factor (
                  Sys,  A_Response, Request.Target,Output);

               Compute_Feasibility_Response_Time (
                  Sys, A_Response, Request.Target,Output);

            end if;

         when Scheduling_Set_Priorities_According_To_Deadline_Monotonic =>

            Set_Priorities (
               Sys, A_Response, Request.Target, Deadline_Monotonic,Output);


         when Scheduling_Set_Priorities_According_To_Rate_Monotonic =>

            Set_Priorities (
               Sys, A_Response, Request.Target, Rate_Monotonic,Output);


         when Buffer_Bound =>

            if (Get_Number_Of_Buffer_From_Processor (Sys.Buffers, Request.
                     Target.
                     Name) /= 0 )
                  then

               Compute_Buffer_Bound (Sys, A_Response, Request.Target,
                  Output);
            end if;


         when Buffer_Utilization   =>
            if (Get_Number_Of_Buffer_From_Processor (Sys.Buffers, Request.
                     Target.
                     Name) /= 0 )
                  then
               Compute_Buffer_Utilization (Sys, A_Response, Request.
                  Target,
                  Output);
            end if;




         when  Resource_Set_Blocking_Time =>
            Compute_Blocking_Time (Sys, A_Response, Request.Target, True,
               Output);
         when  Resource_Compute_Blocking_Time =>
            Compute_Blocking_Time (Sys, A_Response, Request.Target, False,
               Output);


         when Random_Response_Time_Density    =>
            Compute_Response_Time_Density (
               Sys, A_Response, Request.Target, Output);


         when Dependency_Compute_Chetto_Blazewicz_Priority  =>
            Chetto_Parameter_Modifications(
               Sys, A_Response, Request.Target, False, True, False, Output);

         when Dependency_Compute_Chetto_Blazewicz_Deadline   =>
            Chetto_Parameter_Modifications(
               Sys, A_Response, Request.Target, True, False, False, Output);

         when Dependency_Set_Chetto_Blazewicz_Deadline =>
            Chetto_Parameter_Modifications(
               Sys, A_Response, Request.Target, True, False, True, Output);

         when Dependency_Set_Chetto_Blazewicz_Priority  =>
            Chetto_Parameter_Modifications(
               Sys, A_Response, Request.Target, False, True, True, Output);

         when Dependency_Compute_End_To_End_Response_Time_One_Step =>
            Compute_End_To_End_Response_Time (
               Sys, A_Response, Request.Target, False,True, Output);

         when Dependency_Set_End_To_End_Response_Time_One_Step   =>
            Compute_End_To_End_Response_Time (
               Sys, A_Response, Request.Target, True, True, Output);

         when Dependency_Compute_End_To_End_Response_Time_All_Steps  =>
            Compute_End_To_End_Response_Time (
               Sys, A_Response, Request.Target, False, False, Output);

         when Dependency_Set_End_To_End_Response_Time_All_Steps   =>
            Compute_End_To_End_Response_Time (
               Sys, A_Response, Request.Target, True, False, Output);

      end case;
   end Call_A_Request;



   procedure Initialize_Before_Request (
         Request : in     Framework_Request ) is 
   begin
      case Request.Statement is
         when Scheduling_Simulation_Time_Line =>
            Initialize(Sched.all);
         when Scheduling_Simulation_Basics =>
            null;
         when  Scheduling_Simulation_Preemption_Number =>
            null;
         when Scheduling_Simulation_Response_Time =>
            null;
         when Scheduling_Feasibility_Response_Time   =>
            null;
         when Scheduling_Feasibility_Basics =>
            null;
         when Scheduling_Set_Priorities_According_To_Deadline_Monotonic =>
            null;
         when Scheduling_Set_Priorities_According_To_Rate_Monotonic =>
            null;

         when Buffer_Bound =>
            null;
         when Buffer_Utilization   =>
            Initialize(Buff);

         when  Resource_Set_Blocking_Time =>
            null;
         when  Resource_Compute_Blocking_Time =>
            null;


         when Random_Response_Time_Density    =>
            null;

         when Dependency_Set_Chetto_Blazewicz_Deadline =>
            null;
         when Dependency_Compute_End_To_End_Response_Time_One_Step =>
            null;
         when Dependency_Set_End_To_End_Response_Time_One_Step   =>
            null;
         when Dependency_Compute_End_To_End_Response_Time_All_Steps  =>
            null;
         when Dependency_Set_End_To_End_Response_Time_All_Steps   =>
            null;
         when Dependency_Compute_Chetto_Blazewicz_Priority  =>
            null;
         when Dependency_Compute_Chetto_Blazewicz_Deadline   =>
            null;
         when Dependency_Set_Chetto_Blazewicz_Priority  =>
            null;
      end case;
   end Initialize_Before_Request;



   function Build_A_Title (
         Request : in     Framework_Request ) 
     return Unbounded_String is 
   begin

      case Request.Statement is

         when Scheduling_Simulation_Time_Line =>
            return Empty_String;
         when Scheduling_Simulation_Basics =>
            return Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Unbounded_Lf & Lb_Draw(Current_Language);
         when  Scheduling_Simulation_Preemption_Number =>
            return Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Unbounded_Lf & Lb_Draw(Current_Language);
         when Scheduling_Simulation_Response_Time =>
            return Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Unbounded_Lf & Lb_Draw(Current_Language);
         when Scheduling_Feasibility_Response_Time   =>
            return Start_Sched_Feas & Unbounded_Lf & Start_Title &
               Unbounded_Lf &  Lb_Check(Current_Language);
         when Scheduling_Feasibility_Basics =>
            return Start_Sched_Feas & Unbounded_Lf & Start_Title &
               Lb_Check(Current_Language);
         when Scheduling_Set_Priorities_According_To_Deadline_Monotonic =>
            return Start_Set_Prio & Unbounded_Lf & Start_Title &
               Unbounded_Lf & Lb_Set_Priorities(Current_Language);
         when Scheduling_Set_Priorities_According_To_Rate_Monotonic =>
            return Start_Set_Prio & Unbounded_Lf & Start_Title &
               Unbounded_Lf & Lb_Set_Priorities(Current_Language);

         when Buffer_Bound =>
            return Start_Buf_Bound & Unbounded_Lf & Start_Title &
               Unbounded_Lf & Lb_Bound_Buffer(Current_Language);
         when Buffer_Utilization   =>
            return Lb_Utilization_Level(Current_Language);

         when  Resource_Set_Blocking_Time =>
            return Start_Block & Unbounded_Lf & Start_Title & Unbounded_Lf &
               Lb_Blocking_Time(Current_Language);
         when  Resource_Compute_Blocking_Time =>
            return Start_Block & Unbounded_Lf & Start_Title & Unbounded_Lf &
               Lb_Blocking_Time(Current_Language);

         when Random_Response_Time_Density    =>
            return Start_Response_Time_Density & Start_Title &
               Unbounded_Lf & Lb_Response_Time_Density(Current_Language);

         when Dependency_Compute_End_To_End_Response_Time_One_Step =>
            return  Lb_Jitter_From_Response_Time(
               Current_Language);
         when Dependency_Set_End_To_End_Response_Time_One_Step   =>
            return  Lb_Jitter_From_Response_Time(
               Current_Language);
         when Dependency_Compute_End_To_End_Response_Time_All_Steps  =>
            return Start_End_To_End & Unbounded_Lf & Start_Title &
               Unbounded_Lf & Lb_Jitter_From_Response_Time(
               Current_Language);
         when Dependency_Set_End_To_End_Response_Time_All_Steps   =>
            return Start_End_To_End & Unbounded_Lf & Start_Title &
               Unbounded_Lf & Lb_Jitter_From_Response_Time(
               Current_Language);

         when Dependency_Compute_Chetto_Blazewicz_Priority  =>
            return Lb_Chetto(Current_Language);
         when Dependency_Compute_Chetto_Blazewicz_Deadline   =>
            return Lb_Chetto(Current_Language);
         when Dependency_Set_Chetto_Blazewicz_Priority  =>
            return Lb_Chetto(Current_Language);
         when Dependency_Set_Chetto_Blazewicz_Deadline =>
            return Lb_Chetto(Current_Language);

      end case;

   end Build_A_Title;


   procedure Sequential_Framework_Request (
         Sys      : in out System_Ptr;               
         Request  : in     Framework_Request_Table;  
         Response : in out Framework_Response_Table; 
         Order    : in     Perform_Order;            
         Output   : in     Output_Format             ) is 
   begin
      Sequential_Framework_Request(Sys.all, Request, Response, Order,
         Output);
   end  Sequential_Framework_Request;


   procedure Sequential_Framework_Request (
         Sys      : in out System;                   
         Request  : in     Framework_Request_Table;  
         Response : in out Framework_Response_Table; 
         Order    : in     Perform_Order;            
         Output   : in     Output_Format             ) is 

      A_Response  : Framework_Response;  
      A_Processor : Processor_Ptr;  
      My_Iterator : Processors_Iterator;  
      A_Request   : Sub_Framework_Request;  

   begin

      for I in 0..Request.Nb_Entries-1 loop


         Initialize_Before_Request(Request.Entries(I));

         Reset_Iterator(My_Iterator);


         loop
            Current_Element(Sys.Processors,A_Processor,My_Iterator);

            -- Do a request
            --
            if (A_Processor.Name=Request.Entries(I).Target) or
                  (Request.Entries(I).Target=Empty_String) then

               Initialize(A_Response);
               A_Request.Target:=A_Processor;
               A_Request.Param:=Request.Entries(I).Param;
               A_Request.Statement:=Request.Entries(I).Statement;

               Call_A_Request(Sys, A_Request,
                  A_Response.Text, Output);

               if A_Response.Text/=Empty_String then

                  A_Response.Title:=Build_A_Title (Request.Entries(I));
                  if  A_Response.Title/=Empty_String then
                     A_Response.Title:= Unbounded_Lf & Unbounded_Lf &
                        Unbounded_Lf&
                        A_Response.Title & Lb_Comma &
                        Lb_Processor(Current_Language) &
                        Start_Proc & A_Processor.Name & End_Proc &
                        Lb_Colon & Unbounded_Lf & End_Title &
                        Unbounded_Lf;
                  end if;

                  Add(Response, A_Response);
               end if;

            end if;

            exit when Is_Last_Element(Sys.Processors, My_Iterator) =
               True;

            Next_Element(Sys.Processors, My_Iterator);
         end loop;

      end loop;
   end Sequential_Framework_Request;



   procedure Parallel_Framework_Request (
         Sys      : in out System;                                   
         Request  : in     Framework_Request_Table;                  
         Response : in out Framework_Response_Table;                 
         Order    : in     Perform_Order            := Total_Order;  
         Output   : in     Output_Format            := String_Output ) is 
   begin
      null;
   end Parallel_Framework_Request;



   procedure Distributed_Framework_Request (
         Sys      : in out System;                                   
         Request  : in     Framework_Request_Table;                  
         Response : in out Framework_Response_Table;                 
         Order    : in     Perform_Order            := Total_Order;  
         Output   : in     Output_Format            := String_Output ) is 
   begin
      null;
   end Distributed_Framework_Request ;





end Call_Framework;


