------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------


with Ada.Strings.Unbounded;
use  Ada.Strings.Unbounded;

with Parameters;
use Parameters;
use Parameters.Parameters_Table_Package;

with Config;
use Config;


with Systems;
use Systems;

with Tables;
with Convert_Strings;
with Convert_Unbounded_Strings;



package Call_Framework  is


   type Framework_Statement_Type is 
         (Scheduling_Simulation_Basics,                              
          Scheduling_Simulation_Time_Line,                           
          Scheduling_Simulation_Preemption_Number,                   
          Scheduling_Simulation_Response_Time,                       
          Scheduling_Feasibility_Response_Time,                      
          Scheduling_Feasibility_Basics,                             
          Scheduling_Set_Priorities_According_To_Deadline_Monotonic, 
          Scheduling_Set_Priorities_According_To_Rate_Monotonic,     
          Buffer_Bound,                                              
          Buffer_Utilization,                                        
          Resource_Set_Blocking_Time,                                
          Resource_Compute_Blocking_Time,                            
          Random_Response_Time_Density,                              
          Dependency_Compute_End_To_End_Response_Time_One_Step,      
          Dependency_Set_End_To_End_Response_Time_One_Step,          
          Dependency_Compute_End_To_End_Response_Time_All_Steps,     
          Dependency_Set_End_To_End_Response_Time_All_Steps,         
          Dependency_Compute_Chetto_Blazewicz_Priority,              
          Dependency_Compute_Chetto_Blazewicz_Deadline,              
          Dependency_Set_Chetto_Blazewicz_Priority,                  
          Dependency_Set_Chetto_Blazewicz_Deadline); 
   pragma Convention( C,  Framework_Statement_Type);



   procedure To_Framework_Statement is 
   new Convert_Strings(Framework_Statement_Type,
      Scheduling_Simulation_Basics);
   procedure To_Framework_Statement is 
   new Convert_Unbounded_Strings(Framework_Statement_Type,
      Scheduling_Simulation_Basics);





   -- Store a computing request. 
   -- If "target" contains something, the computing should only
   -- done on one processor which is referecendd in "target"
   -- if "target" is a empty string, we compute for all processors
   --
   type Framework_Request is 
      record 
         Statement : Framework_Statement_Type;  
         Param     : Parameters_Table;  
         Target    : Unbounded_String;  
      end record; 


   Invalid_Number_Of_Parameter : exception;  
   Unexpected_Parameter        : exception;  
   Parameter_Type_Error        : exception;  


   procedure Put (
         F : in     Framework_Request ); 
   procedure Initialize (
         F : in out Framework_Request ); 
   pragma Export(C, Put, "put_request");
   pragma Export(C, Initialize, "initialize_request");



   package Framework_Request_Package is new Tables(Framework_Request,
      Max_Request,Put);
   use Framework_Request_Package;

   subtype Framework_Request_Range is Framework_Request_Package.
      Table_Range;
   subtype Framework_Request_Table is Framework_Request_Package.
      Table;




   type Framework_Response is 
      record 
         Title : Unbounded_String;  
         Text  : Unbounded_String;  
      end record; 


   procedure Put (
         F : in     Framework_Response ); 
   procedure Initialize (
         F : in out Framework_Response ); 
   pragma Export(C, Put, "put_response");
   pragma Export(C, Initialize, "initialize_response");


   package Framework_Response_Package is new Tables(Framework_Response,
      Max_Request,Put);
   use Framework_Response_Package;

   subtype Framework_Response_Range is Framework_Response_Package.
      Table_Range;
   subtype Framework_Response_Table is Framework_Response_Package.
      Table;


   type Output_Format is 
         (Xml_Output,    
          String_Output, 
          Raw_Output); 
   pragma Convention( C, Output_Format);


   type Perform_Order is 
         (Total_Order, 
          Causal_Order); 
   pragma Convention( C, Perform_Order);


   procedure Sequential_Framework_Request (
         Sys      : in out System;                   
         Request  : in     Framework_Request_Table;  
         Response : in out Framework_Response_Table; 
         Order    : in     Perform_Order;            
         Output   : in     Output_Format             ); 


   procedure Sequential_Framework_Request (
         Sys      : in out System_Ptr;               
         Request  : in     Framework_Request_Table;  
         Response : in out Framework_Response_Table; 
         Order    : in     Perform_Order;            
         Output   : in     Output_Format             ); 
   pragma Export(C, Sequential_Framework_Request,
      "sequential_framework_request");


   procedure Sequential_Framework_Request (
         Sys      : in out System;                  
         Request  : in     Framework_Request_Table; 
         Response : in out Framework_Response_Table ); 


   procedure Parallel_Framework_Request (
         Sys      : in out System;                                   
         Request  : in     Framework_Request_Table;                  
         Response : in out Framework_Response_Table;                 
         Order    : in     Perform_Order            := Total_Order;  
         Output   : in     Output_Format            := String_Output ); 


   procedure Distributed_Framework_Request (
         Sys      : in out System;                                   
         Request  : in     Framework_Request_Table;                  
         Response : in out Framework_Response_Table;                 
         Order    : in     Perform_Order            := Total_Order;  
         Output   : in     Output_Format            := String_Output ); 



end Call_Framework;


