------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_Io;
use Text_Io;

with Xml_Tag;
use Xml_Tag;

with Tasks;
use Tasks;
with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Resources;
use Resources;
with Resource_Set;
use Resource_Set;
use Resource_Set.Generic_Resource_Set;

with Processors;
use Processors;
with Processor_Set;
use Processor_Set;
use Processor_Set.Generic_Processor_Set;

with Translate;
use Translate;


with Unbounded_Strings;
use Unbounded_Strings;

with Config;
use Config;


with Scheduler;
use Scheduler;
with Scheduler.Fixed_Priority;
use Scheduler.Fixed_Priority;

with Blocking_Time;
use Blocking_Time;



package body Call_Resource_Framework  is




   procedure Compute_Blocking_Time (
         Sys                  : in out System;                           
         Result               : in out Unbounded_String;                 
         A_Processor          : in     Processor_Ptr;                    
         Inject_Blocking_Time : in     Boolean;                          
         Output               : in     Output_Format    := String_Output ) is 


      My_Task_Iterator1     : Tasks_Iterator;  
      My_Task_Iterator2     : Tasks_Iterator;  
      My_Resource_Iterator  : Resources_Iterator;  
      My_Processor_Iterator : Processors_Iterator;  

      Tmp          : Tasks_Set;  
      A_Task       : Task_Ptr;  
      A_Task_Tmp   : Task_Ptr;  
      A_Resource   : Resource_Ptr;  
      A_Processor2 : Processor_Ptr;  

   begin



      -- Check protocol according to the scheduler
      -- If PIP/PCP resource, then scheduler 
      -- can only be RM, DM or HPF
      --
      Reset_Iterator(My_Resource_Iterator);

      loop
         Current_Element(Sys.Resources,A_Resource,My_Resource_Iterator);

         if A_Resource.Location = A_Processor.Name then

            if (A_Resource.Protocol =
                  Priority_Ceiling_Protocol)
                  or  (A_Resource.Protocol =
                  Priority_Inheritance_Protocol)
                  then


               Reset_Iterator(My_Processor_Iterator);

               loop

                  Current_Element(Sys.Processors, A_Processor2,
                     My_Processor_Iterator);

                  if (A_Processor2.Name = A_Resource.Location)
                        then
                     if
                           (Get_Name(A_Processor2.Scheduler.all) /=
                           Rate_Monotonic) and
                           (Get_Name(A_Processor2.Scheduler.all) /=
                           Deadline_Monotonic) and
                           (Get_Name(A_Processor2.Scheduler.all) /=
                           Highiest_Priority_First)
                           then

                        Result := Result &  Unbounded_Lf &
                           Lb_Warning(Current_Language) & A_Resource.Name &
                           Lb_Compute_Blocking_Error2(Current_Language) &
                           Unbounded_Lf;
                     end if;
                  end if;


                  exit when Is_Last_Element(Sys.Processors,
                     My_Processor_Iterator) = True;

                  Next_Element(Sys.Processors, My_Processor_Iterator);

               end loop;
            end if;
         end if;

         exit when Is_Last_Element(Sys.Resources, My_Resource_Iterator) =
            True;

         Next_Element(Sys.Resources, My_Resource_Iterator);

      end loop;



      -- Copy the task set before modification
      --
      Duplicate(Sys.Tasks, Tmp);


      -- Check if scheduler is RM/DM/HPF
      -- and then, change task_set priorities
      --
      Reset_Iterator(My_Processor_Iterator);

      loop

         Current_Element(Sys.Processors, A_Processor2,
            My_Processor_Iterator);

         if (Get_Name(A_Processor2.Scheduler.all) =
               Rate_Monotonic)
               then
            Set_Priority_According_To_Rm(
               Fixed_Priority_Scheduler(A_Processor2.Scheduler.all),
               Tmp,
               A_Processor2.Name);
         else
            if (Get_Name(A_Processor2.Scheduler.all) =
                  Deadline_Monotonic)
                  then
               Set_Priority_According_To_Dm(
                  Fixed_Priority_Scheduler(A_Processor2.Scheduler.all),
                  Tmp,
                  A_Processor2.Name);
            end if;
         end if;

         exit when Is_Last_Element(Sys.Processors, My_Processor_Iterator) =
            True;
         Next_Element(Sys.Processors, My_Processor_Iterator);

      end loop;


      -- And now, compute blocking time
      --

      Reset_Iterator(My_Task_Iterator1);

      Result:= Result & Start_Line;

      loop

         Current_Element(Tmp, A_Task, My_Task_Iterator1);

         if A_Task.Location = A_Processor.Name then

            A_Task.Blocking_Time :=
               Compute_Blocking_Time_Task(Tmp, Sys.Resources, A_Task.Name);

            Result:= Result &  Unbounded_Lf & Start_Task &
               To_Unbounded_String("    ") & A_Task.Name &
               " => " & A_Task.Blocking_Time'Img     &  End_Task &
               Start_Ref & To_Unbounded_String(
               " (") &
               Lb_See(Current_Language) & To_Unbounded_String("[3]).") &
               End_Ref;
         end if;

         exit when Is_Last_Element(Tmp, My_Task_Iterator1) = True;

         Next_Element(Tmp, My_Task_Iterator1);

      end loop;

      Result:= Result & Unbounded_Lf & End_Line & Unbounded_Lf;

      Result := Result & Unbounded_Lf;

      -- Update blocking time
      --
      if (Inject_Blocking_Time=True)
            then

         Reset_Iterator(My_Task_Iterator1);
         loop
            Current_Element(Tmp, A_Task_Tmp, My_Task_Iterator1);

            Reset_Iterator(My_Task_Iterator2);
            loop
               Current_Element(Sys.Tasks, A_Task, My_Task_Iterator2);

               if (A_Task.Name = A_Task_Tmp.Name)
                     then
                  A_Task.Blocking_Time :=
                     A_Task_Tmp.Blocking_Time;
               end if;

               exit when Is_Last_Element(Sys.Tasks, My_Task_Iterator2) =
                  True;
               Next_Element(Sys.Tasks, My_Task_Iterator2);
            end loop;


            exit when Is_Last_Element(Tmp, My_Task_Iterator1) = True;
            Next_Element(Tmp, My_Task_Iterator1);

         end loop;


         Result := Result   & Unbounded_Lf &
            Start_Line & Lb_Inject_Success(Current_Language) & End_Line &
            Unbounded_Lf & End_Block & Unbounded_Lf;
      end if;


   exception
      when Tasks.Compute_Aborded =>
         begin
            Result := Result &  Unbounded_Lf & Lb_Compute_Blocking_Error1(
               Current_Language) & Unbounded_Lf;
         end;

   end Compute_Blocking_Time;



end Call_Resource_Framework;









