------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------



with Config;
use Config;


with Xml_Tag;
use Xml_Tag;
with Scheduler;
use Scheduler;
use Scheduler.Double_Tasks_Parameters;
with Scheduler.Fixed_Priority;
use  Scheduler.Fixed_Priority;
with Scheduler.Fixed_Priority.Dm;
use  Scheduler.Fixed_Priority.Dm;
with Scheduler.Fixed_Priority.Rm;
use  Scheduler.Fixed_Priority.Rm;

with Double_Util;
use Double_Util;
with Translate;
use Translate;
with Tasks;
use Tasks;
with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;
with Unbounded_Strings;
use Unbounded_Strings;


with Multiprocessor_Services;
use Multiprocessor_Services;
use Multiprocessor_Services.Scheduling_Result_Per_Processor;

with  Gnat.Current_Exception;
use  Gnat.Current_Exception;

with Expressions;
use Expressions;


package body Call_Scheduling_Framework  is






   procedure Compute_Simulation_Time_Line (
         Sys                        : in out System;                           
         Result                     : in out Unbounded_String;                 
         Period                     : in     Natural;                          
         Schedule_With_Offsets      : in     Boolean;                          
         Schedule_With_Precedencies : in     Boolean;                          
         Schedule_With_Resources    : in     Boolean;                          
         Seed_Value                 : in     Natural;                          
         Output                     : in     Output_Format    := String_Output ) is 

   begin

      Sort(Sys.Tasks, Increasing_Name'access);
      Scheduling_Base_Period:=Period;


      Build_Multiprocessor_Scheduling(
         Sys.Processors, Sys.Tasks, Sys.Resources, Sys.Dependencies, Sched,
         Period,  Seed_Value,
         Schedule_With_Offsets, Schedule_With_Precedencies,
         Schedule_With_Resources);


   exception
      when Type_Error =>
         Result:= Result & Lb_Minus & Lb_Simulation_Error(
            Current_Language) &
            Lb_Type_Error(Current_Language)  & Lb_Comma &
            Exception_Message & Unbounded_Lf;
      when Statement_Error =>
         Result:= Result & Lb_Minus & Lb_Simulation_Error(
            Current_Language) &
            Lb_Statement_Error(Current_Language) & Lb_Comma &
            Exception_Message & Unbounded_Lf;
      when Variable_Error =>
         Result:= Result & Lb_Minus & Lb_Simulation_Error(
            Current_Language) &
            Lb_Variable_Error(Current_Language)  & Lb_Comma &
            Exception_Message & Unbounded_Lf;
      when Syntax_Error =>
         Result:= Result & Lb_Minus & Lb_Simulation_Error(
            Current_Language) &
            Lb_Syntax_Error(Current_Language) & Lb_Comma &
            Exception_Message & Unbounded_Lf;

   end Compute_Simulation_Time_Line;



   procedure Compute_Simulation_Basics (
         Sys         : in     System;                           
         Result      : in out Unbounded_String;                 
         A_Processor : in     Processor_Ptr;                    
         Output      : in     Output_Format    := String_Output ) is 

      Check : Unbounded_String;  
   begin

      if Output=Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      for J in 0.. Sched.Nb_Entries-1 loop

         if Sched.Entries(J).Data.Processor.Name=A_Processor.Name then


            if Sched.Entries(J).Data.Msg/=Empty_String then
               Result:= Result &  Sched.Entries(J).Data.Msg & Unbounded_Lf;

            else

               Result := Result &  Start_Line &  Lb_Number_Of_Preemption(
                  Current_Language) & Start_Num_Of_Preemp &  Sched.
                  Entries(J).Data.Result.
                  Number_Of_Preemption'Img & End_Num_Of_Preemp & End_Line  &
                  Unbounded_Lf;

               Check:= Empty_String;
               Response_Time_By_Simulation(Sys.Tasks, Sched.Entries(J).
                  Data.
                  Result, Sched.Entries(J).Data.Processor.Name,
                  Check);
               Result:=Result & Check  & Unbounded_Lf & End_Sched_Sim &
                  Unbounded_Lf;
            end if;
         end if;
      end loop;
   end Compute_Simulation_Basics;



   procedure Compute_Simulation_Number_Of_Preemption (
         Sys         : in     System;                           
         Result      : in out Unbounded_String;                 
         A_Processor : in     Processor_Ptr;                    
         Output      : in     Output_Format    := String_Output ) is 
   begin

      for J in 0.. Sched.Nb_Entries-1 loop

         if Sched.Entries(J).Data.Processor.Name=A_Processor.Name then

            Result := Result & Lb_Draw(Current_Language) & Lb_Comma &
               Lb_Processor(Current_Language);
            Result :=Result &
               Sched.Entries(J).Data.Processor.Name & To_Unbounded_String(
               " : ") &
               Unbounded_Lf;
            Result:=Result & Unbounded_Lf;

            if Sched.Entries(J).Data.Msg/=Empty_String then
               Result:= Result &  Sched.Entries(J).Data.Msg;
            else
               Result := Result & Lb_Number_Of_Preemption(
                  Current_Language) & Sched.
                  Entries(J).Data.Result.
                  Number_Of_Preemption'Img & Unbounded_Lf;
            end if;
         end if;
      end loop;



   end Compute_Simulation_Number_Of_Preemption;



   procedure Compute_Simulation_Response_Time (
         Sys         : in     System;                           
         Result      : in out Unbounded_String;                 
         A_Processor : in     Processor_Ptr;                    
         Output      : in     Output_Format    := String_Output ) is 

      Check : Unbounded_String := Empty_String;  

   begin

      for J in 0.. Sched.Nb_Entries-1 loop
         if Sched.Entries(J).Data.Processor.Name=A_Processor.Name then
            Response_Time_By_Simulation(Sys.Tasks, Sched.Entries(J).
               Data.Result,
               Sched.Entries(J).Data.Processor.Name,
               Check);
         end if;
      end loop;

      Result := Result  & Check & Unbounded_Lf;

   end Compute_Simulation_Response_Time;




   procedure Compute_Feasibility_Period (
         Sys         : in     System;                           
         Result      : in out Unbounded_String;                 
         A_Processor : in     Processor_Ptr;                    
         Output      : in     Output_Format    := String_Output ) is 

      Period  : Natural;  
      Util    : Double;  
      Free_Ut : Double;  

   begin

      --definition des balises

      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;


      Period:=Scheduling_Period(Sys.Tasks,
         A_Processor.Name);
      Result:= Result & Start_Line & Lb_Scheduling_Period(
         Current_Language) & Start_Bas_Per &
         Period'
         Img  & End_Bas_Per  & Start_Ref  &
         To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 6). ") & End_Ref &  End_Line &
         Unbounded_Lf;

      Util:=Processor_Utilization_Over_Period(Sys.Tasks,
         A_Processor.Name);

      Free_Ut:=Double(Period)*(1.0-Util);
      Free_Ut:=Double'Max(0.0,Free_Ut);
      Result:= Result & Start_Line & To_Unbounded_String("-")
         & Start_Unit & Integer'Image(Integer(Free_Ut)) & End_Unit &
         Lb_Free_Unit_Time(Current_Language) & End_Line  & Unbounded_Lf;

   exception
      when Task_Set.Task_Must_Be_Periodic=>
         Result:= Result & Lb_Compute_Error_2(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Task_Model_Error =>
         Result:= Result & Lb_Compute_Error_3(Current_Language) &
            Unbounded_Lf;


   end Compute_Feasibility_Period;



   procedure Compute_Feasibility_Processor_Utilization_Factor (
         Sys         : in     System;                           
         Result      : in out Unbounded_String;                 
         A_Processor : in     Processor_Ptr;                    
         Output      : in     Output_Format    := String_Output ) is 


      Util : Double;  
      Msg  : Unbounded_String := Empty_String;  

   begin


      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;


      Util:=Processor_Utilization_Over_Deadline(Sys.Tasks,
         A_Processor.Name);
      Result:= Result & Start_Line &  Lb_Utilization_With_Deadline(
         Current_Language) &
         Start_Deadl_Fact  & Format(Util) & End_Deadl_Fact &  Start_Ref&
         To_Unbounded_String(" (") & Lb_See(
         Current_Language) &
         To_Unbounded_String("[1], page 6). ") & End_Ref & End_Line &
         Unbounded_Lf;


      Util:=Processor_Utilization_Over_Period(Sys.Tasks,
         A_Processor.Name);
      Result:= Result & Start_Line & Lb_Utilization_With_Period(
         Current_Language) &
         Start_Per_Fact & Format(Util) & End_Per_Fact &
         Start_Ref & To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 6). ") & End_Ref & Unbounded_Lf;

      Bound_On_Processor_Utilization(A_Processor.Scheduler.all, Sys.
         Tasks,
         A_Processor.Name, Util, Msg);
      Result:= Result  & Start_Line & Lb_Utilization_Bound(
         Current_Language) &
         Start_Fact_Bound & Format(
         Util) & End_Fact_Bound  & Start_Ref & Msg & End_Ref & End_Line;


   exception
      when Task_Set.Task_Must_Be_Periodic=>
         Result:= Result & Lb_Compute_Error_9(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Task_Model_Error =>
         Result:= Result & Lb_Compute_Error_10(Current_Language) &
            Unbounded_Lf;
      when Invalid_Scheduler =>
         Result:= Result & Lb_Compute_Error_19(Current_Language) &
            Unbounded_Lf;


   end Compute_Feasibility_Processor_Utilization_Factor;




   procedure Compute_Feasibility_Response_Time (
         Sys         : in out System;                           
         Result      : in out Unbounded_String;                 
         A_Processor : in     Processor_Ptr;                    
         Output      : in     Output_Format    := String_Output ) is 

      Rt  : Response_Table;  
      Msg : Unbounded_String := Empty_String;  

   begin

      if Output=Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;
      Compute_Response_Time(A_Processor.Scheduler.all, Sys.Tasks,
         A_Processor.Name, Msg, Rt);

      Result:=Result & Start_Line &Lb_Task_Response_Time(Current_Language) &
         Start_Ref & Msg & End_Ref &
         Unbounded_Lf;



      for I in 0..Response_Range(Get_Number_Of_Task_From_Processor(Sys.
               Tasks,A_Processor.Name)-1) loop
         if (Rt.Entries(I).Item /= null)
               then
            Result:=Result & Start_Task & To_Unbounded_String("    ") & Rt.
               Entries(
               I).
               Item.Name & To_Unbounded_String(" =>") &
               Integer'Image(Integer(Rt.Entries(I).Data))& End_Task;

            if (Natural(Rt.Entries(I).Data) > Rt.Entries(I).Item.
                  Deadline)
                  then
               Result := Result & Lb_Comma & Lb_Check_Deadline(
                  Current_Language) & Rt.Entries(I).Item.Deadline'Img &
                  To_Unbounded_String(")");
            end if;


            Result := Result  &  Unbounded_Lf;

         end if;
      end loop;
      Result:=Result & End_Line  & Unbounded_Lf & End_Sched_Feas  &
         Unbounded_Lf;

   exception
      when Task_Set.Task_Must_Have_Period_Equal_To_Deadline=>
         Result := Result & Lb_Compute_Error_4(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Task_Must_Be_Periodic=>
         Result := Result & Lb_Compute_Error_5(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Task_Model_Error =>
         Result := Result & Lb_Compute_Error_6(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Processor_Utilization_Exceeded =>
         Result:= Result & Lb_Compute_Error_7(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Start_Time_Error =>
         Result:= Result & Lb_Compute_Error_12(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Offset_Error =>
         Result:= Result & Lb_Compute_Error_13(Current_Language) &
            Unbounded_Lf;
      when Invalid_Scheduler =>
         Result:= Result & Lb_Compute_Error_18(Current_Language) &
            Unbounded_Lf;
   end Compute_Feasibility_Response_Time;




   procedure Set_Priorities (
         Sys         : in out System;                           
         Result      : in out Unbounded_String;                 
         A_Processor : in     Processor_Ptr;                    
         A_Scheduler : in     Scheduler_Type;                   
         Output      : in     Output_Format    := String_Output ) is 


      A_Rm_Scheduler   : Rm_Scheduler;  
      A_Dm_Scheduler   : Dm_Scheduler;  
      Ok               : Boolean        := True;  
      A_Task           : Task_Ptr;  
      My_Task_Iterator : Tasks_Iterator;  

   begin

      if Output=Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      begin
         if(A_Scheduler=Rate_Monotonic) then
            Set_Priority_According_To_Rm(A_Rm_Scheduler, Sys.Tasks,
               A_Processor.Name) ;
         else
            Set_Priority_According_To_Dm(A_Dm_Scheduler, Sys.Tasks,
               A_Processor.Name) ;
         end if;


      exception
         when Task_Set.Task_Must_Be_Periodic=>
            Ok:=False;

      end;

      if Ok
            then
         if(A_Scheduler=Rate_Monotonic)
               then
            Result := Result & Start_Line &  Lb_Set_Priorities_Rm(
               Current_Language)& End_Line;
         else
            Result := Result & Start_Line &  Lb_Set_Priorities_Dm(
               Current_Language)& End_Line;
         end if;


         Result :=Result & Unbounded_Lf & Start_Line & Lb_Priorities(
            Current_Language);

         Reset_Iterator(My_Task_Iterator);
         loop
            Current_Element(Sys.Tasks, A_Task, My_Task_Iterator);

            if (A_Task.Location = A_Processor.Name)
                  then
               Result :=Result & Unbounded_Lf & Start_Task &
                  To_Unbounded_String("    ") &
                  A_Task.Name & To_Unbounded_String(" =>") &
                  To_Unbounded_String(A_Task.Priority'Img)& End_Task;
            end if;

            exit when Is_Last_Element(Sys.Tasks, My_Task_Iterator) =
               True;
            Next_Element(Sys.Tasks, My_Task_Iterator);
         end loop;
         Result:=Result & Unbounded_Lf & End_Line &Unbounded_Lf&
            End_Set_Prio;

      else
         Result:= Result &  Lb_Priorities_Error(Current_Language);
      end if;


   end Set_Priorities;




begin

   Sched:=new Scheduling_Table;
   Initialize(Sched.all);




end Call_Scheduling_Framework;




