------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------




with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;


with Unbounded_Strings;
use  Unbounded_Strings;


with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

with Config;
use Config;

with Dependencies;
use Dependencies;
use Dependencies.Dep_Set;
with Task_Dependencies;
use Task_Dependencies;

with Tasks;
use Tasks;
with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Messages;
use Messages;
with Message_Set;
use Message_Set;
use Message_Set.Generic_Message_Set;

with Scheduler;
use Scheduler;
use Scheduler.Double_Tasks_Parameters;

with Network;
use Network;
use Network.Double_Messages_Parameters;


with Processors;
use Processors;
with Processor_Set;
use Processor_Set;
use Processor_Set.Generic_Processor_Set;




package body Distributed_Scheduling is


   procedure Tindell (
         My_Tasks      : in out Tasks_Set;               
         My_Messages   : in out Messages_Set;            
         My_Deps       : in     Tasks_Dependencies;      
         My_Processors : in     Processors_Set;          
         My_Network    : in     Generic_Network'Class;   
         Response_Time :    out Scheduler.Response_Table ) is 


      Rtn,  
      Rtn1        : Scheduler.Response_Table;  
      Rt_Messages : Network.Response_Table;  



      function Get_Response_Time (
            A_Message : Message_Ptr ) 
        return Double is 
      begin
         for I in Network.Response_Range loop
            if (Rt_Messages.Entries(I).Item.Name = A_Message.Name)
                  then
               return Rt_Messages.Entries(I).Data;
            end if;
         end loop;

         raise Internal_Error;
      end Get_Response_Time;


      function Get_Response_Time (
            A_Task : Task_Ptr ) 
        return Double is 
      begin
         for I in Scheduler.Response_Range loop
            if (Rtn.Entries(I).Item.Name = A_Task.Name)
                  then
               return Rtn.Entries(I).Data;
            end if;
         end loop;

         raise Internal_Error;
      end Get_Response_Time;


      procedure Inject_Jitter_Into_Successors (
            Task1 : in     Tasks_Set;  
            Msg1  : in     Depends_Set ) is 

         Max_Jitter : Double           := 0.0;  
         Ite1       : Depends_Iterator;  
         Ite2       : Tasks_Iterator;  
         M1         : Dependency_Ptr;  
         T1         : Task_Ptr;  

      begin
         -- find maximum response time of all messages
         --
         Reset_Iterator(Ite1);
         loop
            Current_Element(Msg1, M1, Ite1);

            Max_Jitter:=Double'Max(Max_Jitter,
               Get_Response_Time(Message_Ptr(M1)));

            exit when Is_Last_Element(Msg1, Ite1) = True;
            Next_Element(Msg1, Ite1);
         end loop;

         -- set jitter with max_jitter
         --
         Reset_Iterator(Ite2);
         loop
            Current_Element(Task1, T1, Ite2);

            Task_Set.Set(My_Tasks, T1.Name,
               Task_Set.Jitter, Natural(Max_Jitter));

            exit when Is_Last_Element(Task1, Ite2) = True;
            Next_Element(Task1, Ite2);
         end loop;

      end Inject_Jitter_Into_Successors;



      procedure Inject_Jitter_Into_Messages (
            Msg1 : in     Depends_Set; 
            T1   : in     Task_Ptr     ) is 

         Ite       : Depends_Iterator;  
         M1        : Dependency_Ptr;  
         T1_Jitter : Double;  
      begin

         T1_Jitter:=Get_Response_Time(T1);

         Reset_Iterator(Ite);
         loop
            Current_Element(Msg1, M1, Ite);

            -- look for response time of t1 in rtn1
            --
            Message_Set.Set(My_Messages, Message_Ptr(M1).Name,
               Message_Set.Jitter,
               Natural(T1_Jitter));

            exit when Is_Last_Element(Msg1, Ite) = True;
            Next_Element(Msg1, Ite);
         end loop;
      end Inject_Jitter_Into_Messages;


      procedure Compute is 

         I     : Natural;  
         Dummy : Unbounded_String := Empty_String;  

         K,  
         J           : Scheduler.Response_Range;  
         A_Processor : Processor_Ptr;  
         My_Iterator : Processor_Set.Generic_Processor_Set.Iterator;  

         type Rt is array (0 .. Max_Processors - 1) of Scheduler.Response_Table; 
         Rt_By_Processor : Rt;  

      begin
         -- Compute tasks and messages response time
         I:=0;
         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Processors, A_Processor, My_Iterator);

            Compute_Response_Time(A_Processor.Scheduler.all, My_Tasks,
               A_Processor.Name, Dummy, Rt_By_Processor(I));

            exit when Is_Last_Element(My_Processors, My_Iterator) = True;

            Next_Element(My_Processors, My_Iterator);
            I:=I+1;
         end loop;

         Compute_Transmission_Delay(My_Network,My_Messages,Rt_Messages);

         -- gather response time information
         --
         Reset_Iterator(My_Iterator);

         I:=0;
         K:=0;

         loop
            J:=0;
            Current_Element(My_Processors, A_Processor, My_Iterator);

            for L in 0 .. (Get_Number_Of_Task_From_Processor(My_Tasks,
                     A_Processor.Name)-1) loop
               Rtn.Entries(K):=Rt_By_Processor(I).Entries(J);
               K:=K+1;
               J:=J+1;
            end loop;

            exit when Is_Last_Element(My_Processors, My_Iterator) = True;

            Next_Element(My_Processors, My_Iterator);
            I:=I+1;


         end loop;

      end Compute;




      procedure Inject_Jitter is 


         T1,  
         T2    : Task_Ptr;  
         Task2 : Tasks_Set;  
         Msg1  : Depends_Set;  
         Ite1,  
         Ite2  : Tasks_Iterator;  

      begin
         -- for each task, look for its successor and insert jitter of messages and
         -- successors
         --
         Reset_Iterator(Ite1);

         loop
            Current_Element(My_Tasks, T1, Ite1);

            -- has successors ?
            --
            if Has_Successor(My_Deps, T1)
                  then
               Task2:=Get_Successors_List(My_Deps, T1);

               -- set messages jitter
               --
               Reset_Iterator(Ite2);
               loop
                  Current_Element(Task2, T2, Ite2);
                  Msg1:=Get_Dependencies(My_Deps, T1, T2);
                  Inject_Jitter_Into_Messages(Msg1, T1);

                  exit when Is_Last_Element(Task2, Ite2) = True;
                  Next_Element(Task2, Ite2);
               end loop;

               -- set task jitter
               Inject_Jitter_Into_Successors(Task2, Msg1);
            end if;

            exit when Is_Last_Element(My_Tasks, Ite1) = True;
            Next_Element(My_Tasks, Ite1);
         end loop;



      end Inject_Jitter;



   begin

      Initialize(Rtn);
      Initialize(Rtn1);

      Compute;


      -- Now, loop until convergence. Insert response time as jitter
      while (Rtn /= Rtn1) loop

         Rtn1:=Rtn;

         Inject_Jitter;

         Compute;

      end loop;


      Response_Time:=Rtn1;


   end Tindell;



end Distributed_Scheduling;

