------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------







use Message_Set.Generic_Message_Set;


package body Message_Set is



   procedure Fill_Descriptor (
         My_Messages : in     Messages_Set;       
         Desc        :    out Message_Descriptor; 
         A_Message   : in     Message_Ptr         ) is 

   begin
      Desc.Name := To_Bounded_String(To_String(A_Message.Name));
      Desc.Size:= A_Message.Size;

      Desc.Period:=0;
      Desc.Deadline:=0;
      Desc.Jitter:=0;

      if (A_Message.Message_Type = Aperiodic_Type)
            then
         Desc.Deadline := Aperiodic_Message_Ptr(A_Message).
            Deadline;

      else
         if (A_Message.Message_Type = Periodic_Type)
               then
            Desc.Period := Periodic_Message_Ptr(A_Message).Period;
            Desc.Deadline := Periodic_Message_Ptr(A_Message).Deadline;
            Desc.Jitter := Periodic_Message_Ptr(A_Message).Jitter;
         end if;
      end if;

   end Fill_Descriptor;


   procedure Extract_Descriptor (
         My_Messages : in     Messages_Set;       
         Desc        : in     Message_Descriptor; 
         A_Message   :    out Message_Ptr         ) is 

      New_Periodic_Message  : Periodic_Message_Ptr;  
      New_Aperiodic_Message : Aperiodic_Message_Ptr;  

   begin
      if (Desc.Period /= 0)
            then
         New_Periodic_Message := new Periodic_Message;
         New_Periodic_Message.Jitter:=Desc.Jitter;
         New_Periodic_Message.Period := Desc.Period;
         A_Message:=Message_Ptr(New_Periodic_Message);

      else
         New_Aperiodic_Message:= new Aperiodic_Message;
         A_Message:=Message_Ptr(New_Aperiodic_Message);
      end if;


      A_Message.Name:=To_Unbounded_String(To_String(Desc.Name));
      A_Message.Size:=Desc.Size;
      A_Message.Deadline:=Desc.Deadline;

   end Extract_Descriptor;


   procedure Add_Message (
         My_Messages : in out Messages_Set;     
         Name        : in     Unbounded_String; 
         Size        : in     Natural;          
         Period      : in     Natural;          
         Deadline    : in     Natural;          
         Jitter      : in     Natural           ) is 

      Dummy : Message_Ptr;  
   begin
      Add_Message(My_Messages, Dummy, Name, Size,
         Period, Deadline, Jitter);
   end Add_Message;


   procedure Add_Message (
         My_Messages : in out Messages_Set;     
         A_Message   : in out Message_Ptr;      
         Name        : in     Unbounded_String; 
         Size        : in     Natural;          
         Period      : in     Natural;          
         Deadline    : in     Natural;          
         Jitter      : in     Natural           ) is 

      My_Iterator : Iterator;  

      Desc : Message_Descriptor;  

   begin

      if (Name = "")
            then
         raise Invalid_Parameter;
      end if;

      if (Size = 0)
            then
         raise Invalid_Parameter;
      end if;

      if (Jitter > Deadline)
            then
         raise Invalid_Parameter;
      end if;

      if(Get_Number_Of_Elements(My_Messages) > 0)
            then

         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Messages, A_Message, My_Iterator);
            if (Name = A_Message.Name)
                  then
               raise Message_Already_Defined;
            end if;

            exit when Is_Last_Element(My_Messages, My_Iterator) = True;

            Next_Element(My_Messages, My_Iterator);
         end loop;
      end if;


      Desc.Name:=To_Bounded_String(To_String(Name));
      Desc.Size:=Size;
      Desc.Period:=Period;
      Desc.Deadline:=Deadline;
      Desc.Jitter:=Jitter;

      Extract_Descriptor(My_Messages, Desc, A_Message);


      Add(My_Messages,A_Message);


   end Add_Message;

   procedure Update_Message (
         My_Messages : in out Messages_Set;     
         Name        : in     Unbounded_String; 
         New_Name    : in     Unbounded_String; 
         Size        : in     Natural;          
         Period      : in     Natural;          
         Deadline    : in     Natural;          
         Jitter      : in     Natural           ) is 
      A_Message          : Message_Ptr;  
      A_Periodic_Message : Periodic_Message_Ptr;  
   begin
      A_Message:=Search_Message(My_Messages,Name);
      A_Message.Name:=New_Name;
      A_Message.Size:=Size;
      A_Message.Deadline:=Deadline;
      if(A_Message.Message_Type=Periodic_Type) then
         A_Periodic_Message := Periodic_Message_Ptr(A_Message);
         A_Periodic_Message.Period:=Period;
         A_Periodic_Message.Jitter:=Jitter;
      end if;

   end Update_Message;


   function Have_Deadlines_Equal_Than_Periods (
         My_Messages    : in     Messages_Set;    
         Processor_Name : in     Unbounded_String ) 
     return Boolean is 

      Equal_Than  : Boolean     := True;  
      A_Message   : Message_Ptr;  
      My_Iterator : Iterator;  

   begin

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Messages,A_Message,My_Iterator);

         if (A_Message.Message_Type /= Periodic_Type)
               then
            return False;
         end if;

         if  (Periodic_Message_Ptr(A_Message).Deadline =
               Periodic_Message_Ptr(A_Message).Period)
               then
            Equal_Than := False;
         end if;

         exit when Is_Last_Element(My_Messages, My_Iterator) = True;

         Next_Element(My_Messages, My_Iterator);

      end loop;

      return Equal_Than;

   end Have_Deadlines_Equal_Than_Periods ;



   procedure Get_Global_Message_Type (
         My_Messages     : in     Messages_Set;     
         Processor_Name  : in     Unbounded_String; 
         Has_Global_Type :    out Boolean;          
         Global_Type     :    out Messages_Type     ) is 

      A_Message   : Message_Ptr;  
      My_Iterator : Iterator;  

   begin

      Has_Global_Type:=True;
      Reset_Iterator(My_Iterator);


      -- find the first message of the processor
      Current_Element(My_Messages,A_Message,My_Iterator);

      Global_Type := A_Message.Message_Type;
      if Is_Last_Element(My_Messages, My_Iterator) = True
            then
         return;
      else
         Next_Element(My_Messages, My_Iterator);
      end if;

      loop
         Current_Element(My_Messages,A_Message,My_Iterator);

         if (Global_Type /= A_Message.Message_Type)
               then
            Has_Global_Type:=False;
            return;
         end if;

         exit when Is_Last_Element(My_Messages, My_Iterator) = True;

         Next_Element(My_Messages, My_Iterator);

      end loop;

   end Get_Global_Message_Type;




   function Search_Message (
         My_Messages : in     Messages_Set;    
         Name        : in     Unbounded_String ) 
     return Message_Ptr is 

      My_Iterator : Iterator;  
      A_Message   : Message_Ptr;  
      Result      : Message_Ptr;  

      Found : Boolean := False;  

   begin

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Messages,A_Message,My_Iterator);

         if(A_Message.Name = Name)
               then
            Found:=True;
            Result:=A_Message;
         end if;

         exit when Is_Last_Element(My_Messages, My_Iterator) = True;

         Next_Element(My_Messages, My_Iterator);

      end loop;

      if not Found
            then
         raise Message_Not_Found;
      end if;

      return Result;
   end Search_Message;


   function Increasing_Period (
         Op1 : in     Message_Ptr; 
         Op2 : in     Message_Ptr  ) 
     return Boolean is 
   begin
      return (Periodic_Message_Ptr(Op1).Period <= Periodic_Message_Ptr(
            Op2).Period);
   end Increasing_Period;


   function Decreasing_Period (
         Op1 : in     Message_Ptr; 
         Op2 : in     Message_Ptr  ) 
     return Boolean is 
   begin
      return (Periodic_Message_Ptr(Op1).Period >= Periodic_Message_Ptr(
            Op2).Period);
   end Decreasing_Period;


   function Increasing_Deadline (
         Op1 : in     Message_Ptr; 
         Op2 : in     Message_Ptr  ) 
     return Boolean is 
   begin
      return (Periodic_Message_Ptr(Op1).Deadline <= Periodic_Message_Ptr(
            Op2).Deadline);
   end Increasing_Deadline;


   function Decreasing_Deadline (
         Op1 : in     Message_Ptr; 
         Op2 : in     Message_Ptr  ) 
     return Boolean is 
   begin
      return (Periodic_Message_Ptr(Op1).Deadline >= Periodic_Message_Ptr(
            Op2).Deadline);
   end Decreasing_Deadline;


   function Get (
         My_Messages  : in     Messages_Set;      
         Message_Name : in     Unbounded_String;  
         Param_Name   : in     Message_Parameters ) 
     return Natural is 

      A_Message   : Message_Ptr;  
      My_Iterator : Iterator;  

   begin

      if ( ( Param_Name /= Size) and
            (Param_Name /= Period) and
            (Param_Name /= Deadline) and
            (Param_Name /= Jitter) )
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Messages,A_Message,My_Iterator);

         if (A_Message.Name = Message_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Messages, My_Iterator) = True;

         Next_Element(My_Messages, My_Iterator);

      end loop;

      if (Param_Name = Size)
            then
         return A_Message.Size;
      else
         if (Param_Name = Deadline)
               then
            return Periodic_Message_Ptr(A_Message).Deadline;
         else
            if (Param_Name = Period)
                  then
               return Periodic_Message_Ptr(A_Message).Period;
            else
               return Periodic_Message_Ptr(A_Message).Jitter;
            end if;
         end if;
      end if;

   end Get;


   procedure Set (
         My_Messages  : in out Messages_Set;       
         Message_Name : in     Unbounded_String;   
         Param_Name   : in     Message_Parameters; 
         Param_Value  : in     Natural             ) is 
      A_Message   : Message_Ptr;  
      My_Iterator : Iterator;  

   begin

      if ( (Param_Name /= Size) and
            (Param_Name /= Period) and
            (Param_Name /= Deadline) and
            (Param_Name /= Jitter) )
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Messages,A_Message,My_Iterator);

         if (A_Message.Name = Message_Name)
               then
            if (Param_Name = Size)
                  then
               A_Message.Size:=Param_Value;
               exit;
            end if;
            if (Param_Name = Deadline)
                  then
               Periodic_Message_Ptr(A_Message).Deadline:=Param_Value;
            end if;
            if (Param_Name = Period)
                  then
               Periodic_Message_Ptr(A_Message).Period:=Param_Value;
               exit;
            end if;
            if (Param_Name = Jitter)
                  then
               Periodic_Message_Ptr(A_Message).Jitter:=Param_Value;
               exit;
            end if;
         end if;
         exit when Is_Last_Element(My_Messages, My_Iterator) = True;

         Next_Element(My_Messages, My_Iterator);

      end loop;
   end Set;



   procedure Export_Xml (
         My_Messages : in     Messages_Set; 
         Into        : in     File_Type     ) is 

      My_Iterator : Messages_Iterator;  
      A_Message   : Message_Ptr;  

   begin

      if not Is_Empty(My_Messages) then

         Put_Line(Into, "<messages>");
         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Messages,A_Message,My_Iterator);

            Put(Into,Ascii.Ht & "<message");
            if A_Message.X/=0
                  then
               Put(Into, " x=""" & A_Message.X'Img & """ ");
            end if;
            if A_Message.Y/=0
                  then
               Put(Into, "y=""" & A_Message.Y'Img & """ ");
            end if;
            Put_Line(Into, ">");




            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<name>" & To_String(
                  A_Message.Name) & "</name>");
            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<size>" & A_Message.Size'
               Img & "</size>");
            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<deadline>" & A_Message.
               Deadline'Img & "</deadline>");


            if (A_Message.Message_Type = Periodic_Type)
                  then
               Put_Line(Into,Ascii.Ht & Ascii.Ht & "<period>" &
                  Periodic_Message_Ptr(A_Message).Period'Img & "</period>");
               Put_Line(Into,Ascii.Ht & Ascii.Ht & "<jitter>" &
                  Periodic_Message_Ptr(A_Message).Jitter'Img & "</jitter>");
            end if;

            Put_Line(Into,Ascii.Ht & "</message>");

            exit when Is_Last_Element(My_Messages, My_Iterator) = True;

            Next_Element(My_Messages, My_Iterator);

         end loop;
         Put_Line(Into, "</messages>");
      end if;

   end Export_Xml;


end Message_Set;
