------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Fixed_Strings;
use Fixed_Strings;

with Config;
use Config;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;


with Messages;
use Messages;

with Text_Io;
use Text_Io;


with Sets;


package Message_Set is


   Message_Model_Error     : exception;  
   Message_Not_Found       : exception;  
   Message_Already_Defined : exception;  
   Invalid_Parameter       : exception;  


   type Message_Descriptor is 
      record 
         Name     : Fixed_String;  
         Size     : Natural;  
         Period   : Natural;  
         Deadline : Natural;  
         Jitter   : Natural;  
      end record; 


   type Message_Parameters is 
         (Size,     
          Period,   
          Deadline, 
          Jitter); 



   package Generic_Message_Set is new Sets (
      Max_Element     => Config.Max_Messages, 
      File_Descriptor => Message_Descriptor,  
      Element         => Message_Ptr,         
      Free            => Free,                
      Copy            => Copy,                
      Put             => Put);


   use Generic_Message_Set;

   type Messages_Set is new Generic_Message_Set.Set with private; 


   subtype Messages_Range is Generic_Message_Set.Element_Range;
   subtype Messages_Iterator is   Generic_Message_Set.Iterator;



   procedure Export_Xml (
         My_Messages : in     Messages_Set; 
         Into        : in     File_Type     ); 


   procedure Fill_Descriptor (
         My_Messages : in     Messages_Set;       
         Desc        :    out Message_Descriptor; 
         A_Message   : in     Message_Ptr         ); 


   procedure Extract_Descriptor (
         My_Messages : in     Messages_Set;       
         Desc        : in     Message_Descriptor; 
         A_Message   :    out Message_Ptr         ); 


   procedure Add_Message (
         My_Messages : in out Messages_Set;     
         Name        : in     Unbounded_String; 
         Size        : in     Natural;          
         Period      : in     Natural;          
         Deadline    : in     Natural;          
         Jitter      : in     Natural           ); 

   procedure Add_Message (
         My_Messages : in out Messages_Set;     
         A_Message   : in out Message_Ptr;      
         Name        : in     Unbounded_String; 
         Size        : in     Natural;          
         Period      : in     Natural;          
         Deadline    : in     Natural;          
         Jitter      : in     Natural           ); 

   procedure Update_Message (
         My_Messages : in out Messages_Set;     
         Name        : in     Unbounded_String; 
         New_Name    : in     Unbounded_String; 
         Size        : in     Natural;          
         Period      : in     Natural;          
         Deadline    : in     Natural;          
         Jitter      : in     Natural           ); 


   function Get (
         My_Messages  : in     Messages_Set;      
         Message_Name : in     Unbounded_String;  
         Param_Name   : in     Message_Parameters ) 
     return Natural; 

   procedure Set (
         My_Messages  : in out Messages_Set;       
         Message_Name : in     Unbounded_String;   
         Param_Name   : in     Message_Parameters; 
         Param_Value  : in     Natural             ); 


   function Have_Deadlines_Equal_Than_Periods (
         My_Messages    : in     Messages_Set;    
         Processor_Name : in     Unbounded_String ) 
     return Boolean; 


   procedure Get_Global_Message_Type (
         My_Messages     : in     Messages_Set;     
         Processor_Name  : in     Unbounded_String; 
         Has_Global_Type :    out Boolean;          
         Global_Type     :    out Messages_Type     ); 


   function Search_Message (
         My_Messages : in     Messages_Set;    
         Name        : in     Unbounded_String ) 
     return Message_Ptr; 



   function Increasing_Deadline (
         Op1 : in     Message_Ptr; 
         Op2 : in     Message_Ptr  ) 
     return Boolean; 

   function Decreasing_Deadline (
         Op1 : in     Message_Ptr; 
         Op2 : in     Message_Ptr  ) 
     return Boolean; 

   function Increasing_Period (
         Op1 : in     Message_Ptr; 
         Op2 : in     Message_Ptr  ) 
     return Boolean; 

   function Decreasing_Period (
         Op1 : in     Message_Ptr; 
         Op2 : in     Message_Ptr  ) 
     return Boolean; 



private

   type Messages_Set is new Generic_Message_Set.Set with null record; 




end Message_Set;
