------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Package description :
--      This package provides tools to support multiprocessing.
--      Today, you only find here stuff to scheduling tasks in several processors
--      with dependencies constraints
------------------------------------------------------------------------------
------------------------------------------------------------------------------





with Processors;
use Processors;

with Processor_Set;
use Processor_Set;
use Processor_Set.Generic_Processor_Set;



with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Resources;
use Resources;
use Resources.Task_Lists;

with Resource_Set;
use Resource_Set;
use Resource_Set.Generic_Resource_Set;


with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;
with Ada.Strings.Unbounded.Text_Io;
use Ada.Strings.Unbounded.Text_Io;

with Config;
use Config;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Scheduling_Sequences;
use Scheduling_Sequences;


with Text_Io;
use Text_Io;

with Unbounded_Strings;
use Unbounded_Strings;

with Scheduler;
use Scheduler;

with Dependencies_Services;
use Dependencies_Services;

with Translate;
use Translate;




package body Multiprocessor_Services is



   procedure Put (
         R : in     Scheduling_Result ) is 
   begin
      Put_Line("Processor : ");
      if (R.Processor /= null)
            then
         Put_Name(R.Processor);
      end if;
      New_Line;
      Put(To_Unbounded_String("Message : ") & R.Msg);
      New_Line;
      Put(R.Result);
   end Put;


   procedure Initialize (
         R :    out Scheduling_Result ) is 
   begin
      R.Msg:=Empty_String;
      R.Processor:=null;
      Reset(R.Result);
   end Initialize;




   procedure Build_Multiprocessor_Scheduling (
         My_Processors      : in     Processors_Set;               
         My_Tasks           : in     Tasks_Set;                    
         My_Resources       : in     Resources_Set;                
         My_Dependencies    : in     Tasks_Dependencies;           
         Result             :    out Scheduling_Table_Ptr;         
         Last_Time          :        Natural;                      
         Common_Seed_Value  :        Integer              := 0;    
         With_Offsets       :        Boolean              := True; 
         With_Precedencies  :        Boolean              := True; 
         With_Resources     :        Boolean              := True; 
         Task_Specific_Seed :        Boolean              := True  ) is 



      Tmp_Tasks     : array (Scheduling_Range) of Tasks_Set;  
      Tmp_Resources : array (Scheduling_Range) of Resources_Set;  
      Si            : Scheduling_Information;  

      Iterator3 : Processors_Iterator;  

      A_Processor : Processor_Ptr;  

      Last_Current_Time : Natural                                     := 
        0;  
      Current_Time_Unit : array (Scheduling_Range) of Time_Unit_Range := 
        (others => 0);  
      Current_Time      : array (Scheduling_Range) of Natural         := 
        (others => 0);  
      Elected           : Tasks_Range                                 := 
        Tasks_Range'First;  
      Number_Of_Tasks   : Tasks_Range;  
      No_Task           : Boolean;  


   begin

      Initialize(Result.all);
      Si.Number_Of_Tasks:=0;
      Si.Number_Of_Resources:=0;


      -- Before scheduling, check if the tasks can be
      -- schedule with the selected scheduler
      --
      Reset_Iterator(Iterator3);

      loop
         Current_Element(My_Processors, A_Processor, Iterator3);

         begin

            Result.Entries(Result.Nb_Entries).Item:=A_Processor;
            Result.Entries(Result.Nb_Entries).Data.Processor:=A_Processor;
            Result.Entries(Result.Nb_Entries).Data.Msg:=Empty_String;

            if (Get_Number_Of_Task_From_Processor(My_Tasks, A_Processor.
                     Name)/=0)
                  then


               Check_Before_Scheduling(A_Processor.Scheduler.all, My_Tasks,
                  A_Processor.Name);


               if With_Precedencies=True then
                  Dependencies_Periods_Control(My_Tasks, My_Dependencies,
                     A_Processor.Name );
               end if;


            else
               Result.Entries(Result.Nb_Entries).Data.Msg:=Lb_Minus &
                  Lb_Define_Tasks_Before(Current_Language);

            end if;


         exception
            when Dependencies_Services.Dependencies_Period_Error=>
               Result.Entries(Result.Nb_Entries).Data.Msg:=
                  Lb_Precedencies_Period_Error(Current_Language) &
                  Unbounded_Lf;
            when Task_Set.Task_Must_Be_Periodic=>
               Result.Entries(Result.Nb_Entries).Data.Msg:=
                  Lb_Compute_Error_15(Current_Language) &
                  Unbounded_Lf;
            when Task_Set.Task_Must_Have_Period_Equal_To_Deadline=>
               Result.Entries(Result.Nb_Entries).Data.Msg:=
                  Lb_Compute_Error_14(Current_Language) &
                  Unbounded_Lf;
            when Task_Set.Task_Model_Error =>
               Result.Entries(Result.Nb_Entries).Data.Msg:=
                  Lb_Compute_Error_1(Current_Language) &
                  Unbounded_Lf;
            when Task_Set.Priority_Error =>
               Result.Entries(Result.Nb_Entries).Data.Msg:=
                  Lb_Compute_Error_11(Current_Language) &
                  Unbounded_Lf;

         end;

         Result.Nb_Entries:=Result.Nb_Entries+1;

         exit when Is_Last_Element(My_Processors, Iterator3) = True;
         Next_Element(My_Processors, Iterator3);

      end loop;



      -- First : gather information
      --
      for J in 0..Result.Nb_Entries-1 loop



         -- Duplicate data before scheduling (if case of data are changed)
         --
         if Result.Entries(J).Data.Msg=Empty_String then

            Current_Processor_Name:=Result.Entries(J).Data.Processor.Name;

            Select_And_Copy(
               My_Tasks, Tmp_Tasks(J), Select_Cpu'access);


            if (With_Resources=True)
                  then
               if (Get_Number_Of_Elements(My_Resources)>0)
                     then
                  Select_And_Copy(
                     My_Resources, Tmp_Resources(J), Select_Cpu'access);
               end if;
            end if;

            Scheduler_Initialization (
               Result.Entries(J).Data.Processor.Scheduler.all,  Si,
               Result.Entries(J).Data.Processor.Name, Tmp_Tasks(J),
               Tmp_Resources(J), Result.Entries(J).Data.Result,
               With_Offsets, With_Precedencies, With_Resources,
               Task_Specific_Seed, Common_Seed_Value, Last_Time);


            Specific_Scheduler_Initialization(
               Result.Entries(J).Data.Processor.Scheduler.all, Si,
               Result.Entries(J).Data.Processor.Name,
               Tmp_Tasks(J), Tmp_Resources(J));



         end if;

      end loop;


      -- Second : compute scheduling
      --
      for J in 0..Result.Nb_Entries -1 loop
         Number_Of_Tasks:=Get_Number_Of_Task_From_Processor(My_Tasks,
            Result.Entries(J).Data.Processor.Name);
         if Number_Of_Tasks/=0
               then
            Current_Time_Unit(J):=Time_Unit_Range(Number_Of_Tasks);
         end if;
      end loop;


      while(Last_Current_Time < Last_Time) loop


         Last_Current_Time:=Natural'Last;

         for J in 0..Result.Nb_Entries -1 loop


            if (Result.Entries(J).Data.Msg = Empty_String) then


               if (Current_Time(J) < Last_Time) then

                  Do_Election(Result.Entries(J).Data.Processor.Scheduler.all,
                     Si,Result.Entries(J).Data.Processor.Name,
                     Current_Time(J), My_Dependencies, With_Offsets,
                     With_Precedencies, With_Resources, Elected,
                     No_Task);


                  -- No task to schedule
                  --
                  if (No_Task = True)
                        then
                     Current_Time(J):=Current_Time(J)+1;

                  else

                     Run_Task(Result.Entries(J).Data.Processor.Scheduler.all,
                        Si, Elected, Result.Entries(J).Data.Result,
                        Current_Time(J), Current_Time_Unit(J), Last_Time,
                        With_Offsets, With_Precedencies, With_Resources);

                  end if;

                  Last_Current_Time:=Natural'Min(Last_Current_Time,
                     Current_Time(J));

               end if;
            end if;

         end loop;


      end loop;

      -- FS
      -- Prevent the use of empty items in scheduling table
      --
      --      for J in reverse 0..Result.Nb_Entries -1 loop
      --         if Result.Entries(J).Data.Msg = "" then
      --            Result.Entries(J).Data.Result.Scheduled.Nb_Entries:=
      --               Result.Entries(J).Data.Result.Scheduled.Nb_Entries-1;
      --            Put(Result.Entries(J).Data.Result.Scheduled);
      --         end if;
      --      end loop;



   end Build_Multiprocessor_Scheduling;







end Multiprocessor_Services;


