------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------






with Ada.Sequential_Io;


with Io_Tools;
use Io_Tools;


package body Network.Predefined is



   procedure Initialize (
         A_Network : in out Predefined_Network ) is 
   begin
      Reset(A_Network);
   end Initialize;

   procedure Finalize (
         A_Network : in out Predefined_Network ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         A_Network : in out Predefined_Network ) is 
   begin
      null;
   end Adjust;



   type Descriptor is 
      record 
         Reception_Latency_Bound : Natural;  
         Emission_Latency_Bound  : Natural;  
         Media_Length            : Natural;  
         Name                    : Unbounded_String;  
         Number_Of_Messages      : Response_Range;  
         Rt                      : Response_Table;  
      end record; 


   procedure Dispatched_Read_From_File (
         My_Network : in out Predefined_Network; 
         File_Name  : in     Unbounded_String    ) is 

      Desc : Descriptor;  

      package Uni_File is new Ada.Sequential_Io( Descriptor );
      use Uni_File;

      A_Uni_File : Uni_File.File_Type;  
   begin
      Open(A_Uni_File,
         Mode => In_File,
         Name => To_String(File_Name));

      if not Is_Open(A_Uni_File)
            then
         raise Can_Not_Open_File_For_Reading;
      end if;

      Read(A_Uni_File,Desc);

      My_Network.Reception_Latency_Bound:=Desc.Reception_Latency_Bound;
      My_Network.Emission_Latency_Bound:=Desc.Emission_Latency_Bound;
      My_Network.Media_Length:=Desc.Media_Length;
      My_Network.Name:=Desc.Name;
      My_Network.Rt:=Desc.Rt;
      My_Network.Number_Of_Messages:=Desc.Number_Of_Messages;

      Close(A_Uni_File);

   end Dispatched_Read_From_File;


   procedure Dispatched_Write_To_File (
         My_Network : in     Predefined_Network; 
         File_Name  : in     Unbounded_String    ) is 

      Desc : Descriptor;  

      package Uni_File is new Ada.Sequential_Io( Descriptor );
      use Uni_File;

      A_Uni_File : Uni_File.File_Type;  
   begin

      Create(A_Uni_File,
         Mode => Out_File,
         Name => To_String(File_Name));

      if not Is_Open(A_Uni_File)
            then
         raise Can_Not_Open_File_For_Writing;
      end if;

      Desc.Reception_Latency_Bound:=My_Network.Reception_Latency_Bound;
      Desc.Emission_Latency_Bound:=My_Network.Emission_Latency_Bound;
      Desc.Media_Length:=My_Network.Media_Length;
      Desc.Name:=My_Network.Name;
      Desc.Rt:=My_Network.Rt;
      Desc.Number_Of_Messages:=My_Network.Number_Of_Messages;

      Write(A_Uni_File,Desc);

      Close(A_Uni_File);

   end Dispatched_Write_To_File;


   procedure Dispatched_Reset (
         My_Network : in out Predefined_Network ) is 
   begin
      My_Network.Number_Of_Messages:=0;
      My_Network.Architecture:=Unknown;
      My_Network.Name:=To_Unbounded_String(
         "Network with predefined response time");
      Initialize(My_Network.Rt);
   end Dispatched_Reset;


   procedure Dispatched_Put (
         My_Network : in     Predefined_Network ) is 
   begin
      Put(My_Network.Rt);
   end Dispatched_Put;


   procedure Dispatched_Transmission_Delay (
         My_Network    : in     Predefined_Network; 
         My_Messages   : in     Messages_Set;       
         Response_Time :    out Response_Table      ) is 

      Msg_Jitter : Double;  
   begin
      Response_Time:=My_Network.Rt;

      -- Add message jitter to predefined message response time
      --
      for I in 0..(My_Network.Number_Of_Messages-1) loop
         Msg_Jitter:=Double(Message_Set.Get(My_Messages, My_Network.Rt.
               Entries(I).
               Item.Name, Jitter));
         Response_Time.Entries(I).Data:=Msg_Jitter+Response_Time.Entries(
            I).Data;
      end loop;


   end Dispatched_Transmission_Delay;



   procedure Register_Message (
         A_Network     : in out Predefined_Network; 
         Message_Name  : in     Unbounded_String;   
         Response_Time : in     Double              ) is 
   begin
      for I in 0..(A_Network.Number_Of_Messages-1) loop
         if (A_Network.Rt.Entries(I).Item.Name = Message_Name)
               then
            A_Network.Rt.Entries(I).Data:=Response_Time;
            return;
         end if;
      end loop;

      A_Network.Rt.Entries(A_Network.Number_Of_Messages).Data:=
         Response_Time;
      A_Network.Rt.Entries(A_Network.Number_Of_Messages).Item.Name:=
         Message_Name;
      A_Network.Number_Of_Messages:=A_Network.Number_Of_Messages+1;

   end Register_Message;


   procedure Register_Messages (
         A_Network     : in out Predefined_Network; 
         Number        : in     Response_Range;     
         Response_Time : in     Response_Table      ) is 
   begin
      for I in 0..(Number-1) loop
         Register_Message(A_Network,Response_Time.Entries(I).Item.Name,
            Response_Time.Entries(I).Data);
      end loop;
   end Register_Messages;




end Network.Predefined;
