------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------





with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Ada.Finalization;

with Double_Util;

with Indexed_Tables;

with Config;

with Messages;
use Messages;

with Message_Set;
use Message_Set;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;


package Network is


   type Network_Architecture is 
         (Single_Processor, 
          Bus,              
          Token_Ring,       
          Star,             
          Unknown); 


   type Network_Parameters is 
         (Reception_Latency_Bound, 
          Emission_Latency_Bound,  
          Transmission_Bandwidth,  
          Media_Length,            
          Name,                    
          Architecture); 


   package Double_Messages_Parameters is new Indexed_Tables(Double,
      Message_Ptr,
      Config.Max_Messages, 0, Double_Util.Put, Double_Util.Initialize,
      Messages.Put_Name);

   subtype Response_Table is Double_Messages_Parameters.Indexed_Table;
   subtype Response_Range is Double_Messages_Parameters.
      Indexed_Table_Range;
   use Double_Messages_Parameters;

   Invalid_Parameter : exception;  


   type Generic_Network is abstract tagged private; 
   type Network_Ptr is access all Generic_Network'Class; 


   -- Controlled objects methods
   --
   procedure Initialize (
         My_Network : in out Generic_Network ); 

   procedure Finalize (
         My_Network : in out Generic_Network ); 

   procedure Adjust (
         My_Network : in out Generic_Network ); 


   --  Methods to set/get network parameters
   --
   function Get (
         My_Network : in     Generic_Network'Class; 
         Param_Name : in     Network_Parameters     ) 
     return Natural; 

   procedure Set (
         My_Network  : in out Generic_Network'Class; 
         Param_Name  : in     Network_Parameters;    
         Param_Value : in     Natural                ); 

   function Get (
         My_Network : in     Generic_Network'Class; 
         Param_Name : in     Network_Parameters     ) 
     return Network_Architecture; 

   function Get_Name (
         My_Network : in     Generic_Network'Class ) 
     return Unbounded_String; 




   -- Dispatching methods
   --
   procedure Write_To_File (
         My_Network : in     Generic_Network'Class; 
         File_Name  : in     Unbounded_String       ); 

   procedure Read_From_File (
         My_Network : in out Generic_Network'Class; 
         File_Name  : in     Unbounded_String       ); 

   procedure Put (
         My_Network : in     Generic_Network'Class ); 

   procedure Reset (
         My_Network : in out Generic_Network'Class ); 

   procedure Compute_Transmission_Delay (
         My_Network    : in     Generic_Network'Class; 
         My_Messages   : in     Messages_Set;          
         Response_Time :    out Response_Table         ); 




   -- And the dispached methods
   --
   procedure Dispatched_Write_To_File (
         My_Network : in     Generic_Network; 
         File_Name  : in     Unbounded_String ) is 
   abstract;

   procedure Dispatched_Read_From_File (
         My_Network : in out Generic_Network; 
         File_Name  : in     Unbounded_String ) is 
   abstract;

   procedure Dispatched_Put (
         My_Network : in     Generic_Network ) is 
   abstract;

   procedure Dispatched_Reset (
         My_Network : in out Generic_Network ) is 
   abstract;

   procedure Dispatched_Transmission_Delay (
         My_Network    : in     Generic_Network; 
         My_Messages   : in     Messages_Set;    
         Response_Time :    out Response_Table   ) is 
   abstract;



private


   type Generic_Network is abstract new Ada.Finalization.Controlled with 
      record 
         Reception_Latency_Bound : Natural;  
         Emission_Latency_Bound  : Natural;  
         Transmission_Bandwidth  : Natural;  
         Media_Length            : Natural;  
         Name                    : Unbounded_String;  
         Architecture            : Network_Architecture;  
      end record; 


end Network;

