------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Package description :
--
-- 
-- 
------------------------------------------------------------------------------
------------------------------------------------------------------------------


with Text_Io;
use Text_Io;

with Ada.Strings.Unbounded.Text_Io;
use Ada.Strings.Unbounded.Text_Io;

with Ada.Exceptions;
use Ada.Exceptions;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

with Config;
use Config;
with Translate;
use Translate;

with Task_Set;
use Task_Set;

with Integer_Util;



package body Expressions is



   function Find_Variable (
         Var : in     Variables_Table_Type; 
         S   : in     Unbounded_String      ) 
     return Variables_Range is 

   begin
      for I in 0..Var.Nb_Entries-1 loop
         if Var.Entries(I).Variable.Name=S
               then
            return I;
         end if;
      end loop;

      Raise_Exception(Expressions.Syntax_Error'Identity, To_String(
            S & Lb_Comma & Lb_Undeclared_Identifier(Current_Language) ));

   end Find_Variable;



   procedure Check_Variable_Declaration (
         Var : in     Variables_Table_Type; 
         S   : in     Unbounded_String      ) is 

      Dummy : Variables_Range;  

   begin
      Dummy:=Find_Variable(Var,S);
   end Check_Variable_Declaration;



   ---------------------------------------------


   function Value_Of (
         C : Constant_Expr;       
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr is 
   begin
      return C.Value;
   end Value_Of;


   function Value_Of (
         C : Variable_Expr;       
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr is 
   begin
      for I in 0..V.Nb_Entries-1 loop
         if V.Entries(I).Variable.Name=C.Name
               then
            return V.Entries(I).Simulation;
         end if;
      end loop;


      Raise_Exception(Variable_Error'Identity,
         To_String(Lb_Undeclared_Identifier(Current_Language) & Lb_Colon &
            C.Name));
   end Value_Of;


   function Value_Of (
         C : Array_Variable_Expr; 
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr is 

      New_Val : Simulation_Value_Ptr;  
      Index   : Simulation_Value_Ptr;  

   begin
      for I in 0..V.Nb_Entries-1 loop
         if V.Entries(I).Variable.Name=C.Name
               then

            -- Two cases : return the all task array table or 
            --		if a index is provided, return
            --		a scalar data
            --

            if (C.Array_Index/=null)


                  -- indexed expression : return scalar data
                  --
                  then

               Index:=Value_Of(C.Array_Index.all,V);
               if Index.Ptype/=Simulation_Integer_Data
                     then
                  Raise_Exception(Type_Error'Identity, To_String(
                        Lb_Index_Error(Current_Language) & Lb_Colon & C.
                        Name));
               end if;

               if C.Variable_Type=Simulation_Array_Integer_Data then
                  New_Val:=new Simulation_Value(Simulation_Integer_Data);
                  New_Val.Integer_Value:=V.Entries(I).Simulation.
                     Integer_Table_Value(Tasks_Range(Index.Integer_Value));
               end if;
               if C.Variable_Type=Simulation_Array_Double_Data then
                  New_Val:=new Simulation_Value(Simulation_Double_Data);
                  New_Val.Double_Value:=V.Entries(I).Simulation.
                     Double_Table_Value(Tasks_Range(Index.Integer_Value));
               end if;
               if C.Variable_Type=Simulation_Array_Boolean_Data then
                  New_Val:=new Simulation_Value(Simulation_Boolean_Data);
                  New_Val.Boolean_Value:=V.Entries(I).Simulation.
                     Boolean_Table_Value(Tasks_Range(Index.Integer_Value));
               end if;

               return New_Val;


            else
               -- Task array table data
               --

               if C.Variable_Type=Simulation_Array_Integer_Data then
                  New_Val:=new Simulation_Value(
                     Simulation_Array_Integer_Data);
                  New_Val.Integer_Table_Value:=V.Entries(I).Simulation.
                     Integer_Table_Value;
               end if;
               if C.Variable_Type=Simulation_Array_Double_Data then
                  New_Val:=new Simulation_Value(
                     Simulation_Array_Double_Data);
                  New_Val.Double_Table_Value:=V.Entries(I).Simulation.
                     Double_Table_Value;
               end if;
               if C.Variable_Type=Simulation_Array_Boolean_Data then
                  New_Val:=new Simulation_Value(
                     Simulation_Array_Boolean_Data);
                  New_Val.Boolean_Table_Value:=V.Entries(I).Simulation.
                     Boolean_Table_Value;
               end if;


               return New_Val;
            end if;


         end if;

      end loop;


      Raise_Exception(Variable_Error'Identity,
         To_String(Lb_Undeclared_Identifier(Current_Language) & Lb_Colon &
            C.Name));
   end Value_Of;



   function Apply_Operator (
         Lvalue : Integer; 
         Rvalue : Integer; 
         Ope    : Operator ) 
     return Integer is 

   begin

      if Ope=Min_Operator then
         if Lvalue<Rvalue then
            return Lvalue;
         else
            return Rvalue;
         end if;
      end if;
      if Ope=Max_Operator then
         if Lvalue>Rvalue then
            return Lvalue;
         else
            return Rvalue;
         end if;
      end if;
      if Ope=Lcm then
         return Integer_Util.Lcm(Lvalue,Rvalue);
      end if;
      if Ope=Plus then
         return Lvalue+Rvalue;
      end if;
      if Ope=Minus then
         return Lvalue-Rvalue;
      end if;
      if Ope=Mult then
         return Lvalue*Rvalue;
      end if;
      if Ope=Div then
         if Rvalue = 0
               then
            Raise_Exception(Variable_Error'Identity,
               To_String(Lb_Operation_Not_Permitted(Current_Language) &
                  Lb_Colon &
                  Lvalue'Img & " / " & Rvalue'Img));
         end if;
         return Lvalue/Rvalue;
      end if;

      if Rvalue = 0
            then
         Raise_Exception(Variable_Error'Identity,
            To_String(Lb_Operation_Not_Permitted(Current_Language) &
               Lb_Colon &
               Lvalue'Img & " mod " & Rvalue'Img));
      end if;

      return Lvalue mod Rvalue;
   end Apply_Operator;


   function Apply_Operator (
         Lvalue : Double;  
         Rvalue : Double;  
         Ope    : Operator ) 
     return Double is 
   begin

      if Ope=Min_Operator then
         if Lvalue<Rvalue then
            return Lvalue;
         else
            return Rvalue;
         end if;
      end if;
      if Ope=Max_Operator then
         if Lvalue>Rvalue then
            return Lvalue;
         else
            return Rvalue;
         end if;
      end if;
      if Ope=Lcm then
         return Double(Integer_Util.Lcm(Integer(Lvalue),Integer(Rvalue)));
      end if;
      if Ope=Plus then
         return Lvalue+Rvalue;
      end if;
      if Ope=Minus then
         return Lvalue-Rvalue;
      end if;
      if Ope=Mult then
         return Lvalue*Rvalue;
      end if;

      if Rvalue = 0.0
            then
         Raise_Exception(Variable_Error'Identity,
            To_String(Lb_Operation_Not_Permitted(Current_Language) &
               Lb_Colon &
               Integer'Image(Integer(Lvalue))  & " / " & Integer'Image(
                  Integer(Rvalue))));
      end if;

      return Lvalue/Rvalue;
   end Apply_Operator;


   function Apply_Operator (
         Lvalue : Double;  
         Rvalue : Double;  
         Ope    : Operator ) 
     return Boolean is 
   begin
      if Ope=Equal then
         return Lvalue=Rvalue;
      end if;
      if Ope=Not_Equal then
         return Lvalue/=Rvalue;
      end if;
      if Ope=Equal_Less then
         return Lvalue<=Rvalue;
      end if;
      if Ope=Equal_Greater then
         return Lvalue>=Rvalue;
      end if;
      if Ope=Inf then
         return Lvalue<Rvalue;
      end if;
      return Lvalue>Rvalue;
   end Apply_Operator;


   function Apply_Operator (
         Lvalue : Integer; 
         Rvalue : Integer; 
         Ope    : Operator ) 
     return Boolean is 
   begin
      if Ope=Equal then
         return Lvalue=Rvalue;
      end if;
      if Ope=Not_Equal then
         return Lvalue/=Rvalue;
      end if;
      if Ope=Equal_Less then
         return Lvalue<=Rvalue;
      end if;
      if Ope=Equal_Greater then
         return Lvalue>=Rvalue;
      end if;
      if Ope=Inf then
         return Lvalue<Rvalue;
      end if;
      return Lvalue>Rvalue;
   end Apply_Operator;


   function Apply_Operator (
         Lvalue : Boolean; 
         Rvalue : Boolean; 
         Ope    : Operator ) 
     return Boolean is 
   begin

      if Ope=Equal then
         return Lvalue=Rvalue;
      end if;
      if Ope=Not_Equal then
         return Lvalue/=Rvalue;
      end if;
      if Ope=Equal_Less then
         return Lvalue<=Rvalue;
      end if;
      if Ope=Equal_Greater then
         return Lvalue>=Rvalue;
      end if;
      if Ope=Inf then
         return Lvalue<Rvalue;
      end if;
      if Ope=Sup then
         return Lvalue>Rvalue;
      end if;
      if Ope=Logic_And then
         return Lvalue and Rvalue;
      end if;
      if Ope=Logic_Or then
         return Lvalue or Rvalue;
      end if;
      return not Lvalue;
   end Apply_Operator;




   function Compute_Composite_Value (
         Lvalue : Simulation_Value_Ptr; 
         Rvalue : Simulation_Value_Ptr; 
         Ope    : Operator              ) 
     return Simulation_Value_Ptr is 

      Res : Simulation_Value_Ptr;  

   begin


      case Lvalue.Ptype  is

         when Simulation_Integer_Data =>

            if Rvalue.Ptype=Simulation_Integer_Data then

               if (Ope=Equal) or
                     (Ope=Not_Equal) or
                     (Ope=Equal_Less) or
                     (Ope=Equal_Greater) or
                     (Ope=Inf) or
                     (Ope=Sup) then
                  Res:=new Simulation_Value(Simulation_Boolean_Data);
                  Res.Boolean_Value:=Apply_Operator(Lvalue.Integer_Value,
                     Rvalue.Integer_Value, Ope);
               else
                  Res:=new Simulation_Value(Simulation_Integer_Data);
                  Res.Integer_Value:=Apply_Operator(Lvalue.Integer_Value,
                     Rvalue.Integer_Value, Ope);
               end if;
            else

               if (Ope=Equal) or
                     (Ope=Not_Equal) or
                     (Ope=Equal_Less) or
                     (Ope=Equal_Greater) or
                     (Ope=Inf) or
                     (Ope=Sup) then
                  Res:=new Simulation_Value(
                     Simulation_Array_Boolean_Data);
               else
                  Res:=new Simulation_Value(
                     Simulation_Array_Integer_Data);
               end if;

               for I in Tasks_Range loop
                  if (Ope=Equal) or
                        (Ope=Not_Equal) or
                        (Ope=Equal_Less) or
                        (Ope=Equal_Greater) or
                        (Ope=Inf) or
                        (Ope=Sup) then
                     Res.Boolean_Table_Value(I):=Apply_Operator(
                        Lvalue.Boolean_Value,
                        Rvalue.Boolean_Table_Value(I),Ope);
                  else
                     Res.Integer_Table_Value(I):=Apply_Operator(
                        Lvalue.Integer_Value,
                        Rvalue.Integer_Table_Value(I),Ope);
                  end if;
               end loop;

            end if;


         when Simulation_Double_Data =>


            if Rvalue.Ptype=Simulation_Double_Data then

               if (Ope=Equal) or
                     (Ope=Not_Equal) or
                     (Ope=Equal_Less) or
                     (Ope=Equal_Greater) or
                     (Ope=Inf) or
                     (Ope=Sup) then
                  Res:=new Simulation_Value(Simulation_Boolean_Data);
                  Res.Boolean_Value:=Apply_Operator(Lvalue.Boolean_Value,
                     Rvalue. Boolean_Value, Ope);
               else
                  Res:=new Simulation_Value(Simulation_Double_Data);
                  Res.Double_Value:=Apply_Operator(Lvalue.Double_Value,
                     Rvalue. Double_Value, Ope);
               end if;

            else

               if (Ope=Equal) or
                     (Ope=Not_Equal) or
                     (Ope=Equal_Less) or
                     (Ope=Equal_Greater) or
                     (Ope=Inf) or
                     (Ope=Sup) then
                  Res:=new Simulation_Value(
                     Simulation_Array_Boolean_Data);
               else
                  Res:=new Simulation_Value(
                     Simulation_Array_Double_Data);
               end if;

               for I in Tasks_Range loop
                  if (Ope=Equal) or
                        (Ope=Not_Equal) or
                        (Ope=Equal_Less) or
                        (Ope=Equal_Greater) or
                        (Ope=Inf) or
                        (Ope=Sup) then
                     Res.Boolean_Table_Value(I):=Apply_Operator(
                        Lvalue.Boolean_Value,
                        Rvalue.Boolean_Table_Value(I),Ope);
                  else
                     Res.Double_Table_Value(I):=Apply_Operator(
                        Lvalue.Double_Value,
                        Rvalue.Double_Table_Value(I),Ope);
                  end if;
               end loop;

            end if;



         when Simulation_Boolean_Data =>


            if Rvalue.Ptype=Simulation_Boolean_Data then

               Res:=new Simulation_Value(Simulation_Boolean_Data);
               Res.Boolean_Value:=Apply_Operator(Lvalue.Boolean_Value,
                  Rvalue.Boolean_Value, Ope);
            else

               Res:=new Simulation_Value(
                  Simulation_Array_Boolean_Data);

               for I in Tasks_Range loop
                  Res.Boolean_Table_Value(I):=Apply_Operator(
                     Lvalue.Boolean_Value,
                     Rvalue.Boolean_Table_Value(I),Ope);
               end loop;

            end if;



         when Simulation_Array_Integer_Data =>


            if (Ope=Equal) or
                  (Ope=Not_Equal) or
                  (Ope=Equal_Less) or
                  (Ope=Equal_Greater) or
                  (Ope=Inf) or
                  (Ope=Sup) then
               Res:=new Simulation_Value(
                  Simulation_Array_Boolean_Data);
            else
               Res:=new Simulation_Value(
                  Simulation_Array_Integer_Data);
            end if;


            for I in Tasks_Range loop

               if Rvalue.Ptype=Simulation_Array_Integer_Data then

                  if (Ope=Equal) or
                        (Ope=Not_Equal) or
                        (Ope=Equal_Less) or
                        (Ope=Equal_Greater) or
                        (Ope=Inf) or
                        (Ope=Sup) then
                     Res.Boolean_Table_Value(I):=Apply_Operator(Lvalue.
                        Boolean_Table_Value(I), Rvalue.
                        Boolean_Table_Value(I), Ope);
                  else
                     Res.Integer_Table_Value(I):=Apply_Operator(Lvalue.
                        Integer_Table_Value(I), Rvalue.
                        Integer_Table_Value(I), Ope);
                  end if;

               else

                  if (Ope=Equal) or
                        (Ope=Not_Equal) or
                        (Ope=Equal_Less) or
                        (Ope=Equal_Greater) or
                        (Ope=Inf) or
                        (Ope=Sup) then
                     Res.Boolean_Table_Value(I):=Apply_Operator(Lvalue.
                        Boolean_Table_Value(I), Rvalue.Boolean_Value, Ope);
                  else
                     Res.Integer_Table_Value(I):=Apply_Operator(Lvalue.
                        Integer_Table_Value(I), Rvalue.Integer_Value, Ope);
                  end if;

               end if;

            end loop;



         when Simulation_Array_Double_Data =>


            if (Ope=Equal) or
                  (Ope=Not_Equal) or
                  (Ope=Equal_Less) or
                  (Ope=Equal_Greater) or
                  (Ope=Inf) or
                  (Ope=Sup) then
               Res:=new Simulation_Value(
                  Simulation_Array_Boolean_Data);
            else
               Res:=new Simulation_Value(
                  Simulation_Array_Double_Data);
            end if;

            for I in Tasks_Range loop

               if Rvalue.Ptype=Simulation_Array_Double_Data then
                  if (Ope=Equal) or
                        (Ope=Not_Equal) or
                        (Ope=Equal_Less) or
                        (Ope=Equal_Greater) or
                        (Ope=Inf) or
                        (Ope=Sup) then
                     Res.Boolean_Table_Value(I):=Apply_Operator(Lvalue.
                        Boolean_Table_Value(I), Rvalue.
                        Boolean_Table_Value(I),Ope);
                  else
                     Res.Double_Table_Value(I):=Apply_Operator(Lvalue.
                        Double_Table_Value(I), Rvalue.Double_Table_Value(
                           I),Ope);
                  end if;
               else
                  if (Ope=Equal) or
                        (Ope=Not_Equal) or
                        (Ope=Equal_Less) or
                        (Ope=Equal_Greater) or
                        (Ope=Inf) or
                        (Ope=Sup) then
                     Res.Boolean_Table_Value(I):=Apply_Operator(Lvalue.
                        Boolean_Table_Value(I), Rvalue.Boolean_Value, Ope);
                  else
                     Res.Double_Table_Value(I):=Apply_Operator(Lvalue.
                        Double_Table_Value(I), Rvalue.Double_Value, Ope);
                  end if;
               end if;

            end loop;


         when Simulation_Array_Boolean_Data =>


            Res:=new Simulation_Value(
               Simulation_Array_Boolean_Data);

            for I in Tasks_Range loop

               if Rvalue.Ptype=Simulation_Array_Boolean_Data then
                  Res.Boolean_Table_Value(I):=Apply_Operator(
                     Lvalue.Boolean_Table_Value(I),
                     Rvalue.Boolean_Table_Value(I),Ope);
               else
                  Res.Boolean_Table_Value(I):=Apply_Operator(
                     Lvalue.
                     Boolean_Table_Value(I),
                     Rvalue.Boolean_Value, Ope);
               end if;

            end loop;

      end case;

      return Res;

   end Compute_Composite_Value;





   function Value_Of (
         C : Composite_Expr;      
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr is 

      Res,  
      Lvalue,  
      Rvalue : Simulation_Value_Ptr;  


   begin

      Lvalue:=Value_Of(C.Lvalue.all,V);
      Rvalue:=Value_Of(C.Rvalue.all,V);


      -- First case : the two values has the same type (boolean, integer, double) and the 
      -- same size (scalar or vector)
      --
      -- Second case :  the two values has the same type but a data is scalar and 
      -- the other is a vector  
      --
      -- Last case : types are differents : it's an error !
      --


      if (
            (Rvalue.Ptype=Lvalue.Ptype)  or
            ( (Lvalue.Ptype=Simulation_Array_Integer_Data) and  (
                  Rvalue.Ptype=Simulation_Integer_Data) ) or
            ( (Lvalue.Ptype=Simulation_Array_Double_Data) and  (
                  Rvalue.Ptype=Simulation_Double_Data) ) or
            ( (Lvalue.Ptype=Simulation_Array_Boolean_Data) and  (
                  Rvalue.Ptype=Simulation_Boolean_Data) ) or
            ( (Lvalue.Ptype=Simulation_Integer_Data) and  (Rvalue.
                  Ptype=Simulation_Array_Integer_Data) )  or
            ( (Lvalue.Ptype=Simulation_Double_Data) and  (Rvalue.
                  Ptype=Simulation_Array_Double_Data) ) or
            ( (Lvalue.Ptype=Simulation_Boolean_Data) and  (Rvalue.
                  Ptype=Simulation_Array_Boolean_Data) )
            )

            then
         Res:=Compute_Composite_Value(Lvalue, Rvalue, C.Ope);

      else
         Raise_Exception(Type_Error'Identity,
            To_String(Lb_Uncompatible_Type_Error(Current_Language) &
               Lb_Colon & To_Unbounded_String(C.Lvalue.all) &
               Lb_Comma &
               To_Unbounded_String(C.Rvalue.all)));
      end if;


      return Res;

   end Value_Of;



   function Value_Of (
         C : Max_To_Index_Expr;   
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr is 
   begin
      return Value_Of(C.Lvalue.all,V);
   end Value_Of;


   function Value_Of (
         C : Min_To_Index_Expr;   
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr is 
   begin
      return Value_Of(C.Lvalue.all,V);
   end Value_Of;


   --------------------------------------------------


   procedure Put (
         S : in     Constant_Expr ) is 
   begin
      Put(To_Unbounded_String(S));
   end Put;

   procedure Put (
         S : in     Constant_Expr_Ptr ) is 
   begin
      Put(To_Unbounded_String(S.all));
   end Put;



   procedure Put (
         S : in     Variable_Expr ) is 
   begin
      Put(S.Name);
      Put("/");
      Put(S.Variable_Type'Img);
   end Put;

   procedure Put (
         S : in     Variable_Expr_Ptr ) is 
   begin
      Put(S.all);
   end Put;


   procedure Put (
         S : in     Max_To_Index_Expr ) is 
   begin
      Put("max_to_index(");
      Put(S.Lvalue.all);
      Put(")");
   end Put;

   procedure Put (
         S : in     Max_To_Index_Expr_Ptr ) is 
   begin
      Put(S.all);
   end Put;


   procedure Put (
         S : in     Min_To_Index_Expr ) is 
   begin
      Put("min_to_index(");
      Put(S.Lvalue.all);
      Put(")");
   end Put;

   procedure Put (
         S : in     Min_To_Index_Expr_Ptr ) is 
   begin
      Put(S.all);
   end Put;


   procedure Put (
         S : in     Array_Variable_Expr ) is 
   begin
      Put(S.Name);
      Put("/");
      Put(S.Variable_Type'Img);
      if (S.Array_Index/=null)  then
         Put(S.Array_Index.all);
      end if;
   end Put;

   procedure Put (
         S : in     Array_Variable_Expr_Ptr ) is 
   begin
      Put(S.all);
   end Put;


   procedure Put (
         S : in     Composite_Expr ) is 
   begin
      if (S.Lvalue/=null)  then
         Put(S.Lvalue.all);
      end if;
      Put(S.Ope);
      if (S.Rvalue/=null)  then
         Put(S.Rvalue.all);
      end if;
   end Put;

   procedure Put (
         S : in     Composite_Expr_Ptr ) is 
   begin
      Put(S.all);
   end Put;


   procedure Put (
         S : in     Operator ) is 
   begin
      case S is
         when Plus=>
            Put(" + ");
         when Minus =>
            Put(" - ");
         when Div  =>
            Put(" / ");
         when Mult =>
            Put(" * ");
         when Exp =>
            Put(" ** ");
         when Modulo =>
            Put(" mod ");
         when Equal =>
            Put(" = ");
         when Not_Equal =>
            Put(" /= ");
         when Equal_Less =>
            Put(" <= ");
         when Equal_Greater =>
            Put(" >= ");
         when Inf =>
            Put(" < ");
         when Sup =>
            Put(" > ");
         when Logic_And =>
            Put(" and ");
         when Logic_Or =>
            Put(" or ");
         when Logic_Not =>
            Put(" not ");
         when Lcm =>
            Put(" lcm ");
         when Min_Operator =>
            Put(" min ");
         when Max_Operator =>
            Put(" max ");
      end case;
   end Put;



   procedure Put (
         V : in     Variable_Record_Ptr ) is 
   begin
      Put(V.Variable);
      Put(V.Simulation);
   end Put;


   ------------------------------------------------------


   procedure Initialize (
         S : in out Generic_Expression ) is 
   begin
      null;
   end Initialize;


   procedure Finalize (
         S : in out Generic_Expression ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         S : in out Generic_Expression ) is 
   begin
      null;
   end Adjust;

   procedure Initialize (
         S : in out Constant_Expr ) is 
   begin
      S.Expression_Type:=Constant_Type;
   end Initialize;


   procedure Finalize (
         S : in out Constant_Expr ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         S : in out Constant_Expr ) is 
   begin
      null;
   end Adjust;


   procedure Initialize (
         S : in out Variable_Expr ) is 
   begin
      S.Expression_Type:=Variable_Type;
   end Initialize;


   procedure Finalize (
         S : in out Variable_Expr ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         S : in out Variable_Expr ) is 
   begin
      null;
   end Adjust;

   procedure Initialize (
         S : in out Array_Variable_Expr ) is 
   begin
      S.Expression_Type:=Array_Variable_Type;
   end Initialize;


   procedure Finalize (
         S : in out Array_Variable_Expr ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         S : in out Array_Variable_Expr ) is 
   begin
      null;
   end Adjust;

   procedure Initialize (
         S : in out Composite_Expr ) is 
   begin
      S.Expression_Type:=Composite_Type;
   end Initialize;


   procedure Finalize (
         S : in out Composite_Expr ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         S : in out Composite_Expr ) is 
   begin
      null;
   end Adjust;

   procedure Initialize (
         S : in out Max_To_Index_Expr ) is 
   begin
      S.Expression_Type:=Max_To_Index_Type;
   end Initialize;


   procedure Finalize (
         S : in out Max_To_Index_Expr ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         S : in out Max_To_Index_Expr ) is 
   begin
      null;
   end Adjust;

   procedure Initialize (
         S : in out Min_To_Index_Expr ) is 
   begin
      S.Expression_Type:=Min_To_Index_Type;
   end Initialize;


   procedure Finalize (
         S : in out Min_To_Index_Expr ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         S : in out Min_To_Index_Expr ) is 
   begin
      null;
   end Adjust;


   ---------------------------------------------------------------

   function To_Unbounded_String (
         C : Constant_Expr ) 
     return Unbounded_String is 
   begin
      return     To_Unbounded_String(C.Value);
   end To_Unbounded_String;


   function To_Unbounded_String (
         C : Variable_Expr ) 
     return Unbounded_String is 
   begin
      return C.Name;
   end To_Unbounded_String;


   function To_Unbounded_String (
         C : Array_Variable_Expr ) 
     return Unbounded_String is 
   begin
      return C.Name;
   end To_Unbounded_String;


   function To_Unbounded_String (
         C : Composite_Expr ) 
     return Unbounded_String is 
   begin
      return (  To_Unbounded_String(C.Lvalue.all) &
         To_Unbounded_String(C.
            Ope'Img) &
         To_Unbounded_String(C.Rvalue.all) );
   end To_Unbounded_String;


   function To_Unbounded_String (
         C : Max_To_Index_Expr ) 
     return Unbounded_String is 
   begin
      return ( To_Unbounded_String("max_to_index(") & To_Unbounded_String(
            C.Lvalue.all) &
         To_Unbounded_String(")" ) );
   end To_Unbounded_String;


   function To_Unbounded_String (
         C : Min_To_Index_Expr ) 
     return Unbounded_String is 
   begin
      return ( To_Unbounded_String("min_to_index(") & To_Unbounded_String(
            C.
            Lvalue.all) & To_Unbounded_String(")") );

   end To_Unbounded_String;




end Expressions;






