------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Package description :
--
-- 
-- 
------------------------------------------------------------------------------
------------------------------------------------------------------------------



with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Config;
use Config;

with Simulations;
use Simulations;

with Tables;

with Ada.Finalization;





package Expressions is




   -- The data type does not match the expected type
   --
   Type_Error : exception;  

   -- Statement exception (sush as unexpected statement)
   --
   Statement_Error : exception;  

   -- Exception raised when an error occurs
   -- on variable  operations
   --
   Variable_Error : exception;  

   -- Syntax error detected
   --
   Syntax_Error : exception;  




   ---------------------------------------------------------
   -- Valid operator used to build expressions
   --
   type Operator is 
         (Plus,          
          Minus,         
          Div,           
          Mult,          
          Exp,           
          Modulo,        
          Equal,         
          Not_Equal,     
          Equal_Less,    
          Equal_Greater, 
          Inf,           
          Sup,           
          Logic_And,     
          Logic_Or,      
          Logic_Not,     
          Min_Operator,  
          Max_Operator,  
          Lcm); 

   procedure Put (
         S : in     Operator ); 


   ---------------------------------------------------------
   -- Table to store the list of variables of
   -- a parametric scheduler. For each variable,
   -- the current simulation values are also stored
   --

   type Variable_Expr; 
   type Variable_Expr_Ptr is access all Variable_Expr'Class; 

   type Variable_Record is 
      record 
         Variable   : Variable_Expr_Ptr;  
         Simulation : Simulation_Value_Ptr;  
      end record; 
   type Variable_Record_Ptr is access Variable_Record; 


   procedure Put (
         V : in     Variable_Record_Ptr ); 


   package Variables_Type_Package is new Tables(Variable_Record_Ptr,
      Max_Variables,Put);
   use Variables_Type_Package;

   subtype Variables_Range is Variables_Type_Package.Table_Range;
   subtype Variables_Table_Type is Variables_Type_Package.Table;



   -- Looking for a variable in the variable table
   -- 
   function Find_Variable (
         Var : in     Variables_Table_Type; 
         S   : in     Unbounded_String      ) 
     return Variables_Range; 



   procedure Check_Variable_Declaration (
         Var : in     Variables_Table_Type; 
         S   : in     Unbounded_String      ); 



   ---------------------------------------------------------

   type Expressions_Type is 
         (Min_To_Index_Type,   
          Max_To_Index_Type,   
          Constant_Type,       
          Variable_Type,       
          Array_Variable_Type, 
          Composite_Type); 


   type Generic_Expression is abstract new Ada.Finalization.Controlled with 
      record 
         Expression_Type : Expressions_Type;  
      end record; 
   type Expression_Ptr is access all Generic_Expression'Class; 


   procedure Initialize (
         S : in out Generic_Expression ); 
   procedure Finalize (
         S : in out Generic_Expression ); 
   procedure Adjust (
         S : in out Generic_Expression ); 


   procedure Put (
         S : in     Generic_Expression ) is 
   abstract;
   procedure Put (
         S : in     Expression_Ptr ) is 
   abstract;


   function Value_Of (
         C : Generic_Expression;  
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr is 
   abstract;

   function To_Unbounded_String (
         C : Generic_Expression ) 
     return Unbounded_String is 
   abstract;







   ------------------------------------------

   type Constant_Expr is new Generic_Expression with 
      record 
         Value : Simulation_Value_Ptr;  
      end record; 

   type Constant_Expr_Ptr is access all Constant_Expr'Class; 

   procedure Put (
         S : in     Constant_Expr ); 
   procedure Put (
         S : in     Constant_Expr_Ptr ); 
   function Value_Of (
         C : Constant_Expr;       
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr; 
   function To_Unbounded_String (
         C : Constant_Expr ) 
     return Unbounded_String; 


   procedure Initialize (
         S : in out Constant_Expr ); 
   procedure Finalize (
         S : in out Constant_Expr ); 
   procedure Adjust (
         S : in out Constant_Expr ); 
   ------------------------------------------

   type Variable_Expr is new Generic_Expression with 
      record 

         -- Name of the variable
         --
         Name : Unbounded_String;  

         -- Type of the variable (integer, boolean, ...)
         --
         Variable_Type : Simulation_Data_Type;  
      end record; 



   procedure Put (
         S : in     Variable_Expr ); 
   procedure Put (
         S : in     Variable_Expr_Ptr ); 
   function Value_Of (
         C : Variable_Expr;       
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr; 
   function To_Unbounded_String (
         C : Variable_Expr ) 
     return Unbounded_String; 

   procedure Initialize (
         S : in out Variable_Expr ); 
   procedure Finalize (
         S : in out Variable_Expr ); 
   procedure Adjust (
         S : in out Variable_Expr ); 
   ------------------------------------------

   type Array_Variable_Expr is new Variable_Expr with 
      record 
         Array_Index : Expression_Ptr;  
      end record; 

   type Array_Variable_Expr_Ptr is access all Array_Variable_Expr'Class; 

   procedure Put (
         S : in     Array_Variable_Expr ); 
   procedure Put (
         S : in     Array_Variable_Expr_Ptr ); 
   function Value_Of (
         C : Array_Variable_Expr; 
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr; 
   function To_Unbounded_String (
         C : Array_Variable_Expr ) 
     return Unbounded_String; 

   procedure Initialize (
         S : in out Array_Variable_Expr ); 
   procedure Finalize (
         S : in out Array_Variable_Expr ); 
   procedure Adjust (
         S : in out Array_Variable_Expr ); 
   ------------------------------------------

   type Composite_Expr is new Generic_Expression with 
      record 
         Ope    : Operator;  
         Rvalue : Expression_Ptr;  
         Lvalue : Expression_Ptr;  
      end record; 
   type Composite_Expr_Ptr is access all Composite_Expr'Class; 

   procedure Put (
         S : in     Composite_Expr ); 
   procedure Put (
         S : in     Composite_Expr_Ptr ); 
   function Value_Of (
         C : Composite_Expr;      
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr; 
   function To_Unbounded_String (
         C : Composite_Expr ) 
     return Unbounded_String; 

   procedure Initialize (
         S : in out Composite_Expr ); 
   procedure Finalize (
         S : in out Composite_Expr ); 
   procedure Adjust (
         S : in out Composite_Expr ); 
   ------------------------------------------

   type Max_To_Index_Expr is new Generic_Expression with 
      record 
         Lvalue : Expression_Ptr;  
      end record; 
   type Max_To_Index_Expr_Ptr is access all Max_To_Index_Expr'Class; 

   procedure Put (
         S : in     Max_To_Index_Expr ); 
   procedure Put (
         S : in     Max_To_Index_Expr_Ptr ); 
   function Value_Of (
         C : Max_To_Index_Expr;   
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr; 
   function To_Unbounded_String (
         C : Max_To_Index_Expr ) 
     return Unbounded_String; 

   procedure Initialize (
         S : in out Max_To_Index_Expr ); 
   procedure Finalize (
         S : in out Max_To_Index_Expr ); 
   procedure Adjust (
         S : in out Max_To_Index_Expr ); 
   ------------------------------------------

   type Min_To_Index_Expr is new Generic_Expression with 
      record 
         Lvalue : Expression_Ptr;  
      end record; 
   type Min_To_Index_Expr_Ptr is access all Min_To_Index_Expr'Class; 

   procedure Put (
         S : in     Min_To_Index_Expr ); 
   procedure Put (
         S : in     Min_To_Index_Expr_Ptr ); 
   function Value_Of (
         C : Min_To_Index_Expr;   
         V : Variables_Table_Type ) 
     return Simulation_Value_Ptr; 
   function To_Unbounded_String (
         C : Min_To_Index_Expr ) 
     return Unbounded_String; 

   procedure Initialize (
         S : in out Min_To_Index_Expr ); 
   procedure Finalize (
         S : in out Min_To_Index_Expr ); 
   procedure Adjust (
         S : in out Min_To_Index_Expr ); 





end Expressions;


