------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Package description :
--  This package provides date structurs to store in memory
--  statements composing a scheduler behavior
-- 
------------------------------------------------------------------------------
------------------------------------------------------------------------------



with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Expressions;
use Expressions;

with Ada.Finalization;


package Statements is



   type Generic_Statement; 
   type Statement_Ptr is access all Generic_Statement'Class; 


   type Statements_Type is 
         (Put_Statement_Type,         
          Nop_Statement_Type,         
          If_Statement_Type,          
          Affectation_Statement_Type, 
          Forall_Statement_Type,      
          Return_Statement_Type,      
          While_Statement_Type); 



   type Generic_Statement is abstract new Ada.Finalization.Controlled with 
      record 
         Statement_Type : Statements_Type;  
         Line_Number    : Natural         := 1;  
         Next_Statement : Statement_Ptr   := null;  
      end record; 


   procedure Initialize (
         S : in out Generic_Statement ); 
   procedure Finalize (
         S : in out Generic_Statement ); 
   procedure Adjust (
         S : in out Generic_Statement ); 


   procedure Put (
         S : in     Generic_Statement ); 
   procedure Recursive_Put (
         S : in     Statement_Ptr ); 


   function Copy (
         S : in     Statement_Ptr ) 
     return Statement_Ptr; 


   ------------------------------------------

   type Nop_Statement is new Generic_Statement with null record; 

   type Nop_Statement_Ptr is access all Nop_Statement'Class; 

   procedure Initialize (
         S : in out Nop_Statement ); 
   procedure Finalize (
         S : in out Nop_Statement ); 
   procedure Adjust (
         S : in out Nop_Statement ); 


   procedure Put (
         S : in     Nop_Statement ); 



   ------------------------------------------

   type Put_Statement is new Generic_Statement with 
      record 
         Identifier : Unbounded_String;  
      end record; 

   type Put_Statement_Ptr is access all Put_Statement'Class; 

   procedure Initialize (
         S : in out Put_Statement ); 
   procedure Finalize (
         S : in out Put_Statement ); 
   procedure Adjust (
         S : in out Put_Statement ); 


   procedure Put (
         S : in     Put_Statement ); 


   ------------------------------------------

   type If_Statement is new Generic_Statement with 
      record 
         Bool_Expression : Expression_Ptr;  
         Else_Statement  : Statement_Ptr;  
         Then_Statement  : Statement_Ptr;  
      end record; 

   type If_Statement_Ptr is access all If_Statement'Class; 

   procedure Initialize (
         S : in out If_Statement ); 
   procedure Finalize (
         S : in out If_Statement ); 
   procedure Adjust (
         S : in out If_Statement ); 


   procedure Put (
         S : in     If_Statement ); 

   ------------------------------------------

   type Affectation_Statement is new Generic_Statement with 
      record 
         Lvalue : Expression_Ptr;  
         Rvalue : Expression_Ptr;  
      end record; 

   type Affectation_Statement_Ptr is access all Affectation_Statement'Class; 

   procedure Initialize (
         S : in out Affectation_Statement ); 
   procedure Finalize (
         S : in out Affectation_Statement ); 
   procedure Adjust (
         S : in out Affectation_Statement ); 

   procedure Put (
         S : in     Affectation_Statement ); 

   ------------------------------------------

   type Forall_Statement is new Generic_Statement with 
      record 
         Included_Statement : Statement_Ptr;  
         Forall_Index       : Variable_Expr_Ptr;  
      end record; 

   type Forall_Statement_Ptr is access all Forall_Statement'Class; 

   procedure Initialize (
         S : in out Forall_Statement ); 
   procedure Finalize (
         S : in out Forall_Statement ); 
   procedure Adjust (
         S : in out Forall_Statement ); 

   procedure Put (
         S : in     Forall_Statement ); 

   ------------------------------------------

   type Return_Statement is new Generic_Statement with 
      record 
         Return_Value : Expression_Ptr;  
      end record; 

   type Return_Statement_Ptr is access all Return_Statement'Class; 

   procedure Initialize (
         S : in out Return_Statement ); 
   procedure Finalize (
         S : in out Return_Statement ); 
   procedure Adjust (
         S : in out Return_Statement ); 

   procedure Put (
         S : in     Return_Statement ); 


   ------------------------------------------

   type While_Statement is new Generic_Statement with 
      record 
         Included_Statement : Statement_Ptr;  
         Condition          : Expression_Ptr;  
      end record; 

   type While_Statement_Ptr is access all While_Statement'Class; 

   procedure Initialize (
         S : in out While_Statement ); 
   procedure Finalize (
         S : in out While_Statement ); 
   procedure Adjust (
         S : in out While_Statement ); 

   procedure Put (
         S : in     While_Statement ); 











end Statements;


