------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------



use Processor_Set.Generic_Processor_Set;

with Scheduler.Dynamic_Priority.Edf;
use Scheduler.Dynamic_Priority.Edf;
with Scheduler.Dynamic_Priority.Llf;
use Scheduler.Dynamic_Priority.Llf;
with Scheduler.Fixed_Priority.Dm;
use Scheduler.Fixed_Priority.Dm;
with Scheduler.Fixed_Priority.Rm;
use Scheduler.Fixed_Priority.Rm;
with Scheduler.Fixed_Priority.Hpf;
use Scheduler.Fixed_Priority.Hpf;
with Scheduler.Parametrics;
use Scheduler.Parametrics;

with Config;
use Config;


package body Processor_Set is




   procedure Fill_Descriptor (
         My_Processor : in     Processors_Set;       
         Desc         :    out Processor_Descriptor; 
         A_Processor  : in     Processor_Ptr         ) is 

   begin

      Desc.Quantum:=Get_Quantum(A_Processor.Scheduler.all);
      Desc.Processor_Name := To_Bounded_String(To_String(A_Processor.Name));
      Desc.Is_Preemptive:=Get_Preemptive(A_Processor.Scheduler.all);
      Desc.Scheduler_Name:=Get_Name(A_Processor.Scheduler.all);

   end Fill_Descriptor;


   procedure Extract_Descriptor (
         My_Processor : in     Processors_Set;       
         Desc         : in     Processor_Descriptor; 
         A_Processor  :    out Processor_Ptr         ) is 

      Scheduler : Scheduler_Ptr;  

   begin
      A_Processor := new Processor;

      if (Desc.Scheduler_Name = Earliest_Deadline_First)
            then
         Scheduler:= new Edf_Scheduler;
      else
         if (Desc.Scheduler_Name = Least_Laxity_First)
               then
            Scheduler:= new Llf_Scheduler;
         else
            if (Desc.Scheduler_Name = Rate_Monotonic)
                  then
               Scheduler:= new Rm_Scheduler;
            else
               if (Desc.Scheduler_Name = Deadline_Monotonic)
                     then
                  Scheduler:= new Dm_Scheduler;
               else
                  if (Desc.Scheduler_Name = Highiest_Priority_First)
                        then
                     Scheduler:= new Hpf_Scheduler;
                  else
                     if (Desc.Scheduler_Name = Parametric)
                           then
                        Scheduler:= new Parametric_Scheduler;
                     end if;
                  end if;
               end if;
            end if;
         end if;
      end if;


      Set_Preemptive(Scheduler.all, Desc.Is_Preemptive);
      Set_Quantum(Scheduler.all, Desc.Quantum);

      A_Processor.Scheduler:=Scheduler;
      A_Processor.Name:=To_Unbounded_String(To_String(Desc.Processor_Name));


   end Extract_Descriptor;


   procedure Add_Processor (
         My_Processors : in out Processors_Set;   
         Name          : in     Unbounded_String; 
         A_Scheduler   : in     Scheduler_Ptr     ) is 

      My_Iterator : Iterator;  
      A_Processor : Processor_Ptr;  

   begin

      if (Name = "")
            then
         raise Invalid_Parameter;
      end if;
      if (A_Scheduler = null)
            then
         raise Invalid_Parameter;
      end if;


      if(Get_Number_Of_Elements(My_Processors) > 0)
            then

         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Processors, A_Processor, My_Iterator);
            if (Name = A_Processor.Name)
                  then
               raise Processor_Already_Defined;
            end if;

            exit when Is_Last_Element(My_Processors, My_Iterator) = True;

            Next_Element(My_Processors, My_Iterator);


         end loop;

      end if;


      A_Processor:= new Processor;
      A_Processor.Name := To_Unbounded_String(To_String(Name));
      A_Processor.Scheduler := A_Scheduler;

      Add(My_Processors, A_Processor);


   end Add_Processor;




   function Get (
         My_Processors  : in     Processors_Set;      
         Processor_Name : in     Unbounded_String;    
         Param_Name     : in     Processor_Parameters ) 
     return Scheduler_Ptr is 

      A_Processor : Processor_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Scheduler)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Processors,A_Processor,My_Iterator);

         if (A_Processor.Name = Processor_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Processors, My_Iterator) = True;

         Next_Element(My_Processors, My_Iterator);

      end loop;

      return A_Processor.Scheduler;

   end Get;



   procedure Set (
         My_Processors  : in out Processors_Set;       
         Processor_Name : in     Unbounded_String;     
         Param_Name     : in     Processor_Parameters; 
         Param_Value    : in     Scheduler_Ptr         ) is 

      A_Processor : Processor_Ptr;  
      My_Iterator : Iterator;  
   begin

      if (Param_Name /= Scheduler)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Processors,A_Processor,My_Iterator);

         if (A_Processor.Name = Processor_Name)
               then
            A_Processor.Scheduler:=Param_Value;
            exit;
         end if;

         exit when Is_Last_Element(My_Processors, My_Iterator) = True;

         Next_Element(My_Processors, My_Iterator);

      end loop;
   end Set;



   procedure Read_From_File (
         My_Processor : in out Processors_Set'Class; 
         File_Name    : in     Unbounded_String      ) is 

      A_Processor : Processor_Ptr;  
      My_Iterator : Processors_Iterator;  
      Behavior    : Unbounded_String;  

   begin
      Read_From_File( Generic_Processor_Set.Set(My_Processor), File_Name);


      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Processor, A_Processor, My_Iterator);


         if Get_Name(A_Processor.Scheduler.all)=Parametric
               then
            Behavior:=Read_Sequential_File( File_Name &
               To_Unbounded_String(".") & A_Processor.Name &
               To_Unbounded_String(".sc"));


            Set_String_Behavior (
               Parametric_Scheduler(A_Processor.Scheduler.all), Behavior);

         end if;


         exit when Is_Last_Element(My_Processor, My_Iterator) = True;

         Next_Element(My_Processor, My_Iterator);

      end loop;


   end Read_From_File;



   procedure Write_To_File (
         My_Processor : in     Processors_Set'Class; 
         File_Name    : in     Unbounded_String;     
         Write_Mode   : in     Io_Tools.File_Mode    ) is 


      A_Processor : Processor_Ptr;  
      My_Iterator : Processors_Iterator;  

   begin
      Write_To_File( Generic_Processor_Set.Set(My_Processor), File_Name,
         Write_Mode);



      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Processor, A_Processor, My_Iterator);


         if Get_Name(A_Processor.Scheduler.all)=Parametric
               then
            Write_Sequential_File(File_Name & To_Unbounded_String(".") &
               A_Processor.Name & To_Unbounded_String(".sc"),
               Get_String_Behavior( Parametric_Scheduler(A_Processor.
                     Scheduler.all)) );
         end if;

         exit when Is_Last_Element(My_Processor, My_Iterator) = True;

         Next_Element(My_Processor, My_Iterator);

      end loop;



   end Write_To_File;



   procedure Export_Xml (
         My_Processors : in     Processors_Set; 
         Into          : in     File_Type       ) is 




      A_Processor   : Processor_Ptr;  
      My_Iterator   : Processors_Iterator;  
      Is_Preemptive : Preemptive_Type;  
      Name          : Scheduler_Type;  
      File          : Unbounded_String;  
      Quantum       : Natural;  

   begin

      if not Is_Empty(My_Processors) then

         Put_Line(Into, "<processors>");
         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Processors, A_Processor, My_Iterator);

            Put_Line(Into,Ascii.Ht & "<processor>");
            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<name>" & To_String(
                  A_Processor.Name) & "</name>");
            Put(Into,Ascii.Ht & Ascii.Ht & "<scheduler" );

            Quantum:=Get_Quantum(A_Processor.Scheduler.all);
            if Quantum/=0
                  then
               Put(Into," quantum=""" & Quantum'Img & """");
            end if;

            Is_Preemptive:=Get_Preemptive(A_Processor.Scheduler.all);
            if Is_Preemptive /= Preemptive
                  then
               Put(Into," is_preemptive=""" & Is_Preemptive'Img & """");
            end if;

            Name:=Get_Name(A_Processor.Scheduler.all);

            if Name=Parametric
                  then
               File:=Get_Behavior_File_Name(Parametric_Scheduler(
                     A_Processor.Scheduler.all));

               if File=To_Unbounded_String("") then
                  File:=A_Processor.Name & ".parametric-cpu.sc";
               end if;

               Put(Into," parametric_file_name=""" & To_String(File) &
                  """");

               begin
                  Write_Sequential_File(File,
                     Get_String_Behavior(Parametric_Scheduler(A_Processor.
                           Scheduler.all)) );
               exception
                  when others =>
                     Put_Line("Warning : can not save parametric file : " &
                        To_String(File));
               end;

            end if;

            Put(Into,">" );
            Put(Into, " " & Name'Img & " ");
            Put_Line(Into, "</scheduler>" );
            Put_Line(Into,Ascii.Ht & "</processor>");


            exit when Is_Last_Element(My_Processors, My_Iterator) = True;

            Next_Element(My_Processors, My_Iterator);

         end loop;
         Put_Line(Into, "</processors>");
      end if;


   end Export_Xml;






end Processor_Set;
