------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_Io;
use Text_Io;

with Random_Tools;
use Random_Tools;

package body Queueing_System.Mg1 is

   -- return the next customer arrival instant
   --
   procedure Next_Customer_Arrival (
         A_Queueing_System : in out Mg1_Queueing_System; 
         New_Arrival       : in out Double;              
         Arrival_Rate      : in     Double;              
         Seed              : in     Generator;           
         Old_Arrival       : in     Double               ) is 

   begin
      New_Arrival := Exponential_Inter_Time (A_Queueing_System,
         Arrival_Rate,Seed, Old_Arrival);

      -- Put_Line("production : " & New_Arrival'img );
   end Next_Customer_Arrival;

   -- return the next customer departure instant
   --
   procedure Next_Customer_Departure (
         A_Queueing_System    : in out Mg1_Queueing_System; 
         New_Arrival          : in out Double;              
         Departure_Rate       : in     Double;              
         Seed                 : in     Generator;           
         Old_Departure        : in     Double;              
         Buffer_Size          : in     Natural;             
         Last_Message_Arrival : in     Double               ) is 

      Resp_Time         : Double;  
      Server_Inter_Time : Double;  
      Xi                : Double;  
   begin
      -- initialization
      Server_Inter_Time := 1.0/Departure_Rate;
      Resp_Time := Double(Get_Rand_Parameter (0,Natural(Server_Inter_Time),
            Seed));
      Xi := 0.0;
      if (Buffer_Size = 0)
            then
         Xi := Resp_Time;
         New_Arrival := Last_Message_Arrival + Xi;
      else
         if (Buffer_Size>0)
               then
            Xi := Server_Inter_Time - A_Queueing_System.Last_Inter_Time +
               Resp_Time;
            New_Arrival := Old_Departure + Xi;
         end if;
      end if;

      -- save the current response time for the next service time
      A_Queueing_System.Last_Inter_Time := Resp_Time;

      -- store the xi to compute average and variance
      A_Queueing_System.Service_Resp_Time.Nb_Entries :=
         A_Queueing_System.Service_Resp_Time.Nb_Entries + 1;
      A_Queueing_System.Service_Resp_Time.Entries(A_Queueing_System.
         Service_Resp_Time.Nb_Entries).Data := Xi;

   end Next_Customer_Departure;

   -- return the theoric average waiting time of one customer in the queue
   --
   function Qs_Average_Waiting_Time (
         A_Queueing_System : in     Mg1_Queueing_System; 
         Arrival_Rate      : in     Double;              
         Departure_Rate    : in     Double               ) 
     return Double is 

      Result         : Double;  
      Rau            : Double;  
      Variation_Coef : Double;  
   begin
      -- initialization
      Rau := Arrival_Rate / Departure_Rate;
      Variation_Coef :=
         A_Queueing_System.Var_Service / ( A_Queueing_System.Avg_Service *
         A_Queueing_System.Avg_Service);

      -- compute the result
      Result :=
         A_Queueing_System.Avg_Service +Rau * A_Queueing_System.
         Avg_Service * (1.0 + Variation_Coef) / (2.0*(1.0 - Rau));

      -- return the result
      return(Result);
   end Qs_Average_Waiting_Time;

   -- return the theoric average number of customer in the queue
   --
   function Qs_Average_Number_Customer (
         A_Queueing_System : in     Mg1_Queueing_System; 
         Arrival_Rate      : in     Double;              
         Departure_Rate    : in     Double               ) 
     return Double is 

      Result         : Double;  
      Rau            : Double;  
      Variation_Coef : Double;  
   begin
      -- initialization
      Rau := Arrival_Rate / Departure_Rate;
      Variation_Coef :=
         A_Queueing_System.Var_Service / ( A_Queueing_System.Avg_Service *
         A_Queueing_System.Avg_Service);

      -- compute the result
      Result := Arrival_Rate * A_Queueing_System.Avg_Service
         + Rau* Arrival_Rate * A_Queueing_System.Avg_Service * (1.0 +
         Variation_Coef) / (2.0*(1.0 - Rau));

      -- return the result
      return(Result);
   end Qs_Average_Number_Customer;

end Queueing_System.Mg1;



