------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_Io;
use Text_Io;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

package body Queueing_System.Mms is

   -- return the theoric average waiting time of one customer in the queue
   --
   function Qs_Average_Waiting_Time (
         A_Queueing_System : in     Mms_Queueing_System; 
         Arrival_Rate      : in     Double;              
         Service_Rate      : in     Double               ) 
     return Double is 

      Result : Double;  
   begin
      Result := 1.0 / (Service_Rate - Arrival_Rate) - (1.0 / Service_Rate);
      return(Result);
   end Qs_Average_Waiting_Time;

   -- return the theoric average number of customer in the queue
   --
   function Qs_Average_Number_Customer (
         A_Queueing_System : in     Mms_Queueing_System; 
         Arrival_Rate      : in     Double;              
         Service_Rate      : in     Double               ) 
     return Double is 
      Utilisation : Double := Arrival_Rate / Service_Rate;  
      Result      : Double;  
   begin
      Result := Utilisation / (1.0 - Utilisation);

      --      Put_Line("average waiting time: " & result'img);
      return(Result);

   end Qs_Average_Number_Customer;

   -- return the probability to be in a certain state
   --
   function Get_Probability_Of_State (
         A_Queueing_System : in     Mms_Queueing_System; 
         N                 : in     Natural              ) 
     return Double is 
      Utilisation : Double;  
      Result      : Double;  
   begin
      Utilisation := A_Queueing_System.Arrival_Rate.Entries(0).Data /
         A_Queueing_System.Service_Rate.Entries(0).Data;
      Result := Pow(Utilisation,Double(N));
      Result := Result * (1.0 - Utilisation);

      --      Put_Line("average waiting time: " & result'img);
      return (Result);
   end;

   -- return the utilistation of queue
   --
   function Get_Utilisation (
         A_Queueing_System : in     Mms_Queueing_System ) 
     return Double is 
      Result : Double;  
   begin
      Result := A_Queueing_System.Arrival_Rate.Entries(0).Data /
         (A_Queueing_System.Service_Rate.Entries(0).Data *
         Double(A_Queueing_System.Nb_Tasks));

      return (Result);
   end;

   -- return the number of tasks for mms queueing system
   --
   function Get_Qs_Nb_Tasks (
         A_Queueing_System : in     Mms_Queueing_System ) 
     return Double is 
   begin
      return  A_Queueing_System.Nb_Tasks;
   end;

   -- set the number of tasks for mms queueing system
   --
   procedure Set_Qs_Nb_Tasks (
         A_Queueing_System : in out Mms_Queueing_System; 
         Value             :        Double               ) is 
   begin
      A_Queueing_System.Nb_Tasks := Value;
   end;

   --P[queueing], Eq. 2.148, Robertazzi
   --
   function Get_Probability_Of_Queueing (
         A_Queueing_System : in     Mms_Queueing_System ) 
     return Double is 
      Utilisation  : Double;  
      Adivs        : Double;  
      Upper_Result : Double := 0.0;  
      Lower_Result : Double := 0.0;  
      Lambda       : Double := 0.0;  

      function Factorial (
            Value : in     Double ) 
        return Double is 
         N      : Double := Value;  
         Result : Double := Value;  
      begin
         while (N > 1.0) loop
            --            Put(" " & N'Img);
            N := N - 1.0;
            Result := Result * N;
         end loop;
         return Result;
      end Factorial;

   begin

      for L in 0 .. A_Queueing_System.Arrival_Rate.Nb_Entries-1 loop
         Lambda := Lambda +  A_Queueing_System.Arrival_Rate.Entries(L).
            Data;
      end loop;

      Utilisation := Lambda /
         A_Queueing_System.Service_Rate.Entries(0).Data;

      Adivs :=  Lambda /
         A_Queueing_System.Service_Rate.Entries(0).Data /
         A_Queueing_System.Nb_Tasks;

      Upper_Result := (1.0 / Factorial(A_Queueing_System.Nb_Tasks)) *
         (Pow(Adivs,A_Queueing_System.Nb_Tasks)) *
         (1.0 / (1.0 - Utilisation));

      Lower_Result := Lower_Result + 1.0;
      for N in 1 .. Natural(A_Queueing_System.Nb_Tasks)-1 loop
         Lower_Result := Lower_Result +
            (1.0 / Factorial(Double(N)) * Pow(Adivs,Double(N)));
      end loop;


      Lower_Result := Lower_Result +
         (1.0 / Factorial(A_Queueing_System.Nb_Tasks)) *
         (Pow(Adivs,A_Queueing_System.Nb_Tasks)) *
         (1.0 / (1.0 - Utilisation));

      return (Upper_Result / Lower_Result);
   end;



end Queueing_System.Mms;



