------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_Io;
use Text_Io;

with Random_Tools;
use Random_Tools;

with Lists;
with Indexed_Tables;
with Natural_Util;
use Natural_Util;
with Double_Util;
use Double_Util;




package body Queueing_System is

   -- return the next customer arrival instant
   --
   procedure Next_Customer_Arrival (
         A_Queueing_System : in out Generic_Queueing_System; 
         New_Arrival       : in out Double;                  
         Arrival_Rate      : in     Double;                  
         Seed              : in     Generator;               
         Old_Arrival       : in     Double                   ) is 

   begin
      New_Arrival := 0.0;
   end Next_Customer_Arrival;

   -- return the next customer departure instant
   --
   procedure Next_Customer_Departure (
         A_Queueing_System    : in out Generic_Queueing_System; 
         New_Arrival          : in out Double;                  
         Departure_Rate       : in     Double;                  
         Seed                 : in     Generator;               
         Old_Departure        : in     Double;                  
         Buffer_Size          : in     Natural;                 
         Last_Message_Arrival : in     Double                   ) is 

   begin
      New_Arrival := 0.0;
   end Next_Customer_Departure;

   -- return the theoric average waiting time of one customer in the queue
   --
   function Qs_Average_Waiting_Time (
         A_Queueing_System : in     Generic_Queueing_System; 
         Arrival_Rate      : in     Double;                  
         Departure_Rate    : in     Double                   ) 
     return Double is 

   begin
      return(0.0);
   end Qs_Average_Waiting_Time;

   -- return the theoric average number of customer in the queue
   --
   function Qs_Average_Number_Customer (
         A_Queueing_System : in     Generic_Queueing_System; 
         Arrival_Rate      : in     Double;                  
         Departure_Rate    : in     Double                   ) 
     return Double is 

   begin
      return(0.0);
   end Qs_Average_Number_Customer;

   function Exponential_Inter_Time (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Rate              : in     Double;                        
         Seed              : in     Generator;                     
         Old_Activation    : in     Double                         ) 
     return Double is 

      Next_Activation : Double;  

   begin
      -- compute next activation
      Next_Activation := Old_Activation + Double(Get_Inter_Arrival_Time(
            Rate,Seed));

      -- return the new date
      return(Next_Activation);

   end Exponential_Inter_Time;

   function Periodic_Inter_Time (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Rate              : in     Double;                        
         Old_Activation    : in     Double                         ) 
     return Double is 

      Next_Activation : Double;  

   begin
      -- compute next activation
      Next_Activation := Old_Activation + 1.0/Rate;

      -- return the new date
      return(Next_Activation);

   end Periodic_Inter_Time;

   function Periodic_Random_Inter_Time (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Rate              : in     Double;                        
         Old_Activation    : in     Double;                        
         Seed              : in     Generator                      ) 
     return Double is 

      Next_Activation : Double;  
      Tmp_Activation  : Double;  
      Tmp_Inter_Time  : Double;  

   begin
      -- init
      Tmp_Inter_Time := 1.0/Rate;

      -- periodic activation
      Tmp_Activation := Double'Ceiling(Double(Old_Activation) * Rate) *
         Tmp_Inter_Time;

      -- compute next activation
      Next_Activation := Tmp_Activation +
         Double(Get_Rand_Parameter (0,Natural(Tmp_Inter_Time),Seed));

      -- return the new date
      return(Next_Activation);

   end Periodic_Random_Inter_Time;

   function Constant_Inter_Time (
         A_Queueing_System    : in     Generic_Queueing_System'Class; 
         Rate                 : in     Double;                        
         Old_Activation       : in     Double;                        
         Buffer_Size          : in     Natural;                       
         Last_Message_Arrival : in     Double                         ) 
     return Double is 

      Next_Activation : Double;  

   begin
      if (Buffer_Size = 0)
            then
         Next_Activation := Last_Message_Arrival + 1.0/Rate;
      else
         if (Buffer_Size>0)
               then
            Next_Activation := Old_Activation + 1.0/Rate;
         end if;
      end if;

      -- return the new date
      return(Next_Activation);

   end Constant_Inter_Time;

   function General_Inter_Time (
         A_Queueing_System    : in     Generic_Queueing_System'Class; 
         Rate                 : in     Double;                        
         Old_Activation       : in     Double;                        
         Buffer_Size          : in     Natural;                       
         Last_Message_Arrival : in     Double                         ) 
     return Double is 

   begin
      return(0.0);
   end General_Inter_Time;


   -- return the probability to be in a certain state
   --
   function Get_Probability_Of_State (
         A_Queueing_System : in     Generic_Queueing_System; 
         N                 : in     Natural                  ) 
     return Double is 
   begin
      return 0.0;
   end;

   --Get probability density of states from 0->N
   --
   function Get_Probability_Density (
         A_Queueing_System : in     Generic_Queueing_System'Class;     
         N                 : in     Natural                       := 0 ) 
     return Double is 
      State  : Natural;  
      Result : Double  := 0.0;  
   begin
      State := N;
      if State > A_Queueing_System.Max_State then
         return (-1.0);
      end if;
      if State=0 then
         State := A_Queueing_System.Max_State;
      end if;

      for I in 0 .. State loop
         Result := Result + Get_Probability_Of_State(A_Queueing_System,I);
      end loop;

      return Result;
   end;


   -- return the probability of buffer will be full
   --
   function Get_Probability_Of_Full_Buffer (
         A_Queueing_System : in     Generic_Queueing_System ) 
     return Double is 
   begin
      return 0.0;
   end;

   -- return the utilistation of queue
   --
   function Get_Utilisation (
         A_Queueing_System : in     Generic_Queueing_System ) 
     return Double is 
   begin
      return 0.0;
   end;

   -- return the probability of all tasks are busy
   -- Eq. 2.153, Robertazzi: Queueing theory
   --
   function Get_Probability_Of_Task_Block (
         A_Queueing_System : in     Generic_Queueing_System ) 
     return Double is 
   begin
      return 0.0;
   end;

   --P[queueing], Eq. 2.148, Robertazzi
   --
   function Get_Probability_Of_Queueing (
         A_Queueing_System : in     Generic_Queueing_System ) 
     return Double is 
   begin
      return 0.0;
   end;




   procedure Simulate (
         A_Queueing_System    : in out Generic_Queueing_System'Class;     
         Arrival_Rate         : in     Arrival_Rate_Table;                
         Server_Period        : in     Natural;                           
         Nb_Inter_Time        : in     Natural;                           
         Max_Buffer_Size      :    out Natural;                           
         Avg_Buffer_Size      :    out Double;                            
         Var_Buffer_Size      :    out Double;                            
         Buffer_Max_Wait_Time :    out Double;                            
         Buffer_Avg_Wait_Time :    out Double;                            
         Dis_Buff_Size        : in     Boolean;                           
         Seed_Value           :        Natural                       := 0 ) is 

      type Activation_Array is array (1 .. Arrival_Rate.Nb_Entries) of Double; 
      type Activation_Bool_Array is array (1 .. Arrival_Rate.Nb_Entries) of Boolean; 

      procedure Put (
            E : in     Double ) is 
      begin
         Put_Line(E'Img);
      end Put;

      package Generic_Index_Lst is new Lists (
         Put     => Put,   
         Element => Double);

      use Generic_Index_Lst;
      -- return the index of the minimum response time


      procedure Min_Activation (
            Min          : in     Activation_Array;   
            Array_Number : in     Arrival_Rate_Range; 
            Minimum      :    out Double;             
            Index        :    out Arrival_Rate_Range  ) is 

      begin
         Minimum := Double'Last;
         for I in 1 .. Array_Number loop
            if (Min(I) < Minimum)
                  then
               Minimum := Min(I);
               Index := I;
            end if;
         end loop;
      end Min_Activation;

      Next_Server_Activation       : Double;  
      Next_Producer_Activation     : Activation_Array;  
      Next_Producer_Activation_Min : Double;  
      Index_Min                    : Arrival_Rate_Range;  
      Next_Producer_Activation_Ptr : Generic_Index_Lst.Element_Ptr;  
      Buffer_Size                  : Natural;  
      Generate_Next_Arrival        : Activation_Bool_Array;  
      Generate_Next_Departure      : Boolean;  
      Plot_Instant                 : Double;  
      Departure_Rate               : Double;  
      Arrival_Number               : Natural;  
      Departure_Number             : Natural;  
      Last_Buffer_Change           : Double;                        
      -- activation_array; 
      Tmp_Wait_Time                : Double;  
      Tmp_Arrival_Date             : Generic_Index_Lst.List;  
      Seed_Arrival                 : Generator;  
      Seed_Departure               : Generator;  
      -- Avg_Service_Time : Double;
      -- Seed_Resp_time : Generator;
   begin
      -- initialization
      Next_Server_Activation := 0.0;
      -- plus server_period for each activation
      for I in 1 .. Arrival_Rate.Nb_Entries loop
         Next_Producer_Activation(I) := 0.0;
         -- next message/customer arrival
         Generate_Next_Arrival(I) := True; -- first arrival of a message
      end loop;
      Next_Producer_Activation_Min := 0.0;
      Index_Min := 0;
      Last_Buffer_Change := 0.0;
      Buffer_Size := 0; -- size of the buffer
      Generate_Next_Departure := True; -- wait for a message
      Max_Buffer_Size := 0;
      Plot_Instant := 0.0;
      Departure_Rate := 1.0/Double(Server_Period);
      Avg_Buffer_Size := 0.0;
      Var_Buffer_Size := 0.0;
      Arrival_Number := 0;
      Departure_Number := 0;
      Buffer_Max_Wait_Time := 0.0;
      Buffer_Avg_Wait_Time := 0.0;
      Tmp_Wait_Time := 0.0;
      -- Avg_Service_Time := 0.0;
      --Next_Producer_Activation_Ptr := new Natural; --Generic_index_lst.Element;

      Reset(Seed_Arrival,Seed_Value);
      Reset(Seed_Departure,Seed_Value);
      -- Reset(Seed_Resp_time,Seed_value);

      -- for I in 1 .. Nb_Inter_Time loop
      while (Departure_Number<=Nb_Inter_Time) loop

         -- Generation of The Arrival and Departure instant
         for I in 0 .. Arrival_Rate.Nb_Entries-1 loop
            if (Generate_Next_Arrival(I) = True)
                  then
               Next_Customer_Arrival (A_Queueing_System,
                  Next_Producer_Activation(I),
                  Arrival_Rate.Entries(I).Data,
                  Seed_Arrival,
                  Next_Producer_Activation(I));

               Generate_Next_Arrival(I) := False;

            end if;
         end loop;

         -- test if the next event is an arrival or a departure
         Min_Activation(Next_Producer_Activation,
            Arrival_Rate.Nb_Entries,
            Next_Producer_Activation_Min,
            Index_Min);

         if (Generate_Next_Departure = True)
               then
            Next_Customer_Departure (A_Queueing_System,
               Next_Server_Activation,
               Departure_Rate,Seed_Departure,Next_Server_Activation,
               Buffer_Size,Next_Producer_Activation_Min);

            -- Avg_Service_Time := Avg_Service_Time + A_Queueing_System.Xi;
            Generate_Next_Departure := False;
         end if;

         --for I in 1 .. Arrival_Rate_number loop
         if (Next_Producer_Activation_Min <= Next_Server_Activation)
               then
            -- debug
            -- Put_Line("Arrival " & Next_Producer_Activation_Min'Img & "  Flux " & Index_Min'img );
            -- average calculation
            Avg_Buffer_Size := Avg_Buffer_Size +
               (Next_Producer_Activation_Min - Last_Buffer_Change)*Double(
               Buffer_Size);
            Last_Buffer_Change := Next_Producer_Activation_Min;

            -- arrival of a message
            Buffer_Size := Buffer_Size + 1;

            -- total number of message arrival
            Arrival_Number := Arrival_Number + 1;

            -- compute the maximum waiting time in the queue
            Next_Producer_Activation_Ptr := new Double;
            Next_Producer_Activation_Ptr.all :=
               Next_Producer_Activation_Min;
            Add (Tmp_Arrival_Date, Next_Producer_Activation_Ptr);

            -- for gnu plot
            if (Dis_Buff_Size = True)
                  then
               Plot_Instant := Next_Producer_Activation_Min;
               Put_Line(Plot_Instant'Img & " " & Buffer_Size'Img);
            end if;

            --Put_Line("# production : " & Next_Producer_Activation'Img
            -- & " " & Buffer_Size'img);

            -- find the maximum buffer size
            if (Max_Buffer_Size < Buffer_Size)
                  then
               Max_Buffer_Size := Buffer_Size;
            end if;
            Generate_Next_Arrival(Index_Min) := True;
         else
            -- Put_Line("Departure " & Next_server_Activation'Img);
            -- average calculation
            Avg_Buffer_Size := Avg_Buffer_Size +
               (Next_Server_Activation - Last_Buffer_Change)*Double(
               Buffer_Size);
            Last_Buffer_Change := Next_Server_Activation;

            -- departure of a message
            if (Buffer_Size>0)
                  then
               -- buffer size calculation
               Buffer_Size := Buffer_Size - 1;

               -- total number of departure
               Departure_Number := Departure_Number + 1;

               -- maximum waiting time of a message in the queue
               Tmp_Wait_Time := Get_Tail(Tmp_Arrival_Date).all;
               Delete(Tmp_Arrival_Date,Get_Tail(Tmp_Arrival_Date));
               -- Put_Line("prod instant " & Tmp_Wait_Time'img );

               Tmp_Wait_Time := Next_Server_Activation - Tmp_Wait_Time;
               Buffer_Avg_Wait_Time := Buffer_Avg_Wait_Time +
                  Tmp_Wait_Time;

               -- Put_Line("waiting time = " & Tmp_Wait_Time'img);
               if (Tmp_Wait_Time> Buffer_Max_Wait_Time)
                     then
                  Buffer_Max_Wait_Time := Tmp_Wait_Time;
               end if;

               -- for gnu plot
               if (Dis_Buff_Size = True)
                     then
                  Plot_Instant := Next_Server_Activation;
                  Put_Line(Plot_Instant'Img & " " & Buffer_Size'Img);
               end if;
            end if;
            --Put_Line("# consumption : " & Next_Server_Activation'Img
            --& " " & Buffer_Size'img);

            Generate_Next_Departure := True;
         end if;
      end loop;

      -- final computation of the average number
      Avg_Buffer_Size := Avg_Buffer_Size / Double(Last_Buffer_Change);
      Buffer_Avg_Wait_Time := Buffer_Avg_Wait_Time/ Double(
         Departure_Number+1);
      -- Put_Line("result simulate " & Avg_Service_Time'img);
      -- Avg_Service_Time := Avg_Service_Time / Double(Departure_Number+1);

      -- display some results
      Put_Line("# total departure number : " & Departure_Number'Img);
      -- it seems that one should add 1
      Put_Line("# total arrival number : " & Arrival_Number'Img);
      -- Put_Line("# average service time : " & Avg_Service_Time'Img);

   end Simulate;

   procedure Theoric_Results (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Arrival_Rate      : in     Arrival_Rate_Table;            
         Server_Period     : in     Natural;                       
         Rau               : in     Double                         ) is 

      Rate : Double;  
   begin
      Rate := 0.0;
      Put_Line("# * Theorical results : ");
      for I in 0 ..Arrival_Rate.Nb_Entries-1 loop

         Rate := Rate + Arrival_Rate.Entries(I).Data;
      end loop;

      Put_Line("TAW" & Rau'Img &
         Qs_Average_Waiting_Time(A_Queueing_System,
            Rate,
            1.0/Double(Server_Period))'Img);
      Put_Line("TAO" & Rau'Img &
         Qs_Average_Number_Customer(A_Queueing_System,
            Rate,
            1.0/Double(Server_Period))'Img);
      Put_Line("TU " & Rau'Img &
         Get_Utilisation(A_Queueing_System)'Img);

      Put_Line("TPS" & Rau'Img &
         Get_Probability_Of_State(A_Queueing_System,1)'Img);

      Put_Line("TPD" & Rau'Img &
         Get_Probability_Density(A_Queueing_System)'Img);

      Put_Line("TPQ" & Rau'Img &
         Get_Probability_Of_Queueing(A_Queueing_System)'Img);
      New_Line;
   end Theoric_Results;

   -- generation of departure rate and arrival rate for a queuing system
   procedure Random_Qs_Project (
         Utilization  : in     Double;             
         Arrival_Rate : in out Arrival_Rate_Table; 
         Service_Rate : in out Service_Rate_Table; 
         --         Arrival_Number        : in Natural;
         --         Departure_Rate        : out Double;
         Seed                  : in     Generator;        
         Max_Interarrival_Time : in     Natural   := 4000 ) is 

      Tmp_Max        : Double;  
      Tmp_Inter_Time : Natural;  

   begin

      -- initialisation
      Tmp_Max := Utilization;
      --      Departure_Rate := 0.0;

      -- compute each arrival rate (max (sum arrival_rate) < utilisation)
      for I in 0 .. Arrival_Rate.Nb_Entries-1 loop
         Tmp_Inter_Time := 0;
         while (Tmp_Inter_Time < 1000) loop
            Tmp_Inter_Time:= Get_Rand_Parameter (Natural(1.5/Tmp_Max),
               Max_Interarrival_Time,
               Seed);
         end loop;

         Arrival_Rate.Entries(I).Data := 1.0/Double(Tmp_Inter_Time);
         Tmp_Max := Tmp_Max - Arrival_Rate.Entries(I).Data;

         Service_Rate.Entries(0).Data := Service_Rate.Entries(0).Data +
            Arrival_Rate.Entries(I).Data;
      end loop;

      -- compute the departure rate
      Service_Rate.Entries(0).Data := Service_Rate.Entries(0).Data /
         Utilization;

   end Random_Qs_Project;

   procedure Set_Qs_Name (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Name              : in     Queueing_System_Type           ) is 
   begin
      A_Queueing_System.Name := Name;
   end Set_Qs_Name;

   function Get_Qs_Name (
         A_Queueing_System : in     Generic_Queueing_System'Class ) 
     return Queueing_System_Type is 
   begin
      return A_Queueing_System.Name;
   end Get_Qs_Name;

   procedure Set_Qs_Arrival_Rate (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Value             :        Double                         ) is 
   begin
      A_Queueing_System.Arrival_Rate.
         Entries(A_Queueing_System.Arrival_Rate.Nb_Entries).Data := Value;
      A_Queueing_System.Arrival_Rate.Nb_Entries :=
         A_Queueing_System.Arrival_Rate.Nb_Entries + 1;
   end Set_Qs_Arrival_Rate;

   function Get_Qs_Arrival_Rate (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Place             :        Arrival_Rate_Range             ) 
     return Double is 
   begin
      return A_Queueing_System.Arrival_Rate.Entries(Place).Data;
   end;

   procedure Set_Qs_Service_Rate (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Value             :        Double                         ) is 
   begin
      A_Queueing_System.Service_Rate.
         Entries(A_Queueing_System.Service_Rate.Nb_Entries).Data := Value;
      A_Queueing_System.Service_Rate.Nb_Entries :=
         A_Queueing_System.Service_Rate.Nb_Entries + 1;
   end;

   function Get_Qs_Service_Rate (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Place             :        Service_Rate_Range             ) 
     return Double is 
   begin
      return A_Queueing_System.Service_Rate.Entries(Place).Data;
   end;

   procedure Set_Qs_Nb_Of_Messages (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Value             :        Double                         ) is 
   begin
      A_Queueing_System.Nb_Of_Messages := Value;
   end;

   function Get_Qs_Nb_Of_Messages (
         A_Queueing_System : in     Generic_Queueing_System'Class ) 
     return Double is 
   begin
      return A_Queueing_System.Nb_Of_Messages;
   end;

   procedure Set_Qs_Waiting_Time (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Value             :        Double                         ) is 
   begin
      A_Queueing_System.Waiting_Time := Value;
   end;

   function Get_Qs_Waiting_Time (
         A_Queueing_System : in     Generic_Queueing_System'Class ) 
     return Double is 
   begin
      return A_Queueing_System.Waiting_Time;
   end;

   procedure Set_Qs_Max_State (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Value             :        Natural                        ) is 
   begin
      A_Queueing_System.Max_State := Value;
   end;

   function Get_Qs_Max_State (
         A_Queueing_System : in     Generic_Queueing_System'Class ) 
     return Natural is 
   begin
      return A_Queueing_System.Max_State;
   end;

   --    procedure Reset ( A_Queueing_System : in out Generic_Queueing_System) is
   --    begin

   -- --         Name :=
   --          Initialize(A_Queueing_System.Arrival_Rate);
   --          Initialize(A_Queueing_System.Service_Rate);
   --          A_Queueing_System.Nb_Of_Messages := 0.0;
   --          A_Queueing_System.Waiting_Time  := 0.0;
   --          A_Queueing_System.Utilisation   := 0.0;
   --          A_Queueing_System.Max_State     := 0;

   --    end;

   -- compute the average response time
   --
   procedure Compute_Average (
         A_Queueing_System : in out Generic_Queueing_System'Class ) is 

      Result : Double;  
   begin
      Result := 0.0;
      for I in 1 .. A_Queueing_System.Service_Resp_Time.Nb_Entries loop
         -- Put_Line("xi =  " & A_Queueing_System.Service_Resp_time.entries(I).Data'img);
         Result := Result + A_Queueing_System.Service_Resp_Time.Entries(I).
            Data;
      end loop;

      Result := Result / Double(A_Queueing_System.Service_Resp_Time.
         Nb_Entries);
      A_Queueing_System.Avg_Service := Result;
   end Compute_Average;

   -- compute the variance
   --
   procedure Compute_Variance (
         A_Queueing_System : in out Generic_Queueing_System'Class ) is 

      Result : Double;  
   begin
      -- Put_Line("variance computation");
      Result := 0.0;
      for I in 1 .. A_Queueing_System.Service_Resp_Time.Nb_Entries loop
         -- Put_Line("x" & I'Img & " = " & A_Queueing_System.Service_Resp_time.entries(I).Data'img);
         Result := Result +
            (A_Queueing_System.Service_Resp_Time.Entries(I).Data -
            A_Queueing_System.Avg_Service)*
            (A_Queueing_System.Service_Resp_Time.Entries(I).Data -
            A_Queueing_System.Avg_Service);
      end loop;

      Result := Result / Double(A_Queueing_System.Service_Resp_Time.
         Nb_Entries - 1);
      A_Queueing_System.Var_Service := Result;
      -- Put_Line("Variance = " & A_Queueing_System.Var_Service'img);
   end Compute_Variance;



   procedure Reset (
         A_Queueing_System : in out Generic_Queueing_System ) is 

   begin
      -- A_Queueing_System.Name;
      Initialize(A_Queueing_System.Arrival_Rate);
      Initialize(A_Queueing_System.Service_Rate);
      A_Queueing_System.Nb_Of_Messages := 0.0;
      A_Queueing_System.Waiting_Time := 0.0;
      A_Queueing_System.Utilisation := 0.0;
      A_Queueing_System.Max_State :=0;

      A_Queueing_System.Last_Inter_Time := 0.0;

      if A_Queueing_System.Service_Resp_Time/=null then
         Free(A_Queueing_System.Service_Resp_Time);
      end if;
      A_Queueing_System.Service_Resp_Time := new Resp_Time_Table;

   end Reset;


   procedure Initialize (
         A_Queueing_System : in out Generic_Queueing_System ) is 
   begin
      Reset(A_Queueing_System);
   end Initialize;




   procedure Finalize (
         A_Queueing_System : in out Generic_Queueing_System ) is 
   begin
      null;
   end Finalize;




   procedure Adjust (
         A_Queueing_System : in out Generic_Queueing_System ) is 
   begin
      null;
   end Adjust;




end Queueing_System;
