------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Numerics.Float_Random;
use Ada.Numerics.Float_Random;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

with Indexed_Tables;
with Natural_Util;
with Double_Util;


with Ada.Finalization;



package Queueing_System is

   type Queueing_System_Type is 
         (Qs_Mm1,  
          Qs_Md1,  
          Qs_Mp1,  
          Qs_Mg1,  
          Qs_Mms,  
          Qs_Mds,  
          Qs_Mps,  
          Qs_Mgs,  
          Qs_Mm1n, 
          Qs_Md1n, 
          Qs_Mp1n, 
          Qs_Mg1n, 
          Qs_Mmsn, 
          Qs_Mdsn, 
          Qs_Mpsn, 
          Qs_Mgsn); 

   type Generic_Queueing_System is abstract tagged private; 
   type Queueing_System_Ptr is access all Generic_Queueing_System'Class; 


   procedure Initialize (
         A_Queueing_System : in out Generic_Queueing_System ); 


   procedure Finalize (
         A_Queueing_System : in out Generic_Queueing_System ); 


   procedure Adjust (
         A_Queueing_System : in out Generic_Queueing_System ); 


   Arrival_Rate_Table_Size : constant Natural := 100;  
   Service_Rate_Table_Size : constant Natural := 100;  
   Resp_Time_Table_Size    : constant Natural := 5000000;  

   package A_Arrival_Rate is new Indexed_Tables(Double,
      Natural,
      Arrival_Rate_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   use A_Arrival_Rate;

   package A_Service_Rate is new Indexed_Tables(Double,
      Natural,
      Service_Rate_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   use A_Service_Rate;

   package A_Resp_Time is new Indexed_Tables(Double,
      Natural,
      Resp_Time_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   use A_Resp_Time;

   subtype Arrival_Rate_Table is A_Arrival_Rate.Indexed_Table;
   subtype Arrival_Rate_Range is A_Arrival_Rate.Indexed_Table_Range;

   subtype Service_Rate_Table is A_Service_Rate.Indexed_Table;
   subtype Service_Rate_Range is A_Service_Rate.Indexed_Table_Range;

   subtype Resp_Time_Table is A_Resp_Time.Indexed_Table;
   subtype Resp_Time_Table_Ptr is A_Resp_Time.Indexed_Table_Ptr;
   subtype Resp_Time_Range is A_Resp_Time.Indexed_Table_Range;


   -- return the next customer arrival instant
   --
   procedure Next_Customer_Arrival (
         A_Queueing_System : in out Generic_Queueing_System; 
         New_Arrival       : in out Double;                  
         Arrival_Rate      : in     Double;                  
         Seed              : in     Generator;               
         Old_Arrival       : in     Double                   ); 

   -- return the next customer departure instant
   --
   procedure Next_Customer_Departure (
         A_Queueing_System    : in out Generic_Queueing_System; 
         New_Arrival          : in out Double;                  
         Departure_Rate       : in     Double;                  
         Seed                 : in     Generator;               
         Old_Departure        : in     Double;                  
         Buffer_Size          : in     Natural;                 
         Last_Message_Arrival : in     Double                   ); 

   -- return the theoric average waiting time of one customer in the queue
   --
   function Qs_Average_Waiting_Time (
         A_Queueing_System : in     Generic_Queueing_System; 
         Arrival_Rate      : in     Double;                  
         Departure_Rate    : in     Double                   ) 
     return Double; 

   -- return the theoric average number of customer in the queue
   --
   function Qs_Average_Number_Customer (
         A_Queueing_System : in     Generic_Queueing_System; 
         Arrival_Rate      : in     Double;                  
         Departure_Rate    : in     Double                   ) 
     return Double; 

   -- return the probability to be in a certain state
   --
   function Get_Probability_Of_State (
         A_Queueing_System : in     Generic_Queueing_System; 
         N                 : in     Natural                  ) 
     return Double; 
   -- return the probability of buffer will be full
   --
   function Get_Probability_Of_Full_Buffer (
         A_Queueing_System : in     Generic_Queueing_System ) 
     return Double; 

   -- return the utilistation of queue
   --
   function Get_Utilisation (
         A_Queueing_System : in     Generic_Queueing_System ) 
     return Double; 

   -- return the probability of all tasks are busy
   -- Eq. 2.153, Robertazzi: Queueing theory
   --
   function Get_Probability_Of_Task_Block (
         A_Queueing_System : in     Generic_Queueing_System ) 
     return Double; 

   --P[queueing], Eq. 2.148, Robertazzi
   --
   function Get_Probability_Of_Queueing (
         A_Queueing_System : in     Generic_Queueing_System ) 
     return Double; 

   --Get probability density of states from 0->N
   --
   function Get_Probability_Density (
         A_Queueing_System : in     Generic_Queueing_System'Class;     
         N                 : in     Natural                       := 0 ) 
     return Double; 


   procedure Simulate (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Arrival_Rate      : in     Arrival_Rate_Table;            
         --        Arrival_Rate_Number : in Natural;
         Server_Period        : in     Natural;     
         Nb_Inter_Time        : in     Natural;     
         Max_Buffer_Size      :    out Natural;     
         Avg_Buffer_Size      :    out Double;      
         Var_Buffer_Size      :    out Double;      
         Buffer_Max_Wait_Time :    out Double;      
         Buffer_Avg_Wait_Time :    out Double;      
         Dis_Buff_Size        : in     Boolean;     
         Seed_Value           :        Natural := 0 ); 

   procedure Theoric_Results (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Arrival_Rate      : in     Arrival_Rate_Table;            
         Server_Period     : in     Natural;                       
         Rau               : in     Double                         ); 

   function Exponential_Inter_Time (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Rate              : in     Double;                        
         Seed              : in     Generator;                     
         Old_Activation    : in     Double                         ) 
     return Double; 

   function Periodic_Inter_Time (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Rate              : in     Double;                        
         Old_Activation    : in     Double                         ) 
     return Double; 

   function Periodic_Random_Inter_Time (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Rate              : in     Double;                        
         Old_Activation    : in     Double;                        
         Seed              : in     Generator                      ) 
     return Double; 

   function Constant_Inter_Time (
         A_Queueing_System    : in     Generic_Queueing_System'Class; 
         Rate                 : in     Double;                        
         Old_Activation       : in     Double;                        
         Buffer_Size          : in     Natural;                       
         Last_Message_Arrival : in     Double                         ) 
     return Double; 

   function General_Inter_Time (
         A_Queueing_System    : in     Generic_Queueing_System'Class; 
         Rate                 : in     Double;                        
         Old_Activation       : in     Double;                        
         Buffer_Size          : in     Natural;                       
         Last_Message_Arrival : in     Double                         ) 
     return Double; 

   procedure Compute_Average (
         A_Queueing_System : in out Generic_Queueing_System'Class ); 

   procedure Compute_Variance (
         A_Queueing_System : in out Generic_Queueing_System'Class ); 


   -- generation of departure rate and arrival rate for a queuing system
   --
   procedure Random_Qs_Project (
         Utilization  : in     Double;             
         Arrival_Rate : in out Arrival_Rate_Table; 
         Service_Rate : in out Service_Rate_Table; 
         --         Departure_Rate        : out Double;
         Seed                  : in     Generator;        
         Max_Interarrival_Time : in     Natural   := 4000 ); 

   procedure Set_Qs_Arrival_Rate (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Value             : in     Double                         ); 

   function Get_Qs_Arrival_Rate (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Place             : in     Arrival_Rate_Range             ) 
     return Double; 

   procedure Set_Qs_Service_Rate (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Value             : in     Double                         ); 

   function Get_Qs_Service_Rate (
         A_Queueing_System : in     Generic_Queueing_System'Class; 
         Place             : in     Service_Rate_Range             ) 
     return Double; 

   procedure Set_Qs_Nb_Of_Messages (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Value             :        Double                         ); 

   function Get_Qs_Nb_Of_Messages (
         A_Queueing_System : in     Generic_Queueing_System'Class ) 
     return Double; 

   procedure Set_Qs_Waiting_Time (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Value             :        Double                         ); 

   function Get_Qs_Waiting_Time (
         A_Queueing_System : in     Generic_Queueing_System'Class ) 
     return Double; 

   procedure Set_Qs_Max_State (
         A_Queueing_System : in out Generic_Queueing_System'Class; 
         Value             :        Natural                        ); 

   function Get_Qs_Max_State (
         A_Queueing_System : in     Generic_Queueing_System'Class ) 
     return Natural; 

   procedure Reset (
         A_Queueing_System : in out Generic_Queueing_System ); 

   --   procedure Set_Qs_Utilisation (
   --            A_Queueing_System : in Generic_Queueing_System'Class;
   --            Value : double);

private


   type Generic_Queueing_System is abstract new Ada.Finalization.Controlled with 
      record 
         Name              : Queueing_System_Type;  
         Arrival_Rate      : Arrival_Rate_Table;   -- average values 
         Service_Rate      : Service_Rate_Table;   -- average values 
         Nb_Of_Messages    : Double;               -- average value 
         Waiting_Time      : Double;               
         -- average value in system 
         Utilisation       : Double;  
         Max_State         : Natural;              
         -- maximum correct state 
         Service_Resp_Time : Resp_Time_Table_Ptr;  
         Var_Service       : Double;               
         -- variance of the departure law 
         Avg_Service       : Double;               
         -- average of the departure law 
         Last_Inter_Time   : Double;               
         -- temporary variable to compute next departure for mg1 qs 
      end record; 


end Queueing_System;




