------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Text_Io;
use Text_Io;

with Config;
use Config;

with Translate;
use Translate;

with Tasks;
use Tasks;

with Scheduler;
use Scheduler;
use Scheduler.Double_Tasks_Parameters;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;


with Unbounded_Strings;
use Unbounded_Strings;
with Ada.Numerics.Float_Random;
use Ada.Numerics.Float_Random;





package body Scheduler.Dynamic_Priority is




   function Build_Tcb (
         My_Scheduler : in     Dynamic_Priority_Scheduler; 
         A_Task       :        Task_Ptr                    ) 
     return Tcb_Ptr is 

      A_Tcb : Tcb_Ptr;  
      Seed  : Generator;  
   begin
      A_Tcb:=new Dynamic_Priority_Tcb;

      A_Tcb.Tsk:=A_Task;
      A_Tcb.Activation:=1;
      A_Tcb.End_Time:=0;
      A_Tcb.Wake_Up_Time:=A_Task.Start_Time;
      A_Tcb.Rest_Of_Capacity:=A_Task.Capacity;

      if A_Task.Task_Type = Periodic_Type
            then
         if(A_Task.Deadline /= 0)
               then
            Dynamic_Priority_Tcb_Ptr(A_Tcb).Dynamic_Deadline:=
               A_Task.Deadline+A_Tcb.Wake_Up_Time;
         else
            Dynamic_Priority_Tcb_Ptr(A_Tcb).Dynamic_Deadline:=
               Periodic_Task_Ptr(
               A_Task).Period+A_Tcb.Wake_Up_Time;
         end if;
      else
         Dynamic_Priority_Tcb_Ptr(A_Tcb).Dynamic_Deadline:=A_Task.
            Deadline;
      end if;

      if A_Task.Task_Type = Poisson_Type
            then
         --         Put_Line("Seed: " & Poisson_Task_Ptr(A_Task).Seed'Img);
         Reset(Seed,Poisson_Task_Ptr(A_Task).Seed);
         Save(Seed,A_Tcb.Seed_State);

         if(A_Task.Deadline /= 0)
               then
            Dynamic_Priority_Tcb_Ptr(A_Tcb).Dynamic_Deadline:=
               A_Task.Deadline+A_Tcb.Wake_Up_Time;
         else
            Dynamic_Priority_Tcb_Ptr(A_Tcb).Dynamic_Deadline:=
               Poisson_Task_Ptr(A_Task).Period+A_Tcb.Wake_Up_Time;
         end if;
      else
         Dynamic_Priority_Tcb_Ptr(A_Tcb).Dynamic_Deadline:=A_Task.
            Deadline;
      end if;


      return A_Tcb;
   end Build_Tcb;


   procedure Check_Before_Scheduling (
         My_Scheduler   : in     Dynamic_Priority_Scheduler; 
         My_Tasks       : in     Tasks_Set;                  
         Processor_Name : in     Unbounded_String            ) is 

   begin
      null;
   end Check_Before_Scheduling;


   procedure Bound_On_Processor_Utilization (
         My_Scheduler   : in     Dynamic_Priority_Scheduler; 
         My_Tasks       : in     Tasks_Set;                  
         Processor_Name : in     Unbounded_String;           
         Result         : in out Double;                     
         Msg            : in out Unbounded_String            ) is 
   begin
      Result:=1.0;
      Msg:=To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 8, ") & Lb_Theorem(
         Current_Language) &
         To_Unbounded_String("2). ") & Unbounded_Lf;

   end Bound_On_Processor_Utilization;



   function Compute_Wi_A_T (
         My_Tasks : in     Tasks_Set; 
         Taski    : in     Task_Ptr;  
         A        : in     Integer;   
         T        : in     Double     ) 
     return Double is 

      Result   : Double         := 0.0;  
      Min      : Double         := 0.0;  
      Iterator : Tasks_Iterator;  
      Taskj    : Task_Ptr;  
   begin

      Reset_Iterator(Iterator);
      loop
         Current_Element(My_Tasks, Taskj, Iterator);

         if ((Taskj.Name /= Taski.Name) and (Taskj.Deadline <= A + Taski.
                  Deadline) )
               then
            Min := 1.0 + Double'Floor(
               (Double(A) + Double(Taski.Deadline) - Double(Taskj.
                     Deadline))
               / Double(Periodic_Task_Ptr(Taskj).Period)
               ) ;

            if ( Double'Ceiling(T/Double(Periodic_Task_Ptr(Taskj).Period)) <
                  Min )
                  then
               Min := Double'Ceiling(T/Double(Periodic_Task_Ptr(Taskj).
                     Period));
            end if ;
            Result := Result + Min * Double(Taskj.Capacity) ;
         end if;

         -- exit loop when there is no more task
         exit when Is_Last_Element(My_Tasks, Iterator) = True;
         Next_Element(My_Tasks, Iterator);
      end loop ;

      return Result ;
   end Compute_Wi_A_T ;

   function Compute_Li_A (
         My_Scheduler : in     Dynamic_Priority_Scheduler; 
         My_Tasks     : in     Tasks_Set;                  
         Taski        : in     Task_Ptr;                   
         A            : in     Integer                     ) 
     return Double is 

      Li_A     : Double         := 0.1;  
      Li_A_1   : Double         := 0.0;  
      Iterator : Tasks_Iterator;  
      Taskj    : Task_Ptr;  
      Max_Cj   : Natural        := 0;  
      Tmp      : Double         := 0.0;  
      Min      : Double         := 0.0;  
      Min1     : Double         := 0.0;  
      Min2     : Double         := 0.0;  
   begin

      if ( My_Scheduler.Is_Preemptive = Preemptive )
            then
         -- preemptive case
         while Li_A_1 /= Li_A loop

            Li_A := Li_A_1 ;

            Li_A_1 := Compute_Wi_A_T(My_Tasks, Taski, A, Li_A) ;

            case Taski.Task_Type is
               when Periodic_Type =>
                  Li_A_1 := Li_A_1 + ( 1.0 + Double'Floor( Double(A) /
                        Double(Periodic_Task_Ptr(Taski).Period) ) )
                     * Double(Taski.Capacity) ;
               when Poisson_Type =>
                  Li_A_1 := Li_A_1 + ( 1.0 + Double'Floor( Double(A) /
                        Double(Poisson_Task_Ptr(Taski).Period) ) )
                     * Double(Taski.Capacity) ;
               when others =>
                  raise Constraint_Error;
            end case;
         end loop ;

      else
         -- non preemptive case

         while Li_A_1 /= Li_A loop

            Li_A := Li_A_1 ;
            Tmp := 0.0 ;
            Reset_Iterator(Iterator);
            loop
               Current_Element(My_Tasks, Taskj, Iterator);

               if ( (A + Taski.Deadline) < Taskj.Deadline )
                     then
                  if ( Taskj.Capacity > Max_Cj)
                        then
                     Max_Cj := Taskj.Capacity ;
                  end if ;
               end if ;

               if ((Taskj.Name /= Taski.Name) and (Taskj.Deadline <= A +
                        Taski.Deadline) )
                     then
                  Min1 := 1.0
                     + Double'Floor( Li_A / Double(Periodic_Task_Ptr(
                           Taskj).Period) ) ;

                  Min2 := 1.0
                     + Double'Floor(Double(A) + Double(Taski.Deadline) -
                     Double(Taskj.Deadline))
                     / Double(Periodic_Task_Ptr(Taskj).Period) ;

                  Min := Min1 ;
                  if(Min2 < Min) then
                     Min := Min2 ;
                  end if ;

                  Tmp := Tmp + Min * Double(Taskj.Capacity)
                     +  Double'Floor(Double(A) / Double(Periodic_Task_Ptr(
                           Taskj).Period)) * Double(Taski.Capacity) ;
               end if ;

               -- exit loop when there is no more task
               exit when Is_Last_Element(My_Tasks, Iterator) = True;
               Next_Element(My_Tasks, Iterator);
            end loop ;

            Li_A_1 := Double(Max_Cj) + Tmp ;
         end loop ;

      end if ;

      return Li_A ;
   end Compute_Li_A ;


   -- This function may present an infinite loop for non-preemptive case
   -- It may come from the upper bound for a: (exit when a >= Li(a))
   --
   procedure Compute_Response_Time (
         My_Scheduler   : in     Dynamic_Priority_Scheduler; 
         My_Tasks       : in out Tasks_Set;                  
         Processor_Name : in     Unbounded_String;           
         Msg            : in out Unbounded_String;           
         Response_Time  :    out Response_Table              ) is 

      Tmp       : Tasks_Set;  
      Iterator1 : Tasks_Iterator;  
      Taski     : Task_Ptr;  
      Iterator2 : Tasks_Iterator;  
      Taskj     : Task_Ptr;  
      I         : Response_Range := 0;  

      A : Integer := 0;  

      B_Sup : Double  := 0.0;  
      Li_A  : Double  := 0.0;  
      K     : Natural := 0;  

   begin
      Double_Tasks_Parameters.Initialize(Response_Time);

      -- check if tasks are periodics
      Periodic_Control(My_Tasks, Processor_Name);

      -- check start time
      Start_Time_Control(My_Tasks, Processor_Name);

      -- check offset
      Offset_Control(My_Tasks, Processor_Name);

      --
      Current_Processor_Name:=Processor_Name;
      Select_And_Copy( My_Tasks, Tmp, Select_Cpu'access );
      Reset_Iterator(Iterator1);


      loop
         -- selection of the current task
         Current_Element(Tmp, Taski, Iterator1);

         -- initialize response time for current task
         Response_Time.Entries(I).Data := 0.0;
         Response_Time.Entries(I).Item := Taski;
         Response_Time.Nb_Entries := Response_Time.Nb_Entries + 1;

         if ( My_Scheduler.Is_Preemptive = Preemptive )
               then
            Response_Time.Entries(I).Data := Double(Taski.Capacity) ;
         end if ;

         -- compute S "on the fly"
         Reset_Iterator(Iterator2);

         loop
            Current_Element(Tmp, Taskj, Iterator2);

            K := 0 ;
            loop
               case Taskj.Task_Type is
                  when Periodic_Type =>

                     A:= (K * Periodic_Task_Ptr(Taskj).Period)
                        + Taskj.Deadline
                        - Taski.Deadline ;
                  when Poisson_Type =>

                     A:= (K * Poisson_Task_Ptr(Taskj).Period)
                        + Taskj.Deadline
                        - Taski.Deadline ;
                  when others =>
                     raise Constraint_Error;
               end case;

               Li_A := Compute_Li_A(My_Scheduler, Tmp, Taski, A) ;

               if ( (Li_A - Double(A) ) > Response_Time.Entries(I).Data )
                     then
                  Response_Time.Entries(I).Data := Li_A - Double(A) ;

                  -- PROBLEM
                  -- with non preemptive case response_time has no upper bound in some cases
                  -- Put(Natural(Response_Time.Entries(i).Data)'Img) ;
               end if ;

               if ( My_Scheduler.Is_Preemptive = Preemptive )
                     then
                  B_Sup := Li_A ;
               else
                  B_Sup := Li_A ;
               end if ;

               exit when (Double(A) >= B_Sup) ;


	       -- PROBLEM : the lines below try yo fix the problem described 10 lines 
               -- before (see morin2003.doc) 
	       -- The "exit when K" statement is not a part of the EDF response time
               -- computation : it's just a way to be sure that the response
               -- time computation never loop for ever .... This Exit statement
               -- HAVE TO BE REMOVED WHEN EDF response time computation will be
	       -- strongly tested by users
               --
	       exit when K > 10000;

               K := K + 1 ;
            end loop ;

            exit when Is_Last_Element(Tmp, Iterator2);
            -- go to the next task
            Next_Element(Tmp, Iterator2);
         end loop ;

         exit when Is_Last_Element(Tmp, Iterator1) = True;

         -- go to the next task
         Next_Element(Tmp, Iterator1);

         I := I + 1 ;
      end loop ;
   end Compute_Response_Time;



   procedure Specific_Scheduler_Initialization (
         My_Scheduler   : in out Dynamic_Priority_Scheduler; 
         Si             : in out Scheduling_Information;     
         Processor_Name : in     Unbounded_String;           
         My_Tasks       : in out Tasks_Set;                  
         My_Resources   : in out Resources_Set               ) is 

   begin
      null;
   end Specific_Scheduler_Initialization;







end Scheduler.Dynamic_Priority ;


