------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------



with Text_Io;
use Text_Io;

with Config;
use Config;

with Translate;
use Translate;

with Tasks;
use Tasks;

with Resources;
use Resources;

with Scheduler;
use Scheduler;
use Scheduler.Double_Tasks_Parameters;


with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;
with Ada.Numerics.Float_Random;
use Ada.Numerics.Float_Random;


with Unbounded_Strings;
use  Unbounded_Strings;



package body Scheduler.Fixed_Priority  is


   function Build_Tcb (
         My_Scheduler : in     Fixed_Priority_Scheduler; 
         A_Task       :        Task_Ptr                  ) 
     return Tcb_Ptr is 

      A_Tcb : Fixed_Priority_Tcb_Ptr;  
      Seed  : Generator;  
   begin
      A_Tcb:=new Fixed_Priority_Tcb;

      A_Tcb.Tsk:=A_Task;
      A_Tcb.Activation:=1;
      A_Tcb.End_Time:=0;
      A_Tcb.Wake_Up_Time:=A_Task.Start_Time;
      A_Tcb.Rest_Of_Capacity:=A_Task.Capacity;

      if A_Tcb.Tsk.Task_Type = Poisson_Type then
         --         Put_Line("Seed: " & Poisson_Task_Ptr(A_Task).Seed'Img);
         Reset(Seed,Poisson_Task_Ptr(A_Task).Seed);
         Save(Seed,A_Tcb.Seed_State);
      end if;

      return Tcb_Ptr(A_Tcb);
   end Build_Tcb;



   procedure Do_Election (
         My_Scheduler      : in out Fixed_Priority_Scheduler; 
         Si                : in out Scheduling_Information;   
         Processor_Name    : in     Unbounded_String;         
         At_Time           :        Natural;                  
         My_Dependencies   : in     Tasks_Dependencies;       
         With_Offsets      :        Boolean;                  
         With_Precedencies :        Boolean;                  
         With_Resources    :        Boolean;                  
         Elected           : in out Tasks_Range;              
         No_Task           : in out Boolean                   ) is 


      Highiest_Priority : Natural     := Natural'First;  
      I                 : Tasks_Range := 0;  
   begin


      -- For each task, call "check_ressources" to
      -- take care of priority modification (PCP and PIP)
      --
      loop
         if (Si.Tcbs(I).Tsk.Location = Processor_Name) then
            if (Si.Tcbs(I).Wake_Up_Time <= At_Time)
                  and (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                  then
               if (With_Offsets=False) or
                     Check_Offset(Si.Tcbs(I), At_Time)
                     then
                  if (With_Precedencies=False) or
                        Check_Precedencies(Si, My_Dependencies, At_Time,
                        Si.Tcbs(I))
                        then
                     Check_Resource(My_Scheduler, Si, Si.Tcbs(I),
                        Fixed_Priority_Tcb_Ptr(Si.Tcbs(I)).Is_Ready);
                  end if;
               end if;
            end if;
         end if;


         I:=I+1;
         exit when Si.Tcbs(I)=null;
      end loop;


      -- We can compute scheduling since priority are ok
      --
      I:=0;
      loop

         if (Si.Tcbs(I).Tsk.Location = Processor_Name) then
            if (Si.Tcbs(I).Wake_Up_Time <= At_Time) and
                  (Natural(Fixed_Priority_Tcb_Ptr(Si.Tcbs(I)).
                     Current_Priority) > Highiest_Priority) and
                  (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                  then
               if (With_Offsets=False) or
                     Check_Offset(Si.Tcbs(I), At_Time)
                     then
                  if (With_Precedencies=False) or
                        Check_Precedencies(Si, My_Dependencies, At_Time,
                        Si.Tcbs(I))
                        then
                     if (With_Resources=False) or (Fixed_Priority_Tcb_Ptr(
                              Si.Tcbs(I)).Is_Ready=True)
                           then

                        Highiest_Priority:=Natural(Fixed_Priority_Tcb_Ptr(
                              Si.Tcbs(I)).Current_Priority);
                        Elected:=I;
                     end if;
                  end if;
               end if;
            end if;
         end if;


         I:=I+1;
         exit when Si.Tcbs(I)=null;
      end loop;

      if Highiest_Priority = Natural'First
            then
         No_Task:=True;
      else
         No_Task:=False;
      end if;

   end Do_Election;



   procedure Bound_On_Processor_Utilization (
         My_Scheduler   : in     Fixed_Priority_Scheduler; 
         My_Tasks       : in     Tasks_Set;                
         Processor_Name : in     Unbounded_String;         
         Result         : in out Double;                   
         Msg            : in out Unbounded_String          ) is 

      Nb : Double := Double (Get_Number_Of_Task_From_Processor (My_Tasks,
        Processor_Name));  
   begin
      Msg:=To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 16, ") & Lb_Theorem(
         Current_Language) &
         To_Unbounded_String("8). ") & Unbounded_Lf;


      Result:=Nb*( Pow(2.0,(1.0/Nb)) - 1.0);
   end Bound_On_Processor_Utilization;




   function Compute_Workload (
         My_Tasks : in     Tasks_Set; 
         Taski    : in     Task_Ptr;  
         Q        : in     Integer    ) 
     return Double is 

      Iterator2 : Tasks_Iterator;  
      Taskj     : Task_Ptr;  
      Wiq_K,  
      Wiq_K1    : Double;  

   begin

      Wiq_K  := -0.1 ;
      Wiq_K1 :=  0.0 ;

      -- approximations for Wiq
      -- exit loop when approximations converge
      while Wiq_K /= Wiq_K1  loop

         Reset_Iterator(Iterator2);

         Wiq_K  := Wiq_K1 ;
         Wiq_K1 := Double((Q+1)*Taski.Capacity) + Double(Taski.
            Blocking_Time) ;

         -- iterate on all tasks
         loop
            Current_Element(My_Tasks, Taskj, Iterator2);

            -- if Taskj is in hp(Taski)
            if(Taskj.Priority> Taski.Priority)
                  then
               Wiq_K1 :=   Wiq_K1
                  + Double'Ceiling((Wiq_K + Double(Periodic_Task_Ptr(
                           Taskj).Jitter))
                  / Double(Periodic_Task_Ptr(Taskj).Period))
                  * Double(Taskj.Capacity) ;
            end if ;

            -- exit loop when there is no more task
            exit when Is_Last_Element(My_Tasks, Iterator2) = True;

            Next_Element(My_Tasks, Iterator2);

         end loop ;

      end loop ;

      return Wiq_K1;

   end Compute_Workload;


   procedure Compute_Response_Time (
         My_Scheduler   : in     Fixed_Priority_Scheduler; 
         My_Tasks       : in out Tasks_Set;                
         Processor_Name : in     Unbounded_String;         
         Msg            : in out Unbounded_String;         
         Response_Time  :    out Response_Table            ) is 

      Tmp       : Tasks_Set;  
      Iterator1 : Tasks_Iterator;  
      Taski     : Task_Ptr;  
      I         : Response_Range := 0;  
      Wiq,  
      Ri        : Double         := 0.0;  
      Q         : Integer;  

   begin

      Double_Tasks_Parameters.Initialize(Response_Time);

      Current_Processor_Name:=Processor_Name;
      Select_And_Copy( My_Tasks, Tmp, Select_Cpu'access );


      -- Set priority according to the scheduler
      if (My_Scheduler.Name = Deadline_Monotonic)
            then
         Set_Priority_According_To_Dm(My_Scheduler, Tmp);
      else
         if (My_Scheduler.Name = Rate_Monotonic)
               then
            Set_Priority_According_To_Rm(My_Scheduler, Tmp);
         end if;
      end if;


      Sort(Tmp, Increasing_Priority'access);


      -- bibliographical references 
      Msg:=To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[2], page 3, ") & Lb_Equation(
         Current_Language);

      if ( My_Scheduler.Is_Preemptive = Not_Preemptive )
            then
         Msg:= Msg & To_Unbounded_String("4 ") & Lb_And(Current_Language) &
            To_Unbounded_String("[2], page 29, ") & Lb_Theorem(
            Current_Language)
            & To_Unbounded_String("15). ");
      else
         Msg:=Msg & To_Unbounded_String("4). ");
      end if ;



      Reset_Iterator(Iterator1);

      loop

         -- selection of the current task
         Current_Element(Tmp, Taski, Iterator1);

         -- initialize response time for current task
         Response_Time.Entries(I).Data := 0.0;
         Response_Time.Entries(I).Item := Taski;
         Response_Time.Nb_Entries := Response_Time.Nb_Entries+1;

         -- find the maximum for Ri in [0..Q]
         Q := 0 ;
         loop

            Wiq := Compute_Workload(Tmp, Taski, Q);

            -- Ri = Wiq + Ji - qTi
            case Taski.Task_Type is
               when Periodic_Type =>
                  Ri :=   Wiq
                     + Double(Periodic_Task_Ptr(Taski).Jitter)
                     - Double(Q) * Double(Periodic_Task_Ptr(Taski).Period) ;
               when Poisson_Type =>
                  Ri :=   Wiq
                     + Double(Poisson_Task_Ptr(Taski).Jitter)
                     - Double(Q) * Double(Poisson_Task_Ptr(Taski).Period) ;
               when others =>
                  raise Constraint_Error;
            end case;


            -- is Ri greater than current maximum for Ri ?
            if ( Ri > Response_Time.Entries(I).Data )
                  then
               Response_Time.Entries(I).Data := Ri ;
            end if ;

            -- exit loop when Q is found ( WiQ<=(Q+1)Ti )
            case Taski.Task_Type is
               when Periodic_Type =>
                  exit when Wiq <= Double(Q+1) *
                     Double(Periodic_Task_Ptr(Taski).Period);
               when Poisson_Type =>
                  exit when Wiq <= Double(Q+1) *
                     Double(Poisson_Task_Ptr(Taski).Period);
               when others =>
                  raise Constraint_Error;
            end case;

            Q := Q + 1 ;

         end loop ;

         -- if the scheduler is non premptive add Ci to Ri
         if ( My_Scheduler.Is_Preemptive = Not_Preemptive )
               then
            Response_Time.Entries(I).Data := Response_Time.Entries(I).
               Data + Double(Periodic_Task_Ptr(Taski).Capacity) ;
         end if ;

         exit when Is_Last_Element(Tmp, Iterator1) = True;

         -- go to the next task
         Next_Element(Tmp, Iterator1);
         I:=I+1 ;

      end loop ;

   end Compute_Response_Time;



   procedure Set_Priority_According_To_Dm (
         My_Scheduler   : in     Fixed_Priority_Scheduler;                
         My_Tasks       : in out Tasks_Set;                               
         Processor_Name : in     Unbounded_String         := Empty_String ) is 

      Iterator1    : Tasks_Iterator;  
      Task1        : Task_Ptr;  
      Current_Prio : Priority_Range := 1;  

   begin


      Periodic_Control( My_Tasks, Processor_Name);

      Sort(My_Tasks, Decreasing_Deadline'access);

      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, Task1, Iterator1);

         if (Task1.Location = Processor_Name)
               or (Processor_Name = Empty_String)
               then
            Task1.Priority:=Current_Prio;
            Current_Prio:=Current_Prio+1;
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;
   end Set_Priority_According_To_Dm;


   procedure Set_Priority_According_To_Rm (
         My_Scheduler   : in     Fixed_Priority_Scheduler;                
         My_Tasks       : in out Tasks_Set;                               
         Processor_Name : in     Unbounded_String         := Empty_String ) is 

      Iterator1    : Tasks_Iterator;  
      Task1        : Task_Ptr;  
      Current_Prio : Priority_Range := 1;  

   begin


      Periodic_Control( My_Tasks, Processor_Name);


      Sort(My_Tasks, Decreasing_Period'access);

      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, Task1, Iterator1);

         if (Task1.Location = Processor_Name)
               or (Processor_Name = Empty_String)
               then

            Task1.Priority:=Current_Prio;
            Current_Prio:=Current_Prio+1;
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;
   end Set_Priority_According_To_Rm;



   procedure Compute_Ceiling_Of_Resources (
         My_Scheduler   : in out Fixed_Priority_Scheduler; 
         Si             : in out Scheduling_Information;   
         Processor_Name : in     Unbounded_String;         
         My_Tasks       : in out Tasks_Set;                
         My_Resources   : in out Resources_Set             ) is 


   begin
      -- Private scheduler data
      --
      My_Scheduler.Used_Resource := False;


      -- Set priority ceiling of resources (only for PCP resources)
      --
      for K in 0 .. Si.Number_Of_Resources-1 loop

         if Si.Shared_Resources(K).Shared.Protocol =
               Priority_Ceiling_Protocol then
            if Si.Shared_Resources(K).Shared.Location = Processor_Name then

               Fixed_Priority_Resource_Ptr(Si.Shared_Resources(K)).
                  Priority_Ceiling:=Priority_Range'First;

               for I in 0..Si.Shared_Resources(K).Shared.Task_Tab.
                     Nb_Entries-1
                     loop
                  --    Looking for the task
                  --
                  for J in 0..Si.Number_Of_Tasks-1 loop

                     if Si.Tcbs(J).Tsk.Location = Processor_Name then

                        if Si.Shared_Resources(K).Shared.Task_Tab.Entries(
                              I).
                              Item = Si.Tcbs(J).
                              Tsk.Name then
                           Fixed_Priority_Resource_Ptr(Si.
                              Shared_Resources(K)).
                              Priority_Ceiling:=Priority_Range'Max(
                              Fixed_Priority_Tcb_Ptr(Si.Tcbs(J)).
                              Current_Priority,
                              Fixed_Priority_Resource_Ptr(Si.
                                 Shared_Resources(K)).Priority_Ceiling);
                        end if;
                     end if;

                  end loop;

               end loop;
            end if;
         end if;

      end loop;


   end Compute_Ceiling_Of_Resources;


   procedure Specific_Scheduler_Initialization (
         My_Scheduler   : in out Fixed_Priority_Scheduler; 
         Si             : in out Scheduling_Information;   
         Processor_Name : in     Unbounded_String;         
         My_Tasks       : in out Tasks_Set;                
         My_Resources   : in out Resources_Set             ) is 


   begin



      -- Set priority according to the scheduler
      --
      if (My_Scheduler.Name = Deadline_Monotonic)
            then
         Set_Priority_According_To_Dm(My_Scheduler, My_Tasks);
      else
         if (My_Scheduler.Name = Rate_Monotonic)
               then
            Set_Priority_According_To_Rm(My_Scheduler, My_Tasks);
         end if;
      end if;

      for I in 0 .. Si.Number_Of_Tasks-1 loop
         if Si.Tcbs(I).Tsk.Location = Processor_Name then
            Fixed_Priority_Tcb_Ptr(Si.Tcbs(I)).Current_Priority:=
               Si.Tcbs(I).Tsk.Priority;
         end if;
      end loop;


      -- Set priority ceiling of resources (only to PCP resource
      --
      Compute_Ceiling_Of_Resources (
         My_Scheduler  ,
         Si            ,
         Processor_Name ,
         My_Tasks      ,
         My_Resources              );

   end Specific_Scheduler_Initialization;


   procedure Release_Resource (
         My_Scheduler      : in out Fixed_Priority_Scheduler; 
         Si                : in out Scheduling_Information;   
         Result            : in out Scheduling_Sequence;      
         Current_Time_Unit : in out Time_Unit_Range;          
         Current_Time      : in     Natural;                  
         A_Tcb             :        Tcb_Ptr                   ) is 

      Current_Capacity : Natural := 0;  

   begin

      if (Si.Number_Of_Resources>0)
            then
         Current_Capacity:=A_Tcb.Tsk.Capacity-A_Tcb.Rest_Of_Capacity+1;

         for Index1 in 0.. Si.Number_Of_Resources-1 loop


            -- For each resource, check if the task hold resources
            -- and then release them
            --
            for Index2 in 0..Si.Shared_Resources(Index1).Shared.
                  Task_Tab.Nb_Entries-1
                  loop


               if (Current_Capacity=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.Task_End)
                     and (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.Name)
                     then
                  Si.Shared_Resources(Index1).Shared.State:=
                     Si.Shared_Resources(Index1).Shared.State+1;
                  for I in 0 .. Si.Shared_Resources(Index1).
                        Nb_Allocated-1 loop
                     if Si.Shared_Resources(Index1).
                           Allocated_By(I)=A_Tcb.Tsk.Name
                           then
                        Si.Shared_Resources(Index1).
                           Allocated_By(I):= Si.
                           Shared_Resources(Index1).Allocated_By(
                           Si.Shared_Resources(Index1).
                           Nb_Allocated-1);
                        Si.Shared_Resources(Index1).
                           Nb_Allocated:=Si.Shared_Resources(
                           Index1).Nb_Allocated-1;
                        exit;
                     end if;
                  end loop;

                  -- If PIP or PCP is used, releasing a shared resource
                  -- should restore initial priority level of the
                  -- task
                  --

                  if (Si.Shared_Resources(Index1).Shared.
                        Protocol = Priority_Inheritance_Protocol) or
                        (Si.Shared_Resources(Index1).Shared.
                        Protocol = Priority_Ceiling_Protocol)
                        then

                     Change_Current_Priority( My_Scheduler,
                        Fixed_Priority_Tcb_Ptr(A_Tcb),
                        A_Tcb.Tsk.Priority);

                  end if;
               end if;


            end loop;


         end loop;
      end if;

   end Release_Resource;





   procedure Allocate_Resource (
         My_Scheduler      : in out Fixed_Priority_Scheduler; 
         Si                : in out Scheduling_Information;   
         Result            : in out Scheduling_Sequence;      
         Current_Time_Unit : in out Time_Unit_Range;          
         Current_Time      : in     Natural;                  
         A_Tcb             :        Tcb_Ptr                   ) is 


      Current_Capacity : Natural := 0;  


   begin

      if (Si.Number_Of_Resources>0)
            then

         Current_Capacity:=A_Tcb.Tsk.Capacity-A_Tcb.Rest_Of_Capacity+1;

         for Index1 in 0 .. Si.Number_Of_Resources -1 loop


            -- For each resource, check if the task request the resource
            -- and if the resource is free
            -- otherwise, block the task
            --
            for Index2 in 0..Si.Shared_Resources(Index1).Shared.
                  Task_Tab.Nb_Entries-1
                  loop


               if (Current_Capacity=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.Task_Begin)
                     and (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.Name)
                     then



                  -- PCP protocol : update current ceiling priority
                  -- the priority of the required resource
                  --

                  if  (Si.Shared_Resources(Index1).Shared.
                        Protocol = Priority_Ceiling_Protocol)
                        then
                     My_Scheduler.Current_Ceiling_Priority:=
                        Priority_Range'Max(My_Scheduler.
                        Current_Ceiling_Priority,
                        Fixed_Priority_Resource_Ptr(Si.Shared_Resources(
                              Index1)).Priority_Ceiling);

                  end if;


                  Si.Shared_Resources(Index1).Shared.State:=
                     Si.Shared_Resources(Index1).Shared.State-1;

                  Si.Shared_Resources(Index1).Allocated_By(
                     Si.Shared_Resources(Index1).Nb_Allocated):=
                     A_Tcb.Tsk.Name;
                  Si.Shared_Resources(Index1).Nb_Allocated:=
                     Si.Shared_Resources(Index1).Nb_Allocated+1;
               end if;

            end loop;


         end loop;
      end if;


   end Allocate_Resource;







   procedure Check_Resource (
         My_Scheduler : in out Fixed_Priority_Scheduler; 
         Si           : in out Scheduling_Information;   
         A_Tcb        : in     Tcb_Ptr;                  
         Is_Ready     :    out Boolean                   ) is 


      Current_Time : Natural := 0;  

   begin

      Is_Ready:=True;

      if (Si.Number_Of_Resources>0)
            then

         Current_Time:=A_Tcb.Tsk.Capacity-A_Tcb.Rest_Of_Capacity+1;

         for Index1 in 0 .. Si.Number_Of_Resources -1 loop


            -- For each resource, check if the task request the resource
            -- and if the resource is free
            --
            for Index2 in 0..Si.Shared_Resources(Index1).Shared.
                  Task_Tab.Nb_Entries-1
                  loop


               if (Current_Time=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.Task_Begin)
                     and (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.Name)
                     then


                  -- PCP blocking : current ceiling priority should be less or equal than
                  -- the priority of the required resource
                  --
                  if (Si.Shared_Resources(Index1).Shared.
                        Protocol = Priority_Inheritance_Protocol) or
                        (Si.Shared_Resources(Index1).Shared.
                        Protocol = Priority_Ceiling_Protocol)
                        then
                     if Fixed_Priority_Resource_Ptr(Si.Shared_Resources(
                              Index1)).Priority_Ceiling
                           > My_Scheduler.Current_Ceiling_Priority
                           then
                        Is_Ready:=False;
                     end if;
                  end if;


                  if (Si.Shared_Resources(Index1).Shared.State<=
                        0)
                        then
                     Is_Ready:=False;

                     -- Find if we must change a priority level of a task
                     -- that takes the resource (PIP protocol)
                     --
                     if (Si.Shared_Resources(Index1).Shared.
                           Protocol = Priority_Inheritance_Protocol) or
                           (Si.Shared_Resources(Index1).Shared.
                           Protocol = Priority_Ceiling_Protocol)
                           then
                        for I in 0 .. Si.Shared_Resources(
                              Index1).Nb_Allocated-1 loop
                           for J in 0..Si.Number_Of_Tasks-1  loop

                              if Si.Shared_Resources(Index1).
                                    Allocated_By(I)=Si.Tcbs(J).
                                    Tsk.Name
                                    then


                                 Change_Current_Priority( My_Scheduler,
                                    Fixed_Priority_Tcb_Ptr(Si.Tcbs(
                                          J)),
                                    Priority_Range'Max(
                                       Fixed_Priority_Tcb_Ptr(A_Tcb).
                                       Current_Priority,
                                       Fixed_Priority_Tcb_Ptr(Si.
                                          Tcbs(J)).Current_Priority) );

                              end if;
                           end loop;
                        end loop;
                     end if;
                  end if;
               end if;


            end loop;


         end loop;
      end if;


   end Check_Resource;





   function Build_Resource (
         My_Scheduler : in     Fixed_Priority_Scheduler; 
         A_Resource   :        Resource_Ptr              ) 
     return Shared_Resource_Ptr is 

      New_A_Resource : Fixed_Priority_Resource_Ptr;  

   begin

      New_A_Resource:= new Fixed_Priority_Resource;
      New_A_Resource.Shared:=A_Resource;

      -- Set priority ceiling of the resource
      --
      New_A_Resource.Priority_Ceiling:=Low_Priority;

      return Shared_Resource_Ptr(New_A_Resource);

   end Build_Resource;




   procedure Change_Current_Priority (
         My_Scheduler : in out Fixed_Priority_Scheduler'Class; 
         A_Tcb        : in     Fixed_Priority_Tcb_Ptr;         
         New_Priority :        Priority_Range                  ) is 
   begin
      Dispatched_Change_Current_Priority(My_Scheduler, A_Tcb, New_Priority);
   end Change_Current_Priority;



   procedure Dispatched_Change_Current_Priority (
         My_Scheduler : in out Fixed_Priority_Scheduler; 
         A_Tcb        : in     Fixed_Priority_Tcb_Ptr;   
         New_Priority :        Priority_Range            ) is 


   begin
      A_Tcb.Current_Priority:=New_Priority;

   end Dispatched_Change_Current_Priority;





end Scheduler.Fixed_Priority;


