------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------


with Io_Tools;
use Io_Tools;

with Buffers;
use Buffers;
use Buffers.Buffer_Roles;


with Messages;
use Messages;

with Fixed_Strings;
use Fixed_Strings;

with Ada.Sequential_Io;

with Stacks;

with Config;
use Config;

with Tasks;
use Tasks;


package body Task_Dependencies is


   function Search_Position (
         My_Dependencies : in     Tasks_Dependencies; 
         Name            : in     Unbounded_String    ) 
     return Dependencies_Range is 


   begin
      for I in  0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (Name = My_Dependencies.Task_Position(I).Name)
               then
            return I;
         end if;
      end loop;
      raise Task_Not_Found;
   end Search_Position;


   function Get_Root_Tasks (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Tasks_Set is 

      Result   : Tasks_Set;  
      Ok       : Boolean;  
      New_Task : Task_Ptr;  

   begin

      for J in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop

         Ok:=True;
         for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop

            if (not Is_Empty(My_Dependencies.Dependencies(I, J)))
                  then
               Ok:=False;
            end if;
         end loop;

         if Ok then
            New_Task:=Copy(My_Dependencies.Task_Position(J));
            Add(Result, New_Task);
         end if;

      end loop;

      return Result;

   end Get_Root_Tasks;


   -- Get a task whitout predecessors
   --
   function Get_A_Root_Task (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Task_Ptr is 

      Ok       : Boolean;  
      New_Task : Task_Ptr;  

   begin

      for J in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop

         Ok:=True;
         for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop

            if (not Is_Empty(My_Dependencies.Dependencies(I, J)))
                  then
               Ok:=False;
            end if;
         end loop;

         if Ok then
            New_Task:=Copy(My_Dependencies.Task_Position(J));
            return New_Task;
         end if;

      end loop;

      raise No_Dependencies;

   end Get_A_Root_Task;


   -- Get a task whitout successors
   -- 
   function Get_A_Leaf_Task (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Task_Ptr is 

      Ok       : Boolean;  
      New_Task : Task_Ptr;  

   begin

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop

         Ok:=True;
         for J in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop


            if (not Is_Empty(My_Dependencies.Dependencies(I, J)))
                  then
               Ok:=False;
            end if;
         end loop;

         if Ok then
            New_Task:=Copy(My_Dependencies.Task_Position(I));
            return New_Task;
         end if;

      end loop;

      raise No_Dependencies;

   end Get_A_Leaf_Task;




   function Get_Leaf_Tasks (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Tasks_Set is 

      Result   : Tasks_Set;  
      Ok       : Boolean;  
      New_Task : Task_Ptr;  

   begin

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop

         Ok:=True;
         for J in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop


            if (not Is_Empty(My_Dependencies.Dependencies(I, J)))
                  then
               Ok:=False;
            end if;
         end loop;

         if Ok then
            New_Task:=Copy(My_Dependencies.Task_Position(I));
            Add(Result, New_Task);
         end if;

      end loop;

      return Result;
   end Get_Leaf_Tasks;



   function Has_Predecessor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Boolean is 

      A_Task : Dependencies_Range;  
      Result : Boolean            := False;  
   begin

      A_Task:=Search_Position(My_Dependencies, My_Task);

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (not Is_Empty(My_Dependencies.Dependencies(I, A_Task)))
               then
            Result:=True;
         end if;
      end loop;
      return Result;
   exception
      when Task_Not_Found =>
         begin
            return False;
         end;
   end Has_Predecessor;


   function Has_Successor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Boolean is 

      A_Task : Dependencies_Range;  
      Result : Boolean            := False;  
   begin

      A_Task:=Search_Position(My_Dependencies, My_Task);

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (not Is_Empty(My_Dependencies.Dependencies(A_Task, I)))
               then
            Result:=True;
         end if;
      end loop;
      return Result;
   exception
      when Task_Not_Found =>
         begin
            return False;
         end;
   end Has_Successor;



   function Get_A_Successor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Task_Ptr is 

      A_Task : Dependencies_Range;  

   begin

      A_Task:=Search_Position(My_Dependencies, My_Task.Name);

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (not Is_Empty(My_Dependencies.Dependencies(A_Task, I)))
               then
            return My_Dependencies.Task_Position(I);
         end if;
      end loop;

      raise No_Dependencies;

   end Get_A_Successor;


   function Get_A_Predecessor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Task_Ptr is 

      A_Task : Dependencies_Range;  

   begin

      A_Task:=Search_Position(My_Dependencies, My_Task.Name);

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (not Is_Empty(My_Dependencies.Dependencies(I, A_Task)))
               then
            return My_Dependencies.Task_Position(I);
         end if;
      end loop;

      raise No_Dependencies;

   end Get_A_Predecessor;



   function Get_Successors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Tasks_Set is 

      Result : Tasks_Set;  
      A_Task : Dependencies_Range;  

   begin

      A_Task:=Search_Position(My_Dependencies, My_Task);

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (not Is_Empty(My_Dependencies.Dependencies(A_Task, I)))
               then
            Add(Result, My_Dependencies.Task_Position(I));
         end if;
      end loop;
      return Result;

   end Get_Successors_List;


   function Get_Predecessors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Tasks_Set is 

      Result : Tasks_Set;  
      A_Task : Dependencies_Range;  

   begin

      A_Task:=Search_Position(My_Dependencies, My_Task);

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (not Is_Empty(My_Dependencies.Dependencies(I, A_Task)))
               then
            Add(Result, My_Dependencies.Task_Position(I));
         end if;
      end loop;
      return Result;

   end Get_Predecessors_List;


   function Get_Successors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Tasks_Set is 

   begin
      return Get_Successors_List(My_Dependencies, My_Task.Name);
   end Get_Successors_List;


   function Get_Predecessors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Tasks_Set is 

   begin
      return Get_Predecessors_List(My_Dependencies, My_Task.Name);
   end Get_Predecessors_List;


   function Has_Predecessor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Boolean is 

   begin
      return Has_Predecessor(My_Dependencies, My_Task.Name);
   end Has_Predecessor;


   function Has_Successor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Boolean is 

   begin
      return Has_Successor(My_Dependencies, My_Task.Name);
   end Has_Successor;


   function No_Dependency (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Boolean is 

      Has_Dependency : Boolean := True;  

   begin
      for I in  Dependencies_Range loop
         for J in  Dependencies_Range loop

            if not Is_Empty(My_Dependencies.Dependencies(I, J))
                  then
               Has_Dependency:=False;
            end if;

         end loop;
      end loop;

      return Has_Dependency;

   end No_Dependency;


   function Get_Dependencies (
         My_Dependencies : in     Tasks_Dependencies; 
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr            ) 
     return Depends_Set is 
   begin
      return Get_Dependencies(My_Dependencies, From_Task.Name, To_Task.
         Name);
   end Get_Dependencies;



   function Get_Dependencies (
         My_Dependencies : in     Tasks_Dependencies; 
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String    ) 
     return Depends_Set is 

      A_Task1,  
      A_Task2 : Dependencies_Range;  

   begin

      A_Task1:=Search_Position(My_Dependencies, From_Task);
      A_Task2:=Search_Position(My_Dependencies, To_Task);

      if (not Is_Empty(My_Dependencies.Dependencies(A_Task1, A_Task2)))
            then
         return My_Dependencies.Dependencies(A_Task1, A_Task2);
      else
         raise No_Dependencies;
      end if;
   end Get_Dependencies;



   function Get_Precedence_Dependency (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Tasks_Set is 
      My_Task_Set : Tasks_Set;  
      My_Iterator : Depends_Iterator;  
      A_Dep       : Dependency_Ptr;  
   begin
      if not No_Dependency(My_Dependencies) then
         for I in Dependencies_Range loop
            for J in Dependencies_Range loop
               if (not Is_Empty(My_Dependencies.Dependencies(I, J))) then
                  Reset_Iterator(My_Iterator);
                  loop
                     Current_Element(My_Dependencies.Dependencies(I, J),
                        A_Dep,My_Iterator);
                     if (A_Dep.Dependence_Type = Precedence_Dependency) then
                        Add(My_Task_Set,My_Dependencies.Task_Position(I));
                        Add(My_Task_Set,My_Dependencies.Task_Position(J));
                     end if;
                     exit when Is_Last_Element(My_Dependencies.
                        Dependencies(I, J),
                        My_Iterator) = True;
                     Next_Element(My_Dependencies.Dependencies(I, J),
                        My_Iterator);
                  end loop;

               end if;

            end loop;
         end loop;
      end if;
      return My_Task_Set;
   end Get_Precedence_Dependency;

   -------------------------------------------------------------

   procedure Private_Add_Half_Dependency (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         A_Task          : in     Unbounded_String;   
         A_Dep           : in     Dependency_Ptr;     
         A_Type          : in     Half_Dep_Type       ) is 

      A_Half_Dep : Half_Dep_Ptr;  
   begin

      A_Half_Dep := new Half_Dep;

      A_Half_Dep.The_Dep := A_Dep;
      A_Half_Dep.The_Task :=
         Search_Task(My_Tasks,A_Task);
      A_Half_Dep.The_Type := A_Type;

      Add(My_Dependencies,A_Half_Dep);
   end Private_Add_Half_Dependency;


   procedure Private_Add_Double_Dependency (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String;   
         My_Deps         : in     Dependency_Ptr      ) is 

      Found_To    : Boolean            := False;  
      Found_From  : Boolean            := False;  
      I,  
      J           : Dependencies_Range;  
      My_Iterator : Tasks_Iterator;  
      A_Task      : Task_Ptr;  
   begin


      -- Check/insert task name
      --
      for I in  0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if(My_Dependencies.Task_Position(I).Name = From_Task)
               then
            Found_From:=True;
         end if;
         if(My_Dependencies.Task_Position(I).Name = To_Task)
               then
            Found_To:=True;
         end if;
      end loop;


      if (not Found_To)
            then
         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Tasks, A_Task, My_Iterator);

            exit when To_Task = A_Task.Name;

            if (Is_Last_Element(My_Tasks, My_Iterator) = True)
                  then
               raise Task_Not_Found;
            end if;

            Next_Element(My_Tasks, My_Iterator);


         end loop;

         My_Dependencies.Task_Position(My_Dependencies.Number_Of_Tasks):=
            A_Task;
         My_Dependencies.Number_Of_Tasks:=My_Dependencies.Number_Of_Tasks +
            1;
      end if;

      if (not Found_From)
            then
         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Tasks,A_Task,My_Iterator);

            exit when From_Task = A_Task.Name;

            if (Is_Last_Element(My_Tasks, My_Iterator) = True)
                  then
               raise Task_Not_Found;
            end if;

            Next_Element(My_Tasks, My_Iterator);

         end loop;

         My_Dependencies.Task_Position(My_Dependencies.Number_Of_Tasks) :=
            A_Task;
         My_Dependencies.Number_Of_Tasks := My_Dependencies.
            Number_Of_Tasks +
            1;
      end if;

      -- Add dependencies

      I:=Search_Position(My_Dependencies,From_Task);
      J:=Search_Position(My_Dependencies,To_Task);



      Add(My_Dependencies.Dependencies(I,J),My_Deps);
   end Private_Add_Double_Dependency;


   -- Add double dependencies
   --
   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task_Name  : in     Unbounded_String;   
         To_Task_Name    : in     Unbounded_String;   
         My_Deps         : in     Dependency_Ptr      ) is 
   begin
      Private_Add_Double_Dependency (
         My_Dependencies, My_Tasks, From_Task_Name,
         To_Task_Name, My_Deps);

      -- Adding a complete dependency imply adding 2 half
      -- dependencies if buffer or communication dependencies
      -- or only one half dependencie if precedence dependency
      --

      if (My_Deps.Dependence_Type=Precedence_Dependency)
            then
         Private_Add_Half_Dependency(
            My_Dependencies, My_Tasks, From_Task_Name,
            My_Deps, To_Task);
      else
         Private_Add_Half_Dependency(
            My_Dependencies, My_Tasks, From_Task_Name,
            My_Deps, To_Task);
         Private_Add_Half_Dependency(
            My_Dependencies, My_Tasks, To_Task_Name,
            My_Deps, To_Task);
      end if;

   end Add_Tasks_Dependencies;


   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task_Name  : in     Unbounded_String;   
         To_Task_Name    : in     Unbounded_String    ) is 

      My_Deps : Precedence_Ptr;  

   begin
      My_Deps:= new Precedence;
      My_Deps.From_Task:=From_Task_Name;
      My_Deps.To_Task:=To_Task_Name;

      Add_Tasks_Dependencies (
         My_Dependencies, My_Tasks,
         From_Task_Name, To_Task_Name, Dependency_Ptr(My_Deps));
   end Add_Tasks_Dependencies;


   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task_Name  : in     Task_Ptr;           
         To_Task_Name    : in     Task_Ptr            ) is 

      My_Deps : Precedence_Ptr;  

   begin
      My_Deps:= new Precedence;
      Add_Tasks_Dependencies (
         My_Dependencies, My_Tasks,
         From_Task_Name.Name, To_Task_Name.Name, Dependency_Ptr(My_Deps));
   end Add_Tasks_Dependencies;

   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task_Name  : in     Task_Ptr;           
         To_Task_Name    : in     Task_Ptr;           
         My_Deps         : in     Dependency_Ptr      ) is 

   begin
      Add_Tasks_Dependencies (
         My_Dependencies, My_Tasks,
         From_Task_Name.Name, To_Task_Name.Name, My_Deps);
   end Add_Tasks_Dependencies;

   -- Add half dependencies
   --
   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         A_Task          : in     Unbounded_String;   
         A_Dep           : in     Dependency_Ptr;     
         A_Type          : in     Half_Dep_Type       ) is 

      My_Iterator : Tasks_Dependencies_Iterator;  
      My_Half_Dep : Half_Dep_Ptr;  

   begin
      Private_Add_Half_Dependency(My_Dependencies,
         My_Tasks, A_Task, A_Dep, A_Type);

      -- Adding a half dependency can lead to create
      -- a new complete dependency
      --
      if not Is_Empty(Half_Dep_Set.Set(My_Dependencies)) then
         Reset_Iterator(My_Iterator);
         loop
            Current_Element(My_Dependencies,My_Half_Dep,My_Iterator);

            -- Looking or the other half dependency
            --
            if (My_Half_Dep.The_Dep = A_Dep)
                  then
               if
                     ((My_Half_Dep.The_Type=To_Task)
                     and (A_Type=To_Dependency))
                     then
                  Private_Add_Double_Dependency (
                     My_Dependencies, My_Tasks, A_Task, My_Half_Dep.
                     The_Task.Name, A_Dep);
               end if;
               if ((My_Half_Dep.The_Type=To_Dependency)
                     and (A_Type=To_Task))
                     then
                  Private_Add_Double_Dependency (
                     My_Dependencies, My_Tasks, My_Half_Dep.The_Task.Name,
                     A_Task, A_Dep);
               end if;

            end if;

            exit when Is_Last_Element(My_Dependencies, My_Iterator) = True;
            Next_Element(My_Dependencies, My_Iterator);
         end loop;
      end if;

   end Add_Tasks_Dependencies;


   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         A_Task          : in     Task_Ptr;           
         A_Dep           : in     Dependency_Ptr;     
         A_Type          : in     Half_Dep_Type       ) is 

   begin
      Add_Tasks_Dependencies(My_Dependencies, My_Tasks,
         A_Task.Name, A_Dep, A_Type);
   end Add_Tasks_Dependencies;



   -------------------------------------------------------------


   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         A_Task          : in     Task_Ptr;           
         A_Dep           : in     Unbounded_String;   
         My_Dep_Type     : in     Dependency_Type;    
         A_Type          : in     Half_Dep_Type       ) is 

   begin
      Delete_Tasks_Dependencies(My_Dependencies, A_Task.Name, A_Dep,
         My_Dep_Type, A_Type);
   end Delete_Tasks_Dependencies;


   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         A_Task          : in     Unbounded_String;   
         A_Dep           : in     Unbounded_String;   
         My_Dep_Type     : in     Dependency_Type;    
         A_Type          : in     Half_Dep_Type       ) is 


      My_Iterator2 : Tasks_Dependencies_Iterator;  
      A_Half_Dep   : Half_Dep_Ptr;  
      Tmp2         : Tasks_Dependencies;  

   begin
      Put_Line("delete " & To_String(A_Task) & "  " & To_String(A_Dep)
         & "  " & A_Type'Img);

      if not Is_Empty(My_Dependencies) then
         Reset_Iterator(My_Iterator2);
         loop
            Current_Element(My_Dependencies, A_Half_Dep ,My_Iterator2);


            if (A_Half_Dep.The_Task.Name = A_Task)
                  and (A_Half_Dep.The_Dep.Dependence_Type = My_Dep_Type)
                  and (A_Half_Dep.The_Type = A_Type)

                  then
               if (A_Half_Dep.The_Dep.Dependence_Type = Buffer_Dependency)
                     then
                  if (Buffer_Ptr(A_Half_Dep.The_Dep).Name = A_Dep)
                        then
                     Add(Tmp2, A_Half_Dep );
                  end if;
               else
                  if (Message_Ptr(A_Half_Dep.The_Dep).Name = A_Dep)
                        then
                     Add(Tmp2, A_Half_Dep );
                  end if;
               end if;
            end if;


            exit when Is_Last_Element(My_Dependencies, My_Iterator2) =
               True;
            Next_Element(My_Dependencies,My_Iterator2);
         end loop;
      end if;


      if not Is_Empty(Tmp2) then
         Reset_Iterator(My_Iterator2);
         loop
            Current_Element(Tmp2, A_Half_Dep ,My_Iterator2);

            Delete (My_Dependencies, A_Half_Dep, False);

            exit when Is_Last_Element(Tmp2, My_Iterator2) = True;
            Next_Element(Tmp2,My_Iterator2);
         end loop;
         Reset(Tmp2,False);
      end if;

   end Delete_Tasks_Dependencies;



   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr            ) is 

   begin
      Delete_Tasks_Dependencies(My_Dependencies,From_Task.Name,To_Task.
         Name);
   end Delete_Tasks_Dependencies;

   Tmp2 : Tasks_Dependencies;  


   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String    ) is 


      I,  
      J            : Dependencies_Range;  
      My_Iterator2 : Tasks_Dependencies_Iterator;  
      A_Half_Dep   : Half_Dep_Ptr;  


   begin
      if not Is_Empty(My_Dependencies) then

         I:=Search_Position(My_Dependencies,From_Task);
         J:=Search_Position(My_Dependencies,To_Task);

         -- Delete all dependencies between I to J
         --
         Reset(My_Dependencies.Dependencies(I,J));


         -- Update the set
         --
         Reset_Iterator(My_Iterator2);
         loop
            Current_Element(My_Dependencies, A_Half_Dep ,My_Iterator2);


            -- Delete precedency dependencies
            --
            if (A_Half_Dep.The_Dep.Dependence_Type = Precedence_Dependency)
                  then
               if (Precedence_Ptr(A_Half_Dep.The_Dep).From_Task =
                     From_Task)
                     and (Precedence_Ptr(A_Half_Dep.The_Dep).To_Task =
                     To_Task)
                     then
                  Add(Tmp2, A_Half_Dep);
               end if;
            end if;


            exit when Is_Last_Element(My_Dependencies, My_Iterator2) =
               True;
            Next_Element(My_Dependencies,My_Iterator2);
         end loop;
      end if;


      if not Is_Empty(Tmp2) then
         Reset_Iterator(My_Iterator2);
         loop
            Current_Element(Tmp2, A_Half_Dep ,My_Iterator2);

            Delete (My_Dependencies, A_Half_Dep, False);

            exit when Is_Last_Element(Tmp2, My_Iterator2) = True;
            Next_Element(Tmp2,My_Iterator2);
         end loop;
         Reset(Tmp2,False);
      end if;




   end Delete_Tasks_Dependencies;




   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr;           
         Dep_Type        : in     Dependency_Type;    
         Name            : in     Unbounded_String    ) is 

   begin
      Delete_Tasks_Dependencies( My_Dependencies,From_Task.Name,To_Task.
         Name,Dep_Type,Name);
   end Delete_Tasks_Dependencies;

   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String;   
         Dep_Type        : in     Dependency_Type;    
         Name            : in     Unbounded_String    ) is 

      My_Iterator : Depends_Iterator;  
      My_Dep      : Dependency_Ptr;  
      I,  
      J           : Dependencies_Range;  

   begin

      if not Is_Empty(My_Dependencies) then
         I:=Search_Position(My_Dependencies,From_Task);
         J:=Search_Position(My_Dependencies,To_Task);

         -- search of the dependencie
         if not Is_Empty(My_Dependencies.Dependencies(I,J)) then
            Reset_Iterator(My_Iterator);
            loop
               Current_Element(My_Dependencies.Dependencies(I,J),My_Dep,
                  My_Iterator);

               if (My_Dep.Dependence_Type = Dep_Type) then
                  if (My_Dep.Dependence_Type = Communication_Dependency) and then (
                        Message_Ptr(My_Dep).Name = Name) then
                     Delete(My_Dependencies.Dependencies(I,J),My_Dep);
                     exit;
                  end if;
                  if (My_Dep.Dependence_Type = Buffer_Dependency) and then (
                        Buffer_Ptr(My_Dep).Name = Name) then
                     Delete(My_Dependencies.Dependencies(I,J),My_Dep);
                     exit;
                  end if;
               end if;
               exit when Is_Last_Element(My_Dependencies.Dependencies(I,J),
                  My_Iterator) = True;
               Next_Element(My_Dependencies.Dependencies(I,J),My_Iterator);
            end loop;
         end if;
      end if;
   end Delete_Tasks_Dependencies;



   procedure Delete_Processors_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         A_Processor     : in     Unbounded_String    ) is 


      My_Iterator2 : Tasks_Dependencies_Iterator;  
      A_Half_Dep   : Half_Dep_Ptr;  
      Tmp2         : Tasks_Dependencies;  

      My_Iterator : Depends_Iterator;  
      My_Dep      : Dependency_Ptr;  
      Tmp         : Depends_Set;  

   begin


      if not Is_Empty(My_Dependencies) then

         ----------------------------------
         ---  Update the matrice first
         ----------------------------------

         for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
            for J in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
               if not Is_Empty(My_Dependencies.Dependencies(I,J)) then
                  Reset_Iterator(My_Iterator);
                  loop
                     Current_Element(My_Dependencies.Dependencies(I,J),
                        My_Dep,My_Iterator);

                     if (My_Dep.Dependence_Type = Buffer_Dependency) then
                        if (Buffer_Ptr(My_Dep).Location = A_Processor) then
                           Add(Tmp, My_Dep);
                        end if;
                     end if;


                     if (My_Dep.Dependence_Type = Precedence_Dependency) then

                        -- Looking for the task in the set
                        -- (to see on which processor the task is located)
                        --
                        if not Is_Empty(My_Dependencies) then
                           Reset_Iterator(My_Iterator2);
                           loop
                              Current_Element(My_Dependencies, A_Half_Dep ,
                                 My_Iterator2);


                              if (A_Half_Dep.The_Task.Location=
                                    A_Processor)
                                    then
                                 Add(Tmp,My_Dep);
                                 exit;
                              end if;

                              exit when Is_Last_Element(My_Dependencies,
                                 My_Iterator2) = True;
                              Next_Element(My_Dependencies,My_Iterator2);
                           end loop;

                        end if;
                     end if;


                     exit when Is_Last_Element(My_Dependencies.
                        Dependencies(
                           I,J), My_Iterator) = True;
                     Next_Element(My_Dependencies.Dependencies(I,J),
                        My_Iterator);
                  end loop;


                  -- Clear tmp	
                  --
                  if not Is_Empty(Tmp) then
                     Reset_Iterator(My_Iterator);
                     loop
                        Current_Element(Tmp, My_Dep ,My_Iterator);


                        Delete(My_Dependencies.Dependencies(I,J),My_Dep,
                           False);

                        exit when Is_Last_Element(Tmp, My_Iterator) = True;
                        Next_Element(Tmp,My_Iterator);
                     end loop;
                     Reset(Tmp,False);
                  end if;



               end if;
            end loop;
         end loop;


         ----------------------------------
         ---  Update the set
         ----------------------------------

         if not Is_Empty(My_Dependencies) then
            Reset_Iterator(My_Iterator2);
            loop
               Current_Element(My_Dependencies, A_Half_Dep ,My_Iterator2);


               if (A_Half_Dep.The_Task.Location=  A_Processor)
                     then
                  Add(Tmp2, A_Half_Dep );
               end if;

               exit when Is_Last_Element(My_Dependencies, My_Iterator2) =
                  True;
               Next_Element(My_Dependencies,My_Iterator2);
            end loop;
         end if;


         if not Is_Empty(Tmp2) then
            Reset_Iterator(My_Iterator2);
            loop
               Current_Element(Tmp2, A_Half_Dep ,My_Iterator2);

               Delete (My_Dependencies, A_Half_Dep, False);

               exit when Is_Last_Element(Tmp2, My_Iterator2) = True;
               Next_Element(Tmp2,My_Iterator2);
            end loop;
            Reset(Tmp2,False);
         end if;

      end if;

   end Delete_Processors_Dependencies;






   procedure Delete_All_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Dep_Type     : in     Dependency_Type;    
         My_Dep_Name     : in     Unbounded_String    ) is 

      My_Iterator2 : Tasks_Dependencies_Iterator;  
      A_Half_Dep   : Half_Dep_Ptr;  
      Tmp2         : Tasks_Dependencies;  

      My_Iterator : Depends_Iterator;  
      My_Dep      : Dependency_Ptr;  
      I,  
      J           : Dependencies_Range;  
      Found_From  : Boolean            := False;  
      Found_To    : Boolean            := False;  

   begin

      if not Is_Empty(My_Dependencies) then

         ----------------------------------
         ---  Update the set
         ----------------------------------
         Reset_Iterator(My_Iterator2);
         loop
            Current_Element(My_Dependencies, A_Half_Dep ,My_Iterator2);


            if (My_Dep_Type = Precedence_Dependency)
                  then
               if Dependency_Ptr(A_Half_Dep.The_Dep).Dependence_Type=
                     Precedence_Dependency
                     then
                  if (Precedence_Ptr(A_Half_Dep.The_Dep).From_Task =
                        My_Dep_Name)
                        or  (Precedence_Ptr(A_Half_Dep.The_Dep).To_Task =
                        My_Dep_Name)
                        then
                     Add(Tmp2, A_Half_Dep);
                  end if;
               end if;
            end if;

            if (My_Dep_Type = Buffer_Dependency)
                  then
               if Dependency_Ptr(A_Half_Dep.The_Dep).Dependence_Type=
                     Buffer_Dependency
                     then
                  if (Buffer_Ptr(A_Half_Dep.The_Dep).Name =  My_Dep_Name)
                        then
                     Add(Tmp2, A_Half_Dep);
                  end if;
               end if;
            end if;

            if (My_Dep_Type = Communication_Dependency)
                  then
               if Dependency_Ptr(A_Half_Dep.The_Dep).Dependence_Type=
                     Communication_Dependency
                     then
                  if (Message_Ptr(A_Half_Dep.The_Dep).Name =  My_Dep_Name)
                        then
                     Add(Tmp2, A_Half_Dep);
                  end if;
               end if;
            end if;

            exit when Is_Last_Element(My_Dependencies, My_Iterator2) =
               True;
            Next_Element(My_Dependencies,My_Iterator2);
         end loop;



         -- Clear tmp2
         --
         if not Is_Empty(Tmp2) then
            Reset_Iterator(My_Iterator2);
            loop
               Current_Element(Tmp2, A_Half_Dep ,My_Iterator2);

               Delete (My_Dependencies, A_Half_Dep, False);

               exit when Is_Last_Element(Tmp2, My_Iterator2) = True;
               Next_Element(Tmp2, My_Iterator2);
            end loop;
            Reset(Tmp2);
         end if;



         ----------------------------------
         ---  Update the matrice
         ----------------------------------

         if (My_Dep_Type = Precedence_Dependency) then

            -- search task name in from and to
            --
            for I in  0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
               if(My_Dependencies.Task_Position(I).Name = My_Dep_Name)
                     then
                  Found_From:=True;
               end if;
               if(My_Dependencies.Task_Position(I).Name = My_Dep_Name)
                     then
                  Found_To:=True;
               end if;
            end loop;

            -- search and delete on line I
            --
            if Found_From then
               I:=Search_Position(My_Dependencies,My_Dep_Name);
               for J in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
                  Reset(My_Dependencies.Dependencies(I,J),False);
               end loop;
            end if;
            -- search and delete on column I
            --
            if Found_To then
               J:=Search_Position(My_Dependencies,My_Dep_Name);
               for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
                  Reset(My_Dependencies.Dependencies(I,J),False);
               end loop;
            end if;

         else
            -- delete in the task_dependencies
            --
            if (My_Dep_Type = Buffer_Dependency) then
               for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
                  for J in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
                     if not Is_Empty(My_Dependencies.Dependencies(I,J)) then
                        Reset_Iterator(My_Iterator);
                        loop
                           Current_Element(My_Dependencies.Dependencies(I,
                                 J),
                              My_Dep,My_Iterator);
                           if (My_Dep.Dependence_Type = Buffer_Dependency) then
                              if (Buffer_Ptr(My_Dep).Name = My_Dep_Name) then
                                 Delete(My_Dependencies.Dependencies(I,J),
                                    My_Dep);
                                 exit;
                              end if;
                           end if;
                           exit when Is_Last_Element(My_Dependencies.
                              Dependencies(I,J), My_Iterator) = True;
                           Next_Element(My_Dependencies.Dependencies(I,J),
                              My_Iterator);
                        end loop;
                     end if;
                  end loop;
               end loop;
            else
               if (My_Dep_Type = Communication_Dependency) then
                  for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
                     for J in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
                        if not Is_Empty(My_Dependencies.Dependencies(I,J)) then
                           Reset_Iterator(My_Iterator);
                           loop
                              Current_Element(My_Dependencies.
                                 Dependencies(I,
                                    J),My_Dep,My_Iterator);
                              if (My_Dep.Dependence_Type =
                                    Communication_Dependency) then
                                 if (Message_Ptr(My_Dep).Name =
                                       My_Dep_Name) then
                                    Delete(My_Dependencies.Dependencies(I,
                                          J),
                                       My_Dep);
                                    exit;
                                 end if;
                              end if;
                              exit when Is_Last_Element(My_Dependencies.
                                 Dependencies(I,J), My_Iterator) = True;
                              Next_Element(My_Dependencies.Dependencies(I,
                                    J),
                                 My_Iterator);
                           end loop;
                        end if;
                     end loop;
                  end loop;
               end if;
            end if;
         end if;


      end if;
   end Delete_All_Dependencies;

   -------------------------------------------------------------

   function Copy (
         My_Half : in     Half_Dep_Ptr ) 
     return Half_Dep_Ptr is 

      Ptr : Half_Dep_Ptr;  
   begin
      Ptr:=new Half_Dep;
      Ptr.all:=My_Half.all;

      return Ptr;
   end Copy;



   procedure Put (
         My_Half_Dep : in     Half_Dep ) is 
   begin
      if (My_Half_Dep.The_Dep.Dependence_Type = Buffer_Dependency) then
         if My_Half_Dep.The_Type = To_Task then
            Put_Line(To_String(Buffer_Ptr(My_Half_Dep.The_Dep).Name) &
               " -> " & To_String(My_Half_Dep.The_Task.Name));
         else
            Put_Line(To_String(My_Half_Dep.The_Task.Name) & " -> " &
               To_String(Buffer_Ptr(My_Half_Dep.The_Dep).Name));
         end if;
      else

         if (My_Half_Dep.The_Dep.Dependence_Type =
               Communication_Dependency) then
            if My_Half_Dep.The_Type = To_Task then
               Put_Line(To_String(Message_Ptr(My_Half_Dep.The_Dep).Name) &
                  " -> " & To_String(My_Half_Dep.The_Task.Name));
            else
               Put_Line(To_String(My_Half_Dep.The_Task.Name) & " -> " &
                  To_String(Message_Ptr(My_Half_Dep.The_Dep).Name));
            end if;

         else
            Put_Line(To_String(Precedence_Ptr(My_Half_Dep.The_Dep).
                  From_Task) & " -> "  &
               To_String(Precedence_Ptr(My_Half_Dep.The_Dep).To_Task) );
         end if;
      end if;
   end Put;


   procedure Put (
         My_Half : in     Half_Dep_Ptr ) is 
   begin
      Put(My_Half.all);
   end Put;


   -------------------------------------------------------------

   package Dep_File is new Ada.Sequential_Io(Half_Dep_Descriptor);
   use Dep_File;

   procedure Read_From_File (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         My_Messages     : in     Messages_Set;       
         My_Buffers      : in     Buffers_Set;        
         File_Name       : in     Unbounded_String    ) is 

      Desc       : Half_Dep_Descriptor;  
      A_Dep_File : Dep_File.File_Type;  

      A_Buff        : Buffer_Ptr;  
      A_Mess        : Message_Ptr;  
      Buff_Iterator : Buffers_Iterator;  
      Mess_Iterator : Messages_Iterator;  
   begin

      Reset(My_Dependencies);

      Open(A_Dep_File,
         Mode => In_File,
         Name => To_String(File_Name));


      if not Is_Open(A_Dep_File)
            then
         raise Can_Not_Open_File_For_Reading;
      end if;


      while not End_Of_File(A_Dep_File)
            loop
         Read(A_Dep_File,Desc);

         if (Desc.Depend_Type=Precedence_Dependency)
               then
            Add_Tasks_Dependencies(My_Dependencies,
               My_Tasks,
               To_Unbounded_String(To_String(Desc.Task_Name)),
               To_Unbounded_String(To_String(Desc.Depend_Name)));
         else
            if (Desc.Depend_Type=Buffer_Dependency)
                  then

               Reset_Iterator(Buff_Iterator);

               loop
                  Current_Element(My_Buffers, A_Buff ,Buff_Iterator);

                  if A_Buff.Name = To_Unbounded_String(To_String(Desc.
                           Depend_Name))
                        then

                     Add_Tasks_Dependencies(My_Dependencies,
                        My_Tasks,
                        To_Unbounded_String(To_String(Desc.Task_Name)),
                        Dependency_Ptr(A_Buff),
                        Desc.Depend_Position);
                     exit;
                  end if;
                  if Is_Last_Element(My_Buffers, Buff_Iterator) = True
                        then
                     raise Buffer_Set.Buffer_Not_Found;
                  end if;

                  Next_Element(My_Buffers, Buff_Iterator);
               end loop;
            else
               Reset_Iterator(Mess_Iterator);

               loop
                  Current_Element(My_Messages, A_Mess, Mess_Iterator);

                  if A_Mess.Name = To_Unbounded_String(To_String(Desc.
                           Depend_Name))
                        then

                     Add_Tasks_Dependencies(My_Dependencies,
                        My_Tasks,
                        To_Unbounded_String(To_String(Desc.Task_Name)),
                        Dependency_Ptr(A_Mess),
                        Desc.Depend_Position);
                     exit;

                  end if;
                  if Is_Last_Element(My_Messages, Mess_Iterator) = True
                        then
                     raise Message_Set.Message_Not_Found;
                  end if;

                  Next_Element(My_Messages, Mess_Iterator);

               end loop;

            end if;
         end if;

      end loop;

      Close(A_Dep_File);

   end Read_From_File;


   procedure Fill_Descriptor (
         My_Dependencies : in     Tasks_Dependencies;  
         Desc            :    out Half_Dep_Descriptor; 
         A_Dependency    : in     Half_Dep_Ptr         ) is 
   begin

      Desc.Task_Name:=To_Bounded_String(To_String(A_Dependency.The_Task.
            Name));
      Desc.Depend_Type:=A_Dependency.The_Dep.Dependence_Type;
      Desc.Depend_Position:=A_Dependency.The_Type;

      if (Desc.Depend_Type = Communication_Dependency)
            then
         Desc.Depend_Name:=To_Bounded_String(To_String(Message_Ptr(
                  A_Dependency.The_Dep).Name));
      else
         if (Desc.Depend_Type = Buffer_Dependency)
               then
            Desc.Depend_Name:=To_Bounded_String(To_String(Buffer_Ptr(
                     A_Dependency.The_Dep).Name));
         else
            Desc.Task_Name:=To_Bounded_String(To_String(Precedence_Ptr(
                     A_Dependency.The_Dep).From_Task));
            Desc.Depend_Name:=To_Bounded_String(To_String(Precedence_Ptr(
                     A_Dependency.The_Dep).To_Task));
         end if;
      end if;

   end Fill_Descriptor;



   procedure Duplicate (
         Src  : in     Tasks_Dependencies; 
         Dest : in out Tasks_Dependencies  ) is 

   begin

      Duplicate(Half_Dep_Set.Set(Src), Half_Dep_Set.Set(Dest));

      for I in  Dependencies_Range loop
         for J in  Dependencies_Range loop

            Duplicate(Src.Dependencies(I,J),
               Dest.Dependencies(I,J)  );
         end loop;
      end loop;

      Dest.Number_Of_Tasks:=Src.Number_Of_Tasks;
      Dest.Task_Position:=Src.Task_Position;

   end Duplicate;



   procedure Reset (
         My_Dependencies : in out Tasks_Dependencies;         
         Free_Object     : in     Boolean            := False ) is 

      Root_Task : Task_Ptr;  

   begin

      Reset(Half_Dep_Set.Set(My_Dependencies), Free_Object);

      My_Dependencies.Number_Of_Tasks:=0;

      Root_Task := new Periodic_Task;
      Root_Task.Name:=To_Unbounded_String("Root");

      My_Dependencies.Task_Position(Root_Index):=Root_Task;

      for I in  Dependencies_Range loop
         for J in  Dependencies_Range loop

            Reset(My_Dependencies.Dependencies(I,J),Free_Object);
         end loop;
      end loop;

   end Reset;



   function Is_Empty (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Boolean is 


      Set_Is_Empty : Boolean := Is_Empty (Half_Dep_Set.Set (My_Dependencies));
         





   begin

      -- A easy and quick test : no double dependencies defined
      --
      if ((My_Dependencies.Number_Of_Tasks=0) and Set_Is_Empty)
            then
         return True;
      end if;


      -- We must each row/colum to see if there is no dependencies
      -- 
      for I in  Dependencies_Range loop
         for J in  Dependencies_Range loop

            if not Is_Empty(My_Dependencies.Dependencies(I,J))
                  then
               return False;
            end if;
         end loop;
      end loop;



      if Set_Is_Empty
            then
         return True;
      else
         return False;
      end if;



   end Is_Empty;



   procedure Put (
         My_Dependencies : in     Tasks_Dependencies ) is 

      My_Iterator : Depends_Iterator;  
      A_Dep       : Dependency_Ptr;  
   begin


      Put(Half_Dep_Set.Set(My_Dependencies));


      if My_Dependencies.Number_Of_Tasks /= 0
            then
         Put("Task list/index : ");
         New_Line;
         for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
            Put(" Task " & To_String(My_Dependencies.Task_Position(I).
                  Name) &
               " index is " & I'Img);
            New_Line;
         end loop;

         Put(" Task " & To_String(My_Dependencies.Task_Position(
                  Root_Index).Name) &
            " index is " & Root_Index'Img);
         New_Line;


         Put("Dependencies : ");
         New_Line;
         for I in  Dependencies_Range loop
            for J in  Dependencies_Range loop

               if (not Is_Empty(My_Dependencies.Dependencies(I, J)))
                     then
                  Put(To_String(My_Dependencies.Task_Position(I).Name) &
                     " => "
                     & To_String(My_Dependencies.Task_Position(J).Name));
                  Put(" ( ");
                  Reset_Iterator(My_Iterator);
                  loop
                     Current_Element(My_Dependencies.Dependencies(I, J),
                        A_Dep,My_Iterator);
                     if (A_Dep.Dependence_Type = Precedence_Dependency) then
                        Put("precedence ");
                     else
                        if (A_Dep.Dependence_Type = Buffer_Dependency)  then
                           Put("buffer/");
                           Put_Name(Buffer_Ptr(A_Dep));
                           Put(" ");
                        else
                           Put("communication/");
                           Put_Name(Message_Ptr(A_Dep));
                           Put(" ");
                        end if;
                     end if;
                     exit when Is_Last_Element(My_Dependencies.
                        Dependencies(I, J),
                        My_Iterator) = True;
                     Next_Element(My_Dependencies.Dependencies(I, J),
                        My_Iterator);
                  end loop;

                  Put(")");
                  New_Line;
               end if;

            end loop;
         end loop;

      end if;
   end Put;



   -- Dependencies set used to internal computation
   -- Can not be declared in the stack due to its size
   --
   Tmp_Deps : Tasks_Dependencies;  


   package Internal_Stacks is new Stacks(Config.Max_Tasks,
      Task_Ptr, Put);
   use Internal_Stacks;


   -- Check is a task dependencies is acyclic or not
   --
   function Is_Acyclic (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Boolean is 


      To_Visit  : Internal_Stacks.Stack;  
      Leaf,  
      Visited,  
      Backwards : Tasks_Set;  
      Right,  
      Left,  
      Current,  
      Back      : Task_Ptr;  
      Ite1,  
      Ite2,  
      Ite3      : Tasks_Iterator;  

   begin

      Reset(Tmp_Deps);
      Duplicate(My_Dependencies, Tmp_Deps);


      -- Push every leaves of the graph
      --
      Leaf:=Get_Leaf_Tasks(Tmp_Deps);

      Reset_Iterator(Ite1);
      for I in 0 .. Get_Number_Of_Elements(Leaf)-1 loop

         Current_Element(Leaf, Current, Ite1);

         Reset(To_Visit);
         Push(To_Visit, Current);
         Reset(Visited);
         Add(Visited, Current);

         while not Is_Empty(To_Visit) loop


            Right:=Get_Top(To_Visit);
            if Has_Predecessor(Tmp_Deps, Right)
                  then


               Left:=Get_A_Predecessor(Tmp_Deps, Right);
               Reset(Backwards);
               Backwards:=Get_Predecessors_List(Tmp_Deps, Left);

               -- Check if backward arrows exists. If so, the 
               -- graph is cyclic
               --
               if not Is_Empty(Backwards) then
                  Reset_Iterator(Ite2);
                  for I in 0 .. Get_Number_Of_Elements(Visited)-1 loop
                     Current_Element(Visited, Current, Ite2);

                     -- Looking for backward arrows
                     --
                     Reset_Iterator(Ite3);
                     for J in 0 .. Get_Number_Of_Elements(Backwards)-1 loop
                        Current_Element(Backwards, Back, Ite3);
                        if Back.Name = Current.Name
                              then
                           return False;
                        end if;
                        Next_Element(Backwards, Ite3);
                     end loop;

                     Next_Element(Visited, Ite2);
                  end loop;
               end if;

               Delete_Tasks_Dependencies(Tmp_Deps, Left, Right);
               if  Has_Predecessor(Tmp_Deps, Left) then
                  Add(Visited, Left);
                  Push(To_Visit, Left);
               end if;



            else
               Pop(To_Visit, Right);
            end if;

         end loop;

         Next_Element(Leaf, Ite1);
      end loop;



      -- No more leaf but possible cyclic dependencies 
      -- Test them ...
      --
      for I in  Dependencies_Range loop
         for J in  Dependencies_Range loop

            if not Is_Empty(Tmp_Deps.Dependencies(I, J))
                  then
               return False;
            end if;

         end loop;
      end loop;



      return True;
   end Is_Acyclic;



   -- Build dependencies from the definition
   -- of a buffer
   --
   procedure Build_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         A_Buffer        : in     Buffer_Ptr          ) is 

      My_Half_Dep_Type : Half_Dep_Type;  

   begin

      Delete_All_Dependencies(My_Dependencies,
         Buffer_Dependency, A_Buffer.Name);

      for I in 0..A_Buffer.Roles.Nb_Entries-1 loop
         if A_Buffer.Roles.Entries(I).Data.The_Role=Producer
               then

            My_Half_Dep_Type := To_Dependency;
         else
            My_Half_Dep_Type := To_Task;
         end if;

         Add_Tasks_Dependencies(My_Dependencies,
            My_Tasks,
            A_Buffer.Roles.Entries(I).Item,
            Dependency_Ptr(A_Buffer),
            My_Half_Dep_Type);
      end loop;


   end Build_Dependencies;



   procedure Export_Xml (
         My_Dependencies : in     Tasks_Dependencies; 
         Into            : in out Text_Io.File_Type   ) is 


      My_Iterator : Tasks_Dependencies_Iterator;  
      A_Half_Dep  : Half_Dep_Ptr;  


   begin

      if not Is_Empty(My_Dependencies) then

         Reset_Iterator(My_Iterator);


         Put_Line(Into, "<dependencies>");
         loop
            Current_Element(My_Dependencies,A_Half_Dep,My_Iterator);

            Put(Into,  Ascii.Ht & "<dependency");

            if A_Half_Dep.The_Dep.Dependence_Type=Buffer_Dependency then
               if A_Half_Dep.The_Type = To_Dependency then
                  Put(Into, " to_type=""buffer""> " &
                     To_String(A_Half_Dep.The_Task.Name) & " " & To_String(
                        Buffer_Ptr(A_Half_Dep.The_Dep).Name) & " ");
               else
                  Put(Into, " from_type=""buffer""> " &
                     To_String(Buffer_Ptr(A_Half_Dep.The_Dep).Name) & " " &
                     To_String(A_Half_Dep.The_Task.Name) & " ");
               end if;
            else
               if A_Half_Dep.The_Dep.Dependence_Type=
                     Communication_Dependency then
                  if A_Half_Dep.The_Type = To_Dependency then
                     Put(Into, " to_type=""message""> " &
                        To_String(A_Half_Dep.The_Task.Name) & " " &
                        To_String(Message_Ptr(A_Half_Dep.The_Dep).Name) &
                        " ");
                  else
                     Put(Into, " from_type=""message""> " &
                        To_String(Message_Ptr(A_Half_Dep.The_Dep).Name) &
                        " " & To_String(A_Half_Dep.The_Task.Name) & " ");
                  end if;
               else
                  if A_Half_Dep.The_Type = To_Dependency then
                     Put(Into, "> " & To_String(A_Half_Dep.The_Task.Name) &
                        " " & To_String(Precedence_Ptr(A_Half_Dep.The_Dep).
                           To_Task) & " ");
                  else
                     Put(Into, "> " & To_String(Precedence_Ptr(A_Half_Dep.
                              The_Dep).From_Task) & " " & To_String(
                           Precedence_Ptr(A_Half_Dep.The_Dep).To_Task) &
                        " ");
                  end if;
               end if;
            end if;

            Put_Line(Into,  "</dependency>");

            exit when Is_Last_Element(My_Dependencies, My_Iterator) = True;

            Next_Element(My_Dependencies, My_Iterator);

         end loop;
         Put_Line(Into, "</dependencies>");


      end if;


   end Export_Xml;



end Task_Dependencies;

