------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------



with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Fixed_Strings;
use Fixed_Strings;

with Text_Io;
use Text_Io;

with Config;
use Config;

with Sets;

with Tasks;
use Tasks;
with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Buffers;
use Buffers;
with Buffer_Set;
use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Message_Set;
use Message_Set;
use Message_Set.Generic_Message_Set;


with Dependencies;
use Dependencies;
use Dependencies.Dep_Set;


with Unchecked_Deallocation;


package Task_Dependencies is




   Bad_Use_Of_Procedure : exception;  
   Exist_Already        : exception;  
   Task_Not_Found       : exception;  
   Message_Not_Found    : exception;  
   Buffer_Not_Found     : exception;  
   No_Dependencies      : exception;  



   ---------------------------------------
   --  Define dependencies type :
   --  it's set extended by a 2D array
   --  of generic dependencies set
   ---------------------------------------
   ---------------------------------------
   -- First of all : a half dependencies 
   -- between two tasks
   ---------------------------------------
   type Half_Dep_Type is 
         (To_Task,      
          To_Dependency); 

   type Half_Dep is tagged 
      record 
         The_Task : Task_Ptr;  
         The_Dep  : Dependency_Ptr;  
         The_Type : Half_Dep_Type;  
      end record; 

   procedure Put (
         My_Half_Dep : in     Half_Dep ); 

   type Half_Dep_Ptr is access all Half_Dep'Class; 

   procedure Free is 
   new Unchecked_Deallocation
      (Half_Dep'Class, Half_Dep_Ptr);

   procedure Put (
         My_Half : in     Half_Dep_Ptr ); 

   function Copy (
         My_Half : in     Half_Dep_Ptr ) 
     return Half_Dep_Ptr; 


   --------------------------------------------
   -- A set of half dependencies
   --------------------------------------------


   type Half_Dep_Descriptor is 
      record 
         Task_Name       : Fixed_String;  
         Depend_Name     : Fixed_String;  
         Depend_Type     : Dependency_Type;  
         Depend_Position : Half_Dep_Type;  
      end record; 

   package Half_Dep_Set is
      new Sets (
      Max_Element     => Config.Max_Tasks,    
      File_Descriptor => Half_Dep_Descriptor, 
      Element         => Half_Dep_Ptr,        
      Free            => Free,                
      Copy            => Copy,                
      Put             => Put);
   use Half_Dep_Set;

   type Tasks_Dependencies is new Half_Dep_Set.Set with private; 

   subtype Tasks_Dependencies_Range is Half_Dep_Set.Element_Range;
   subtype Tasks_Dependencies_Iterator is Half_Dep_Set.Iterator;



   -- Some procedures/functions that must be
   -- overidden since a tasks_dependencies
   -- contains more than a set
   --

   procedure Export_Xml (
         My_Dependencies : in     Tasks_Dependencies; 
         Into            : in out File_Type           ); 


   procedure Read_From_File (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         My_Messages     : in     Messages_Set;       
         My_Buffers      : in     Buffers_Set;        
         File_Name       : in     Unbounded_String    ); 


   function Is_Empty (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Boolean; 


   procedure Reset (
         My_Dependencies : in out Tasks_Dependencies;         
         Free_Object     : in     Boolean            := False ); 


   procedure Duplicate (
         Src  : in     Tasks_Dependencies; 
         Dest : in out Tasks_Dependencies  ); 


   procedure Put (
         My_Dependencies : in     Tasks_Dependencies ); 


   procedure Fill_Descriptor (
         My_Dependencies : in     Tasks_Dependencies;  
         Desc            :    out Half_Dep_Descriptor; 
         A_Dependency    : in     Half_Dep_Ptr         ); 




   ---------------------------------------
   --  Services below allow you to delete
   --  dependencies.
   ---------------------------------------



   -- Delete ALL dependencies (double and half) of the
   -- "a_processor" processor
   --
   procedure Delete_Processors_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         A_Processor     : in     Unbounded_String    ); 



   -- Delete ALL dependencies (double and half) :
   --  1) of a task if my_dep_type is  "precedence_dependency" and my_dep_name is 
   --      a task name
   --  2) of a message if my_dep_type is  "message_dependency" and my_dep_name is 
   --      a message name
   --  3) of a buffer if my_dep_type is  "buffer_dependency" and my_dep_name is 
   --      a buffer name
   --
   procedure Delete_All_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Dep_Type     : in     Dependency_Type;    
         My_Dep_Name     : in     Unbounded_String    ); 



   -- Remove one or several double dependencies
   -- The two first methods remove ALL dependencies between two tasks
   -- and the two others remove only ONE dependency between two tasks
   --
   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr            ); 
   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String    ); 
   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr;           
         Dep_Type        : in     Dependency_Type;    
         Name            : in     Unbounded_String    ); 
   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String;   
         Dep_Type        : in     Dependency_Type;    
         Name            : in     Unbounded_String    ); 


   -- Remove a half dependency between a task and
   --   a message or a buffer
   --
   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         A_Task          : in     Task_Ptr;           
         A_Dep           : in     Unbounded_String;   
         My_Dep_Type     : in     Dependency_Type;    
         A_Type          : in     Half_Dep_Type       ); 
   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         A_Task          : in     Unbounded_String;   
         A_Dep           : in     Unbounded_String;   
         My_Dep_Type     : in     Dependency_Type;    
         A_Type          : in     Half_Dep_Type       ); 




   ---------------------------------------
   --  Services below allow to add
   --  dependencies.
   ---------------------------------------




   -- Build dependencies from the definition
   -- of a buffer
   --
   procedure Build_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         A_Buffer        : in     Buffer_Ptr          ); 




   -- Add a half dependency between a task and 
   --  a task, a message or a buffer
   --
   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         A_Task          : in     Task_Ptr;           
         A_Dep           : in     Dependency_Ptr;     
         A_Type          : in     Half_Dep_Type       ); 
   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         A_Task          : in     Unbounded_String;   
         A_Dep           : in     Dependency_Ptr;     
         A_Type          : in     Half_Dep_Type       ); 



   -- Add a precedency constraint between two tasks
   --
   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task_Name  : in     Unbounded_String;   
         To_Task_Name    : in     Unbounded_String    ); 
   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task_Name  : in     Task_Ptr;           
         To_Task_Name    : in     Task_Ptr            ); 

   -- Add a buffer/message/predecency constaint between two tasks
   --
   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task_Name  : in     Task_Ptr;           
         To_Task_Name    : in     Task_Ptr;           
         My_Deps         : in     Dependency_Ptr      ); 
   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task_Name  : in     Unbounded_String;   
         To_Task_Name    : in     Unbounded_String;   
         My_Deps         : in     Dependency_Ptr      ); 



   ---------------------------------------
   --  Services below allow to test
   --  existing dependencies.
   ---------------------------------------

   function Get_A_Successor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Task_Ptr; 

   function Get_A_Predecessor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Task_Ptr; 

   function Get_Successors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Tasks_Set; 

   function Get_Predecessors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Tasks_Set; 

   function Get_Successors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Tasks_Set; 

   function Get_Predecessors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Tasks_Set; 

   function No_Dependency (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Boolean; 


   function Has_Predecessor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Boolean; 

   function Has_Successor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Boolean; 

   function Has_Predecessor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Boolean; 

   function Has_Successor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Boolean; 

   function Get_Dependencies (
         My_Dependencies : in     Tasks_Dependencies; 
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr            ) 
     return Depends_Set; 

   function Get_Dependencies (
         My_Dependencies : in     Tasks_Dependencies; 
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String    ) 
     return Depends_Set; 

   function Get_Precedence_Dependency (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Tasks_Set; 


   -- Get a task whitout predecessors
   --
   function Get_A_Root_Task (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Task_Ptr; 


   -- Get a task whitout successors
   -- 
   function Get_A_Leaf_Task (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Task_Ptr; 


   -- Get all tasks whitout predecessors
   --
   function Get_Root_Tasks (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Tasks_Set; 


   -- Get all tasks whitout successors
   --
   function Get_Leaf_Tasks (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Tasks_Set; 


   -- Check is a task dependencies is acyclic or not
   --
   function Is_Acyclic (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Boolean; 



private

   Root_Index : constant := Config.Max_Tasks + 1;  
   type Dependencies_Range is new Natural range 0 .. Root_Index; 

   type Tasks_Table is array (Dependencies_Range) of Task_Ptr; 
   type Dependencies_Table is array (Dependencies_Range, Dependencies_Range) of Depends_Set; 


   type Tasks_Dependencies is new Half_Dep_Set.Set with 
      record 
         Dependencies    : Dependencies_Table;  
         Task_Position   : Tasks_Table;  
         Number_Of_Tasks : Dependencies_Range := 0;  
      end record; 


end Task_Dependencies;




