------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------






with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;
with Unbounded_Strings;
use Unbounded_Strings;

with Translate;
use Translate;

with Tasks;
use Tasks;

with Time_Units;
use  Time_Units;
use  Time_Units.A_Time_Unit;

with Parameters;
use Parameters;

with Natural_Util;
use Natural_Util;
with Double_Util;
use Double_Util;



package body Task_Set is



   procedure Fill_Descriptor (
         My_Tasks : in     Tasks_Set;       
         Desc     :    out Task_Descriptor; 
         A_Task   : in     Task_Ptr         ) is 

   begin

      for I in 0..A_Task.Param.Nb_Entries-1 loop
         Desc.Param(I).Name:=To_Bounded_String(To_String(A_Task.Param.
               Entries(I).Name));
         Desc.Param(I).Ptype:=A_Task.Param.Entries(I).Ptype;
         if Desc.Param(I).Ptype = Boolean_Parameter then
            Desc.Param(I).Boolean_Value:=A_Task.Param.Entries(I).
               Boolean_Value;
         else
            if Desc.Param(I).Ptype = Integer_Parameter then
               Desc.Param(I).Integer_Value:=A_Task.Param.Entries(I).
                  Integer_Value;
            else
               if Desc.Param(I).Ptype = Double_Parameter then
                  Desc.Param(I).Double_Value:=A_Task.Param.Entries(I).
                     Double_Value;
               else
                  Desc.Param(I).String_Value:=To_Bounded_String(To_String(
                        A_Task.Param.Entries(I).String_Value));
               end if;
            end if;
         end if;
      end loop;

      Desc.Offset:=A_Task.Offset;
      Desc.Name := To_Bounded_String(To_String(A_Task.Name));
      Desc.Location:=To_Bounded_String(To_String(A_Task.Location));
      Desc.Capacity := A_Task.Capacity;
      Desc.Start_Time := A_Task.Start_Time;
      Desc.Priority:= A_Task.Priority;
      Desc.Policy:= A_Task.Policy;
      Desc.Task_Type:= A_Task.Task_Type;
      Desc.Deadline := A_Task.Deadline;
      Desc.Blocking_Time:= A_Task.Blocking_Time;


      case A_Task.Task_Type is
         when Periodic_Type =>
            Desc.Period := Periodic_Task_Ptr(A_Task).Period;
            Desc.Jitter := Periodic_Task_Ptr(A_Task).Jitter;
            Desc.Seed := 0;
         when Poisson_Type  =>
            Desc.Period := Poisson_Task_Ptr(A_Task).Period;
            Desc.Jitter := Poisson_Task_Ptr(A_Task).Jitter;
            Desc.Seed := Poisson_Task_Ptr(A_Task).Seed;
         when Parametric_Type  =>
            Desc.Parametric_File_Name:=
               To_Bounded_String(To_String(Parametric_Task_Ptr(A_Task).
                  File_Name));
            Desc.Period:=0;
            Desc.Jitter:=0;
            Desc.Seed := 0;
         when others =>
            Desc.Period:=0;
            Desc.Jitter:=0;
            Desc.Seed := 0;
      end case;

   end Fill_Descriptor;


   procedure Extract_Descriptor (
         My_Tasks : in     Tasks_Set;       
         Desc     : in     Task_Descriptor; 
         A_Task   :    out Task_Ptr         ) is 

      New_Periodic_Task   : Periodic_Task_Ptr;  
      New_Aperiodic_Task  : Aperiodic_Task_Ptr;  
      New_Poisson_Task    : Poisson_Task_Ptr;  
      New_Parametric_Task : Parametric_Task_Ptr;  

   begin

      case Desc.Task_Type is
         when Periodic_Type   =>
            New_Periodic_Task:=new Periodic_Task;
            New_Periodic_Task.Jitter:=Desc.Jitter;
            New_Periodic_Task.Period:=Desc.Period;
            A_Task:=Task_Ptr(New_Periodic_Task);
         when Aperiodic_Type  =>
            New_Aperiodic_Task:=new Aperiodic_Task;
            A_Task:=Task_Ptr(New_Aperiodic_Task);
         when Poisson_Type    =>
            New_Poisson_Task:=new Poisson_Task;
            New_Poisson_Task.Jitter:=Desc.Jitter;
            New_Poisson_Task.Period:=Desc.Period;
            New_Poisson_Task.Seed:=Desc.Seed;
            A_Task:=Task_Ptr(New_Poisson_Task);
         when Parametric_Type =>
            New_Parametric_Task:=new Parametric_Task;
            A_Task:=Task_Ptr(New_Parametric_Task);
            Parametric_Task_Ptr(A_Task).File_Name:=
               To_Unbounded_String(To_String(Desc.Parametric_File_Name));
         when others =>
            null;
      end case;

      A_Task.Offset:=Desc.Offset;
      A_Task.Deadline:=Desc.Deadline;
      A_Task.Name:=To_Unbounded_String(To_String(Desc.Name));
      A_Task.Capacity:=Desc.Capacity;
      A_Task.Start_Time:=Desc.Start_Time;
      A_Task.Location:=To_Unbounded_String(To_String(Desc.Location));
      A_Task.Priority:=Desc.Priority;
      A_Task.Policy:=Desc.Policy;
      A_Task.Task_Type:=Desc.Task_Type;
      A_Task.Blocking_Time:=Desc.Blocking_Time;

      A_Task.Param.Nb_Entries:=0;
      for I in Parameters_Range loop

         if Desc.Param(I).Name= To_Bounded_String("")
               then
            exit;
         end if;

         A_Task.Param.Nb_Entries:=A_Task.Param.Nb_Entries+1;
         A_Task.Param.Entries(I) := new Parameter(Desc.Param(I).Ptype);
         A_Task.Param.Entries(I).Name:=To_Unbounded_String(To_String(Desc.
               Param(I).Name));
         if Desc.Param(I).Ptype = Boolean_Parameter then
            A_Task.Param.Entries(I).Boolean_Value:=Desc.Param(I).
               Boolean_Value;
         else
            if Desc.Param(I).Ptype = Integer_Parameter then
               A_Task.Param.Entries(I).Integer_Value:=Desc.Param(I).
                  Integer_Value;
            else
               if Desc.Param(I).Ptype = Double_Parameter then
                  A_Task.Param.Entries(I).Double_Value:=Desc.Param(I).
                     Double_Value;
               else
                  A_Task.Param.Entries(I).String_Value:=
                     To_Unbounded_String(To_String(Desc.Param(I).
                        String_Value));
               end if;
            end if;
         end if;
      end loop;

   end Extract_Descriptor;

   procedure Add_Task (
         My_Tasks             : in out Tasks_Set;                        
         Name                 : in     Unbounded_String;                 
         Location             : in     Unbounded_String;                 
         Task_Type            : in     Tasks_Type;                       
         Start_Time           : in     Natural;                          
         Capacity             : in     Natural;                          
         Period               : in     Natural;                          
         Deadline             : in     Natural;                          
         Jitter               : in     Natural;                          
         Blocking_Time        : in     Natural;                          
         Priority             : in     Priority_Range;                   
         Policy               : in     Policies;                         
         Offset               :        Offsets_Table;                    
         Param                : in     Parameters_Table := No_Parameter; 
         Parametric_File_Name : in     Unbounded_String := Empty_String; 
         Seed_Value           : in     Integer          := 0             ) is 


      Dummy : Task_Ptr;  

   begin
      Add_Task(My_Tasks, Dummy, Name, Location, Task_Type, Start_Time,
         Capacity, Period, Deadline, Jitter, Blocking_Time,
         Priority, Policy, Offset, Param, Parametric_File_Name,
         Seed_Value);
   end Add_Task;


   procedure Add_Task (
         My_Tasks             : in out Tasks_Set;                        
         A_Task               : in out Task_Ptr;                         
         Name                 : in     Unbounded_String;                 
         Location             : in     Unbounded_String;                 
         Task_Type            : in     Tasks_Type;                       
         Start_Time           : in     Natural;                          
         Capacity             : in     Natural;                          
         Period               : in     Natural;                          
         Deadline             : in     Natural;                          
         Jitter               : in     Natural;                          
         Blocking_Time        : in     Natural;                          
         Priority             : in     Priority_Range;                   
         Policy               : in     Policies;                         
         Offset               :        Offsets_Table;                    
         Param                : in     Parameters_Table := No_Parameter; 
         Parametric_File_Name : in     Unbounded_String := Empty_String; 
         Seed_Value           : in     Integer          := 0             ) is 



      My_Iterator : Iterator;  
      Desc        : Task_Descriptor;  


   begin

      if (Name = "")
            then
         raise Invalid_Parameter;
      end if;

      if (Location = "")
            then
         raise Invalid_Parameter;
      end if;

      if (Capacity = 0)
            then
         raise Invalid_Parameter;
      end if;

      if (Jitter > Deadline)
            then
         raise Invalid_Parameter;
      end if;


      if (Priority /= 0) and
            (Policy=Sched_Others)
            then
         raise Invalid_Parameter;
      end if;

      if (Priority = 0) and
            (Policy/=Sched_Others)
            then
         raise Invalid_Parameter;
      end if;


      if(Get_Number_Of_Elements(My_Tasks) > 0)
            then

         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Tasks, A_Task, My_Iterator);
            if (Name = A_Task.Name)
                  then
               raise Task_Already_Defined;
            end if;

            exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

            Next_Element(My_Tasks, My_Iterator);
         end loop;
      end if;


      Desc.Offset:=Offset;
      Desc.Name:=To_Bounded_String(To_String(Name));
      Desc.Location:=To_Bounded_String(To_String(Location));
      Desc.Task_Type:=Task_Type;
      Desc.Start_Time:=Start_Time;
      Desc.Capacity:=Capacity;
      Desc.Period:=Period;
      Desc.Deadline:=Deadline;
      Desc.Jitter:=Jitter;
      Desc.Priority:=Priority;
      Desc.Policy:=Policy;
      Desc.Blocking_Time:=Blocking_Time;
      Desc.Parametric_File_Name:=To_Bounded_String(
         To_String(Parametric_File_Name));
      Desc.Seed := Seed_Value;


      for I in 0..Param.Nb_Entries-1 loop
         Desc.Param(I).Name:=To_Bounded_String(To_String(Param.Entries(I).
               Name));
         Desc.Param(I).Ptype:=Param.Entries(I).Ptype;
         if Desc.Param(I).Ptype = Boolean_Parameter then
            Desc.Param(I).Boolean_Value:=Param.Entries(I).Boolean_Value;
         else
            if Desc.Param(I).Ptype = Integer_Parameter then
               Desc.Param(I).Integer_Value:=Param.Entries(I).
                  Integer_Value;
            else
               if Desc.Param(I).Ptype = Double_Parameter then
                  Desc.Param(I).Double_Value:=Param.Entries(I).
                     Double_Value;
               else
                  Desc.Param(I).String_Value:=To_Bounded_String(To_String(
                        Param.Entries(I).String_Value));
               end if;
            end if;
         end if;
      end loop;

      Extract_Descriptor(My_Tasks, Desc, A_Task);


      Add(My_Tasks,A_Task);

   end Add_Task;



   function Search_Task (
         My_Tasks : in     Tasks_Set;       
         Name     : in     Unbounded_String ) 
     return Task_Ptr is 

      My_Iterator : Iterator;  
      A_Task      : Task_Ptr;  
      Result      : Task_Ptr;  

      Found : Boolean := False;  

   begin

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if(A_Task.Name = Name)
               then
            Found:=True;
            Result:=A_Task;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      if not Found
            then
         raise Task_Not_Found;
      end if;

      return Result;
   end Search_Task;



   function Increasing_Priority (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Op1.Priority<= Op2.Priority);
   end Increasing_Priority;

   function Decreasing_Priority (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Op1.Priority>= Op2.Priority);
   end Decreasing_Priority;


   function Increasing_Period (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      case Op1.Task_Type is
         when Periodic_Type =>
            return (Periodic_Task_Ptr(Op1).Period <=
               Periodic_Task_Ptr(Op2).Period);
         when Poisson_Type =>
            return (Poisson_Task_Ptr(Op1).Period <=
               Poisson_Task_Ptr(Op2).Period);
         when others =>
            return True;
      end case;

   end Increasing_Period;


   function Decreasing_Period (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      case Op1.Task_Type is
         when Periodic_Type =>
            return (Periodic_Task_Ptr(Op1).Period >= Periodic_Task_Ptr(
                  Op2).
               Period);
         when Poisson_Type =>
            return (Poisson_Task_Ptr(Op1).Period >= Poisson_Task_Ptr(Op2).
               Period);
         when others =>
            return True;
      end case;

   end Decreasing_Period;


   function Increasing_Deadline (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Op1.Deadline <= Op2.Deadline);
   end Increasing_Deadline;


   function Decreasing_Deadline (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Op1.Deadline >= Op2.Deadline);
   end Decreasing_Deadline;

   function Increasing_Name (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Op1.Name<=Op2.Name);
   end Increasing_Name;



   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Unbounded_String is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Location)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      return A_Task.Location;

   end Get;


   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Priority_Range is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Priority)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      return A_Task.Priority;

   end Get;


   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Natural is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if ( (Param_Name /= Start_Time) and
            (Param_Name /= Blocking_Time) and
            (Param_Name /= Capacity) and
            (Param_Name /= Period) and
            (Param_Name /= Deadline) and
            (Param_Name /= Jitter) )
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      if (Param_Name = Start_Time)
            then
         return A_Task.Start_Time;
      else
         if (Param_Name = Deadline)
               then
            return A_Task.Deadline;
         else
            if (Param_Name = Period)
                  then
               return Periodic_Task_Ptr(A_Task).Period;
            else
               if (Param_Name = Jitter)
                     then
                  return Periodic_Task_Ptr(A_Task).Jitter;
               else
                  if (Param_Name = Capacity)
                        then
                     return A_Task.Capacity;
                  else
                     return A_Task.Blocking_Time;
                  end if;
               end if;
            end if;
         end if;
      end if;

   end Get;


   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Unbounded_String  ) is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Location)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            A_Task.Location:=Param_Value;
            exit;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

   end Set;


   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Priority_Range    ) is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Priority)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            A_Task.Priority:=Param_Value;
            exit;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;
   end Set;


   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Natural           ) is 
      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if ( (Param_Name /= Start_Time) and
            (Param_Name /= Blocking_Time) and
            (Param_Name /= Capacity) and
            (Param_Name /= Period) and
            (Param_Name /= Deadline) and
            (Param_Name /= Jitter) )
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            if (Param_Name = Start_Time)
                  then
               A_Task.Start_Time:=Param_Value;
               exit;
            end if;
            if (Param_Name = Deadline)
                  then
               A_Task.Deadline:=Param_Value;
            end if;
            if (Param_Name = Period and A_Task.Task_Type = Periodic_Type)
                  then
               Periodic_Task_Ptr(A_Task).Period:=Param_Value;
               exit;
            end if;
            if (Param_Name = Jitter and A_Task.Task_Type = Periodic_Type)
                  then
               Periodic_Task_Ptr(A_Task).Jitter:=Param_Value;
               exit;
            end if;
            if (Param_Name = Period and A_Task.Task_Type = Poisson_Type)
                  then
               Poisson_Task_Ptr(A_Task).Period:=Param_Value;
               exit;
            end if;
            if (Param_Name = Jitter and A_Task.Task_Type = Poisson_Type)
                  then
               Poisson_Task_Ptr(A_Task).Jitter:=Param_Value;
               exit;
            end if;
            if (Param_Name = Capacity)
                  then
               A_Task.Capacity:=Param_Value;
               exit;
            end if;
            if (Param_Name = Blocking_Time)
                  then
               A_Task.Blocking_Time:=Param_Value;
               exit;
            end if;
         end if;
         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;
   end Set;


   function Get_Number_Of_Task_From_Processor (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Tasks_Range is 

      Number      : Tasks_Range := 0;  
      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Location  = Processor_Name)
               then
            Number:=Number+1;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      return Number;
   end Get_Number_Of_Task_From_Processor;


   procedure Periodic_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) is 

      Iterator1 : Tasks_Iterator;  
      Task1     : Task_Ptr;  

   begin

      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, Task1, Iterator1);


         if (Task1.Location = Processor_Name)
               or (Processor_Name = Empty_String )
               then
            if  (Task1.Task_Type /= Periodic_Type)
                  then
               raise Task_Must_Be_Periodic;
            end if;
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;

   end Periodic_Control;


   procedure Priority_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) is 


      Iterator1 : Tasks_Iterator;  
      Task1     : Task_Ptr;  
      Iterator2 : Tasks_Iterator;  
      Task2     : Task_Ptr;  

   begin
      -- No duplicate priority
      --

      Reset_Iterator(Iterator1);


      loop
         Current_Element(My_Tasks, Task1, Iterator1);


         if (Task1.Location = Processor_Name)
               or (Processor_Name = Empty_String )
               then
            Reset_Iterator(Iterator2);

            loop
               Current_Element(My_Tasks, Task2, Iterator2);


               if (Task2.Location = Task1.Location)
                     and (Task1.Name /= Task2.Name)
                     and (Task1.Priority = Task2.Priority)
                     then
                  raise Priority_Error;
               end if;

               exit when Is_Last_Element(My_Tasks, Iterator2) = True;
               Next_Element(My_Tasks, Iterator2);

            end loop;


         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;
   end Priority_Control;


   -- Check start time
   --
   procedure Start_Time_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) is 

      Iterator1 : Tasks_Iterator;  
      Task1     : Task_Ptr;  

   begin

      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, Task1, Iterator1);


         if ( (Task1.Location = Processor_Name)
               or (Processor_Name = Empty_String )  )
               and (Task1.Start_Time /= 0)
               then
            raise Start_Time_Error;
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;
   end Start_Time_Control;


   -- Check offset
   --
   procedure Offset_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) is 

      Iterator1 : Tasks_Iterator;  
      Task1     : Task_Ptr;  

   begin

      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, Task1, Iterator1);


         if (Task1.Location = Processor_Name)
               or (Processor_Name = Empty_String )
               then
            for I in 0..Task1.Offset.Nb_Entries-1 loop
               if (Task1.Offset.Entries(I).Activation /= 0)
                     then
                  raise Offset_Error;
               end if;
            end loop;
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;
   end Offset_Control;


   procedure Have_Deadlines_Equal_Than_Periods_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if ( not ((A_Task.Task_Type = Periodic_Type) or
                  (A_Task.Task_Type = Poisson_Type)) and
               (A_Task.Location = Processor_Name) )
               then
            raise Task_Must_Be_Periodic;
         end if;

         if   (A_Task.Location = Processor_Name)
               or (Processor_Name = Empty_String )
               then
            if A_Task.Task_Type = Periodic_Type then
               if (A_Task.Deadline /=
                     Periodic_Task_Ptr(A_Task).Period)
                     then
                  raise Task_Must_Have_Period_Equal_To_Deadline;
               end if;
            end if;

            if A_Task.Task_Type = Poisson_Type then  -- check this
               if (A_Task.Deadline /=
                     Poisson_Task_Ptr(A_Task).Period)
                     then
                  raise Task_Must_Have_Period_Equal_To_Deadline;
               end if;
            end if;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

   end Have_Deadlines_Equal_Than_Periods_Control ;


   function Deadline_Inferior_To_Period (
         My_Tasks : in     Tasks_Set ) 
     return Boolean is 

      My_Iterator : Iterator;  
      A_Task      : Task_Ptr;  
      Result      : Boolean  := True;  

   begin
      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if A_Task.Task_Type = Periodic_Type then
            if(Periodic_Task_Ptr(A_Task).Period <  A_Task.Deadline)
                  then
               Result:=False;
            end if;
         end if;

         if A_Task.Task_Type = Poisson_Type then -- check this
            if(Poisson_Task_Ptr(A_Task).Period <  A_Task.Deadline)
                  then
               Result:=False;
            end if;
         end if;


         -- next task
         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;
         Next_Element(My_Tasks, My_Iterator);
      end loop;

      return Result;

   end Deadline_Inferior_To_Period;



   function Is_Harmonic (
         My_Tasks : in     Tasks_Set ) 
     return Boolean is 

      My_Iterator1,  
      My_Iterator2 : Iterator;  
      A_Task1,  
      A_Task2      : Task_Ptr;  
      Result       : Boolean  := True;  
      Tmp_Mod      : Natural;  

   begin
      -- First loop
      --
      Reset_Iterator(My_Iterator1);

      loop
         Current_Element(My_Tasks,A_Task1,My_Iterator1);

         -- second loop
         Reset_Iterator(My_Iterator2);

         loop
            Current_Element(My_Tasks,A_Task2,My_Iterator2);

            if A_Task1.Task_Type = Periodic_Type then
               if (Periodic_Task_Ptr(A_Task1).Period > Periodic_Task_Ptr(
                        A_Task2).Period)
                     then
                  Tmp_Mod := Periodic_Task_Ptr(A_Task1).Period mod
                     Periodic_Task_Ptr(A_Task2).Period;
               else
                  Tmp_Mod := Periodic_Task_Ptr(A_Task2).Period mod
                     Periodic_Task_Ptr(A_Task1).Period;
               end if;

               if ( A_Task1.Name /= A_Task2.Name and Tmp_Mod /= 0 )
                     then
                  Result:=False;
               end if;
            end if;


            if A_Task1.Task_Type = Poisson_Type then -- check this
               if (Poisson_Task_Ptr(A_Task1).Period > Poisson_Task_Ptr(
                        A_Task2).Period)
                     then
                  Tmp_Mod := Poisson_Task_Ptr(A_Task1).Period mod
                     Poisson_Task_Ptr(A_Task2).Period;
               else
                  Tmp_Mod := Poisson_Task_Ptr(A_Task2).Period mod
                     Poisson_Task_Ptr(A_Task1).Period;
               end if;

               if ( A_Task1.Name /= A_Task2.Name and Tmp_Mod /= 0 )
                     then
                  Result:=False;
               end if;
            end if;

            -- next task
            exit when Is_Last_Element(My_Tasks, My_Iterator2) = True;
            Next_Element(My_Tasks, My_Iterator2);

         end loop;

         -- next task
         exit when Is_Last_Element(My_Tasks, My_Iterator1) = True;
         Next_Element(My_Tasks, My_Iterator1);
      end loop;

      return Result;
   end Is_Harmonic;



   procedure Delete_Processor (
         My_Tasks    : in out Tasks_Set;       
         A_Processor : in     Unbounded_String ) is 

      Tmp         : Tasks_Set;  
      A_Task      : Task_Ptr;  
      My_Iterator : Tasks_Iterator;  

   begin


      if(Get_Number_Of_Elements(My_Tasks) > 0)
            then


         Reset_Iterator(My_Iterator);
         loop
            Current_Element(My_Tasks, A_Task, My_Iterator);

            if(A_Task.Location=A_Processor)
                  then
               Add(Tmp, A_Task);
            end if;

            exit when Is_Last_Element(My_Tasks, My_Iterator) = True;
            Next_Element(My_Tasks, My_Iterator);
         end loop;

         if not Is_Empty(Tmp) then
            Reset_Iterator(My_Iterator);
            loop
               Current_Element(Tmp, A_Task, My_Iterator);

               Delete(My_Tasks,A_Task);

               exit when Is_Last_Element(Tmp, My_Iterator) = True;
               Next_Element(Tmp, My_Iterator);
            end loop;

            Reset(Tmp,False);
         end if;

      end if;


   end Delete_Processor;

   function Scheduling_Period (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Natural is 

      Period : Natural := 1;  

      A_Task      : Task_Ptr;  
      My_Iterator : Tasks_Iterator;  

   begin


      -- check if tasks are periodics
      --      Periodic_Control(My_Tasks, Processor_Name);


      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks, A_Task, My_Iterator);

         if (A_Task.Location = Processor_Name and
               A_Task.Task_Type = Periodic_Type)
               then
            Period:= Natural_Util.Lcm(Period,
               Periodic_Task_Ptr(A_Task).Period);
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);
      end loop;


      return Period;

   end Scheduling_Period;


   function Processor_Utilization_Over_Deadline (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Double is 

      Utilization : Double := 0.0;  

      My_Iterator : Tasks_Iterator;  
      A_Task      : Task_Ptr;  

   begin

      -- check if tasks are periodics
      Periodic_Control(My_Tasks, Processor_Name);


      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Location = Processor_Name)
               then
            Utilization:=Utilization+(Double(A_Task.Capacity) /
               Double(A_Task.Deadline));
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;


      return Utilization;

   end Processor_Utilization_Over_Deadline;


   function Processor_Utilization_Over_Period (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Double is 

      Utilization : Double := 0.0;  

      My_Iterator : Tasks_Iterator;  
      A_Task      : Task_Ptr;  

   begin

      -- Check if tasks are periodics
      --
      Periodic_Control(My_Tasks, Processor_Name);



      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Location = Processor_Name)
               then
            if A_Task.Task_Type = Periodic_Type then
               Utilization:=Utilization+(Double(A_Task.Capacity) /
                  Double(Periodic_Task_Ptr(A_Task).Period));
            end if;
            if A_Task.Task_Type = Poisson_Type then -- check this
               Utilization:=Utilization+(Double(A_Task.Capacity) /
                  Double(Poisson_Task_Ptr(A_Task).Period));
            end if;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      return Utilization;

   end Processor_Utilization_Over_Period;


   procedure Put (
         My_Task_Occurence_Table : Task_Occurence_Table ) is 
   begin
      for I in My_Task_Occurence_Table'range loop
         Put(My_Task_Occurence_Table(I));
         Put("Arrival Date"  & My_Task_Occurence_Table(I).Arrival_Date'
            Img);
         Put("Response Time" & My_Task_Occurence_Table(I).Response_Time'
            Img);
      end loop;
   end Put ;


   procedure Export_Xml (
         My_Tasks : in     Tasks_Set; 
         Into     : in     File_Type  ) is 

      My_Iterator : Tasks_Iterator;  
      A_Task      : Task_Ptr;  

   begin


      if not Is_Empty(My_Tasks) then

         Put_Line(Into, "<tasks>");

         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Tasks,A_Task,My_Iterator);

            Put(Into,Ascii.Ht & "<task");
            if A_Task.X/=0
                  then
               Put(Into, " x=""" & A_Task.X'Img & """ ");
            end if;
            if A_Task.Y/=0
                  then
               Put(Into, "y=""" & A_Task.Y'Img & """ ");
            end if;

            case A_Task.Task_Type is
               when Periodic_Type =>
                  Put(Into, " task_type=""PERIODIC_TYPE"" ");
               when Poisson_Type  =>
                  Put(Into, " task_type=""POISSON_TYPE"" ");
               when Parametric_Type  =>
                  Put(Into, " task_type=""PARAMETRIC_TYPE"" ");
               when others =>
                  null;
            end case;
            Put_Line(Into, ">");


            if A_Task.Param.Nb_Entries>0 then
               Put_Line(Into,Ascii.Ht & Ascii.Ht & "<parameters>");
               for I in 0..A_Task.Param.Nb_Entries-1 loop

                  Put(Into,Ascii.Ht & Ascii.Ht & Ascii.Ht &"<parameter");
                  if A_Task.Param.Entries(I).Ptype = Boolean_Parameter then
                     Put(Into, " parameter_type=""boolean""> ");
                     Put(Into, " " & A_Task.Param.Entries(I).
                        Boolean_Value'
                        Img & " " );
                  else
                     if A_Task.Param.Entries(I).Ptype = Integer_Parameter then
                        Put(Into, " parameter_type=""integer""> ");
                        Put(Into, " " & A_Task.Param.Entries(I).
                           Integer_Value'Img & " " );
                     else
                        if A_Task.Param.Entries(I).Ptype =
                              Double_Parameter then
                           Put(Into, " parameter_type=""double""> ");
                           Put(Into, " " & To_String(Format(A_Task.Param.
                                    Entries(I).
                                    Double_Value)) & " " );
                        else
                           Put(Into, " parameter_type=""string""> ");
                           Put(Into, " " & To_String(A_Task.Param.
                                 Entries(
                                    I).
                                 String_Value) & " " );
                        end if;
                     end if;
                  end if;
                  Put(Into, To_String(" " & A_Task.Param. Entries(I).
                        Name) &
                     " ");
                  Put_Line(Into, "</parameter>");
               end loop;
               Put_Line(Into,Ascii.Ht & Ascii.Ht & "</parameters>");
            end if;


            if A_Task.Offset.Nb_Entries>0 then
               Put_Line(Into,Ascii.Ht & Ascii.Ht & "<offsets>");
               for I in 0..A_Task.Offset.Nb_Entries-1 loop
                  Put(Into,Ascii.Ht & Ascii.Ht & Ascii.Ht & "<offset>");
                  Put(Into, " " & A_Task.Offset.Entries(I).Activation'
                     Img &
                     " ");
                  Put(Into, " " & A_Task.Offset.Entries(I).Value'Img &
                     " ");
                  Put_Line(Into, "</offset>");
               end loop;
               Put_Line(Into,Ascii.Ht & Ascii.Ht & "</offsets>");
            end if;


            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<location>" &
               To_String(
                  A_Task.Location) & "</location>");
            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<name>" & To_String(
                  A_Task.Name) & "</name>");
            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<capacity>" & A_Task.
               Capacity'Img & "</capacity>");
            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<start_time>" & A_Task.
               Start_Time'Img & "</start_time>");
            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<policy>" & A_Task.
               Policy'Img & "</policy>");
            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<deadline>" & A_Task.
               Deadline'Img & "</deadline>");
            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<blocking_time>" &
               A_Task.
               Blocking_Time'Img & "</blocking_time>");
            Put_Line(Into,Ascii.Ht & Ascii.Ht & "<priority>" & A_Task.
               Priority'Img & "</priority>");


            case A_Task.Task_Type is
               when Periodic_Type =>
                  Put_Line(Into,Ascii.Ht & Ascii.Ht & "<period>" &
                     Periodic_Task_Ptr(A_Task).Period'Img & "</period>");
                  Put_Line(Into,Ascii.Ht & Ascii.Ht & "<jitter>" &
                     Periodic_Task_Ptr(A_Task).Jitter'Img & "</jitter>");
               when Poisson_Type  =>
                  Put_Line(Into,Ascii.Ht & Ascii.Ht & "<period>" &
                     Poisson_Task_Ptr(A_Task).Period'Img & "</period>");
                  Put_Line(Into,Ascii.Ht & Ascii.Ht & "<jitter>" &
                     Poisson_Task_Ptr(A_Task).Jitter'Img & "</jitter>");
                  Put_Line(Into,Ascii.Ht & Ascii.Ht & "<seed>" &
                     Poisson_Task_Ptr(A_Task).Seed'Img & "</seed>");
               when Parametric_Type  =>
                  Put_Line(Into, Ascii.Ht & Ascii.Ht &
                     "<parametric_file_name>"  & To_String(
                        Parametric_Task_Ptr(A_Task).File_Name) &
                     "</parametric_file_name>");
               when others =>
                  null;
            end case;



            Put_Line(Into,Ascii.Ht & "</task>");

            exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

            Next_Element(My_Tasks, My_Iterator);

         end loop;
         Put_Line(Into, "</tasks>");
      end if;

   end Export_Xml;


   procedure Response_Time_By_Simulation (
         My_Task : in     Periodic_Task_Ptr;   
         Sched   : in     Scheduling_Sequence; 
         Result  : in out Unbounded_String     ) is 

      Max_Occurence_Number : Natural := Max_Scheduling_Period;  

      Task_Occurence_Number            : Natural          := 0;  
      Awoken_Task_Occurence_Number     : Natural          := 0;  
      Running_Awoken_Task_Index        : Natural          := 0;  
      Next_My_Task_Occurence_Awake     : Natural          := 0;  
      My_Task_Worst_Case_Response_Time : Natural          := 0;  
      Running_Task                     : Task_Ptr;  
      Effective_End                    : Natural          := 0;  
      Deadline_Message                 : Unbounded_String := To_Unbounded_String
        ("");  
      Current_Date                     : Natural          := 0;  

      My_Task_Occurence_Table : Task_Occurence_Table (0 .. Max_Occurence_Number);
         







   begin
      -- initialize all occurences
      -- dates are taken from scheduling table
      for J in 0..Sched.Scheduled.Nb_Entries-1 loop

         if Sched.Scheduled.Entries(J).Item.Etype = Activation then
            if Sched.Scheduled.Entries(J).Item.Activated_Task.Name =
                  My_Task.Name then
               Initialize(My_Task_Occurence_Table(Task_Occurence_Number),
                  Periodic_Task_Ptr(My_Task),
                  Sched.Scheduled.Entries(J).Data ) ;
               Task_Occurence_Number := Task_Occurence_Number + 1;
            end if;
         end if;
      end loop ;


      for T in 0 .. Sched.Scheduled.Nb_Entries-1 loop

         -- the current date
         Current_Date := Sched.Scheduled.Entries(T).Data ;

         -- the current task running is the running occurence
         case Sched.Scheduled.Entries(T).Item.Etype is
            when Running =>
               Running_Task := Sched.Scheduled.Entries(T).Item.
                  Running_Task;
               --               Put();

               -- New occurence of My_task
               if ( Current_Date = My_Task_Occurence_Table(
                        Next_My_Task_Occurence_Awake).Arrival_Date ) then

                  -- one more task is awoken
                  Awoken_Task_Occurence_Number :=
                     Awoken_Task_Occurence_Number + 1 ;

                  -- activate the occurence
                  My_Task_Occurence_Table(Next_My_Task_Occurence_Awake).
                     Is_Active := True ;

                  -- if there is not active occurence, run the new occurence
                  if ( Awoken_Task_Occurence_Number = 0 ) then
                     Running_Awoken_Task_Index :=
                        Next_My_Task_Occurence_Awake ;
                  end if ;

                  -- the next occurence to awake
                  Next_My_Task_Occurence_Awake :=
                     Next_My_Task_Occurence_Awake + 1 ;

               end if ;

               -- is there any activated occurences
               if ( Awoken_Task_Occurence_Number > 0 ) then

                  -- increment response time for all active occurences
                  -- for I_Occurence in Running_Awoken_Task_Index ..
                  -- (Running_Awoken_Task_Index +
                  -- Awoken_Task_Occurence_Number - 1) loop

                  for I_Occurence in 0 .. Max_Occurence_Number loop
                     if ( My_Task_Occurence_Table(I_Occurence).Is_Active )
                           then
                        My_Task_Occurence_Table(I_Occurence).
                           Response_Time :=
                           My_Task_Occurence_Table(I_Occurence).
                           Response_Time + 1 ;
                     end if ;
                  end loop ;

                  if ( Running_Task.Name = My_Task.Name ) then

                     My_Task_Occurence_Table(Running_Awoken_Task_Index).
                        Residual_Capacity :=
                        My_Task_Occurence_Table(Running_Awoken_Task_Index).
                        Residual_Capacity - 1 ;

                     -- is The Current Occurence finished
                     if (My_Task_Occurence_Table(
                              Running_Awoken_Task_Index).
                           Residual_Capacity = 0) then
                        Effective_End := Natural(Current_Date) + 1 ;

                        -- Check deadline missed
                        if ( My_Task_Occurence_Table(
                                 Running_Awoken_Task_Index).Response_Time
                              > My_Task.Deadline ) then
                           Deadline_Message := Deadline_Message
                              & Lb_Check_Deadline(Current_Language)
                              & My_Task_Occurence_Table(
                              Running_Awoken_Task_Index).
                              Absolute_Deadline'Img
                              & Lb_Completion_Time(Current_Language)
                              & Effective_End'Img
                              & To_Unbounded_String(")");

                        end if ;

                        -- check worst case response time
                        if (My_Task_Occurence_Table(
                                 Running_Awoken_Task_Index).Response_Time
                              > My_Task_Worst_Case_Response_Time) then
                           My_Task_Worst_Case_Response_Time :=
                              My_Task_Occurence_Table(
                              Running_Awoken_Task_Index).Response_Time ;
                        end if ;

                        -- desactivate the current occurence
                        My_Task_Occurence_Table(Running_Awoken_Task_Index).
                           Is_Active := False ;

                        -- give hand to the next awoken occurence
                        Running_Awoken_Task_Index :=
                           Running_Awoken_Task_Index + 1 ;

                        -- one less Awoken Occurence
                        Awoken_Task_Occurence_Number :=
                           Awoken_Task_Occurence_Number - 1 ;

                     end if ;

                  end if ;

               end if ;

            when Activation =>
               Running_Task := Sched.Scheduled.Entries(T).
                  Item.Activated_Task;
            when Allocate_Resource =>
               null;
            when Release_Resource =>
               null;

         end case;


      end loop ;

      Result := Result
         & To_Unbounded_String("    ")
         & To_String(My_Task.Name)
         & To_Unbounded_String(" =>")
         & Integer'Image(Integer(My_Task_Worst_Case_Response_Time))
         & Deadline_Message
         & Unbounded_Lf ;

   end Response_Time_By_Simulation;


   procedure Response_Time_By_Simulation (
         My_Task : in     Aperiodic_Task_Ptr;  
         Sched   : in     Scheduling_Sequence; 
         Result  : in out Unbounded_String     ) is 

      My_Task_Occurence : Task_Occurence;  
      Running_Task      : Task_Ptr;  
      Effective_End     : Natural          := 0;  
      Deadline_Message  : Unbounded_String := To_Unbounded_String (""); 
        

      Current_Date : Natural := 0;  
   begin

      Initialize(My_Task_Occurence , Aperiodic_Task_Ptr(My_Task) ) ;

      for T in 0 .. Sched.Scheduled.Nb_Entries-1 loop
         -- the current date
         Current_Date := Sched.Scheduled.Entries(T).Data ;

         -- the current task running is the running occurence
         case Sched.Scheduled.Entries(T).Item.Etype is
            when Running =>
               Running_Task := Sched.Scheduled.Entries(T).
                  Item.Running_Task;

               -- if the occurence of My_task starts
               if ( Current_Date = My_Task_Occurence.Arrival_Date ) then
                  My_Task_Occurence.Is_Active := True ;
               end if ;

               if ( Running_Task.Name = My_Task.Name ) then

                  My_Task_Occurence.Residual_Capacity :=  My_Task_Occurence.
                     Residual_Capacity - 1 ;

                  -- is The Current Occurence finished
                  if (My_Task_Occurence.Residual_Capacity = 0) then

                     My_Task_Occurence.Is_Active := False ;

                     Effective_End := Current_Date + 1 ;

                     -- Check deadline missed
                     if ( My_Task_Occurence.Response_Time >
                           My_Task.Deadline ) then
                        Deadline_Message := Deadline_Message
                           & Lb_Check_Deadline(Current_Language)
                           & My_Task_Occurence.Absolute_Deadline'Img
                           & Lb_Completion_Time(Current_Language)
                           & Effective_End'Img
                           & To_Unbounded_String(")");
                     end if ;

                  end if;

               end if;


               -- increment response time if the occurence is Active
               if ( My_Task_Occurence.Is_Active ) then
                  My_Task_Occurence.Response_Time := My_Task_Occurence.
                     Response_Time + 1 ;
               end if ;
            -- if the current task running is the occurence of My_Task
            when Activation =>
               Running_Task := Sched.Scheduled.Entries(T).
                  Item.Activated_Task;
            when Allocate_Resource =>
               null;
            when Release_Resource =>
               null;
         end case;


      end loop;

      Result := Result
         & To_Unbounded_String("    ")
         & To_String(My_Task.Name)
         & To_Unbounded_String(" =>")
         & Integer'Image(Integer(My_Task_Occurence.Response_Time))
         & Deadline_Message
         & Unbounded_Lf ;

   end Response_Time_By_Simulation;


   procedure Response_Time_By_Simulation (
         My_Task : in     Poisson_Task_Ptr;    
         Sched   : in     Scheduling_Sequence; 
         Result  : in out Unbounded_String     ) is 

      Awoken_Task_Occurence_Number     : Natural          := 0;  
      Running_Awoken_Task_Index        : Natural          := 0;  
      Next_My_Task_Occurence_Awake     : Natural          := 0;  
      My_Task_Worst_Case_Response_Time : Natural          := 0;  
      Task_Occurence_Number            : Natural          := 0;  
      Running_Task                     : Task_Ptr;  
      Effective_End                    : Natural          := 0;  
      Deadline_Message                 : Unbounded_String := To_Unbounded_String
        ("");  
      Current_Date                     : Natural          := 0;  

      My_Task_Occurence_Table : Task_Occurence_Table (0 .. Max_Scheduling_Period);
         






   begin

      for J in 0..Sched.Scheduled.Nb_Entries-1 loop
         if Sched.Scheduled.Entries(J).Item.Etype = Activation then
            if Sched.Scheduled.Entries(J).Item.Activated_Task.Name =
                  My_Task.Name then
               Initialize(My_Task_Occurence_Table(Task_Occurence_Number),
                  Poisson_Task_Ptr(My_Task),
                  Sched.Scheduled.Entries(J).Data ) ;
               Task_Occurence_Number := Task_Occurence_Number + 1;
            end if;
         end if;
      end loop ;

      for T in 0 .. Sched.Scheduled.Nb_Entries-1 loop

         -- the current date
         Current_Date := Sched.Scheduled.Entries(T).Data ;

         -- the current task running is the running occurence
         case Sched.Scheduled.Entries(T).Item.Etype is
            when Running =>
               Running_Task := Sched.Scheduled.Entries(T).Item.
                  Running_Task ;

               -- New occurence of My_task
               if ( Current_Date = My_Task_Occurence_Table(
                        Next_My_Task_Occurence_Awake).Arrival_Date ) then

                  -- one more task is awoken
                  Awoken_Task_Occurence_Number :=
                     Awoken_Task_Occurence_Number + 1 ;

                  -- activate the occurence
                  My_Task_Occurence_Table(Next_My_Task_Occurence_Awake).
                     Is_Active := True ;

                  -- if there is not active occurence, run the new occurence
                  if ( Awoken_Task_Occurence_Number = 0 ) then
                     Running_Awoken_Task_Index :=
                        Next_My_Task_Occurence_Awake ;
                  end if ;

                  -- the next occurence to awake
                  Next_My_Task_Occurence_Awake :=
                     Next_My_Task_Occurence_Awake + 1 ;

               end if ;

            when Activation =>
               Running_Task := Sched.Scheduled.Entries(T).Item.
                  Activated_Task ;
            when Allocate_Resource =>
               null;
            when Release_Resource =>
               null;


         end case;

         -- is there any activated occurences
         if ( Awoken_Task_Occurence_Number > 0 ) then

            -- increment response time for all active occurences
            -- for I_Occurence in Running_Awoken_Task_Index .. (Running_Awoken_Task_Index + Awoken_Task_Occurence_Number - 1) loop
            for I_Occurence in 0 .. Max_Scheduling_Period loop
               if ( My_Task_Occurence_Table(I_Occurence).Is_Active ) then
                  My_Task_Occurence_Table(I_Occurence).Response_Time :=
                     My_Task_Occurence_Table(I_Occurence).
                     Response_Time + 1 ;
               end if ;
            end loop ;

            if ( Running_Task.Name = My_Task.Name ) then

               My_Task_Occurence_Table(Running_Awoken_Task_Index).
                  Residual_Capacity :=
                  My_Task_Occurence_Table(Running_Awoken_Task_Index).
                  Residual_Capacity - 1 ;

               -- is The Current Occurence finished
               if (My_Task_Occurence_Table(Running_Awoken_Task_Index).
                     Residual_Capacity = 0) then
                  Effective_End := Natural(Current_Date) + 1 ;

                  -- Check deadline missed
                  if ( My_Task_Occurence_Table(
                           Running_Awoken_Task_Index).Response_Time
                        > My_Task.Deadline ) then
                     Deadline_Message := Deadline_Message
                        & Lb_Check_Deadline(Current_Language)
                        & My_Task_Occurence_Table(
                        Running_Awoken_Task_Index).Absolute_Deadline'
                        Img
                        & Lb_Completion_Time(Current_Language)
                        & Effective_End'Img
                        & To_Unbounded_String(")");

                  end if ;

                  -- check worst case response time
                  if (My_Task_Occurence_Table(
                           Running_Awoken_Task_Index).Response_Time
                        > My_Task_Worst_Case_Response_Time) then
                     My_Task_Worst_Case_Response_Time :=
                        My_Task_Occurence_Table(
                        Running_Awoken_Task_Index).Response_Time ;
                  end if ;

                  -- desactivate the current occurence
                  My_Task_Occurence_Table(Running_Awoken_Task_Index).
                     Is_Active := False ;

                  -- give hand to the next awoken occurence
                  Running_Awoken_Task_Index :=
                     Running_Awoken_Task_Index + 1 ;

                  -- one less Awoken Occurence
                  Awoken_Task_Occurence_Number :=
                     Awoken_Task_Occurence_Number - 1 ;

               end if ;

            end if ;

         end if ;

      end loop ;

      Result := Result
         & To_Unbounded_String("    ")
         & To_String(My_Task.Name)
         & To_Unbounded_String(" =>")
         & Integer'Image(Integer(My_Task_Worst_Case_Response_Time))
         & Deadline_Message
         & Unbounded_Lf ;

   end Response_Time_By_Simulation;


   procedure Compute_Response_Time_Distribution (
         My_Task : in     Task_Ptr;            
         Sched   : in     Scheduling_Sequence; 
         Result  : in out Density_Table        ) is 


      Awoken_Task_Occurence_Number : Natural      := 0;  
      Task_Occurence_Number        : Natural      := 0;  
      Running_Awoken_Task_Index    : Natural      := 0;  
      Next_My_Task_Occurence_Awake : Natural      := 0;  
      Running_Task                 : Task_Ptr;  
      Effective_End                : Natural      := 0;  
      Current_Date                 : Natural      := 0;  
      Found                        : Boolean      := False;  
      A_Density                    : Density_Item;  
      Number_Of_Response_Time      : Double;  

      My_Task_Occurence_Table : Task_Occurence_Table (0 .. Max_Scheduling_Period);
         






   begin


      -- initialize all occurences
      case My_Task.Task_Type is
         when Periodic_Type =>

            -- initialize all occurences
            -- dates are taken from scheduling table
            for J in 0..Sched.Scheduled.Nb_Entries-1 loop

               if Sched.Scheduled.Entries(J).Item.Etype = Activation then
                  if Sched.Scheduled.Entries(J).Item.Activated_Task.Name =
                        My_Task.Name then
                     Initialize(My_Task_Occurence_Table(
                           Task_Occurence_Number),
                        Periodic_Task_Ptr(My_Task),
                        Sched.Scheduled.Entries(J).Data ) ;
                     Task_Occurence_Number := Task_Occurence_Number + 1;
                  end if;
               end if;
            end loop ;

         when Poisson_Type =>


            for J in 0..Sched.Scheduled.Nb_Entries-1 loop
               if Sched.Scheduled.Entries(J).Item.Etype = Activation then
                  if Sched.Scheduled.Entries(J).Item.Activated_Task.Name =
                        My_Task.Name then
                     Initialize(My_Task_Occurence_Table(
                           Task_Occurence_Number),
                        Poisson_Task_Ptr(My_Task),
                        Sched.Scheduled.Entries(J).Data ) ;
                     Task_Occurence_Number := Task_Occurence_Number + 1;
                  end if;
               end if;
            end loop ;


         when Aperiodic_Type =>

            Initialize(My_Task_Occurence_Table(Task_Occurence_Number),
               Aperiodic_Task_Ptr(My_Task) ) ;
            Task_Occurence_Number := Task_Occurence_Number + 1;

         when Parametric_Type =>


            raise Task_Model_Error;


      end case;





      for T in 0 .. Sched.Scheduled.Nb_Entries-1 loop

         if Sched.Scheduled.Entries(T).Item.Etype = Running then

            -- the current date
            Current_Date := Sched.Scheduled.Entries(T).Data ;

            -- the current task running is the running occurence
            Running_Task := Sched.Scheduled.Entries(T).Item.Running_Task ;
            -- New occurence of My_task
            if ( Current_Date = My_Task_Occurence_Table(
                     Next_My_Task_Occurence_Awake).Arrival_Date ) then

               -- one more task is awoken
               Awoken_Task_Occurence_Number :=
                  Awoken_Task_Occurence_Number + 1 ;

               -- activate the occurence
               My_Task_Occurence_Table(Next_My_Task_Occurence_Awake).
                  Is_Active := True ;

               -- if there is not active occurence, run the new occurence
               if ( Awoken_Task_Occurence_Number = 0 ) then
                  Running_Awoken_Task_Index :=
                     Next_My_Task_Occurence_Awake ;
               end if ;

               -- the next occurence to awake
               Next_My_Task_Occurence_Awake :=
                  Next_My_Task_Occurence_Awake + 1 ;

            end if ;

            -- is there any activated occurences
            if ( Awoken_Task_Occurence_Number > 0 ) then

               -- increment response time for all active occurences
               -- for I_Occurence in Running_Awoken_Task_Index .. (Running_Awoken_Task_Index + Awoken_Task_Occurence_Number - 1) loop
               for I_Occurence in 0 .. Max_Scheduling_Period loop
                  if ( My_Task_Occurence_Table(I_Occurence).Is_Active ) then
                     My_Task_Occurence_Table(I_Occurence).Response_Time :=
                        My_Task_Occurence_Table(I_Occurence).
                        Response_Time + 1 ;
                  end if ;
               end loop ;


               if ( Running_Task.Name = My_Task.Name ) then

                  My_Task_Occurence_Table(Running_Awoken_Task_Index).
                     Residual_Capacity :=
                     My_Task_Occurence_Table(Running_Awoken_Task_Index).
                     Residual_Capacity
                     - 1 ;

                  -- is The Current Occurence finished
                  if (My_Task_Occurence_Table(Running_Awoken_Task_Index).
                        Residual_Capacity = 0) then
                     Effective_End := Natural(Current_Date) + 1 ;


                     -- We have a response time. We look for the
                     --   density table its response time is already
                     -- known. If so, increment the frequency, if
                     -- not, add it to the table
                     --
                     Found:=False;
                     for K in 0..Result.Nb_Entries-1 loop
                        if Result.Entries(K).Response_Time=
                              My_Task_Occurence_Table(
                              Running_Awoken_Task_Index).Response_Time
                              then
                           Result.Entries(K).Probability:=Result.Entries(
                              K).Probability+1.0;
                           Found:=True;
                        end if;
                     end loop;
                     if not Found then
                        A_Density.Response_Time:=My_Task_Occurence_Table(
                           Running_Awoken_Task_Index).Response_Time;
                        A_Density.Probability:=1.0;
                        Add(Result, A_Density);
                     end if;

                     -- desactivate the current occurence
                     My_Task_Occurence_Table(
                        Running_Awoken_Task_Index).
                        Is_Active := False ;

                     -- give hand to the next awoken occurence
                     Running_Awoken_Task_Index :=
                        Running_Awoken_Task_Index + 1 ;

                     -- one less Awoken Occurence
                     Awoken_Task_Occurence_Number :=
                        Awoken_Task_Occurence_Number - 1 ;

                  end if ;

               end if ;

            end if ;

         end if;

      end loop ;


      -- Now, compute density : find number of activation done and
      -- divide all entries of the table by this number
      --
      Number_Of_Response_Time:=0.0;
      for K in 0..Result.Nb_Entries-1 loop
         Number_Of_Response_Time:=Number_Of_Response_Time+Result.Entries(
            K).Probability;
      end loop;
      for K in 0..Result.Nb_Entries-1 loop
         Result.Entries(K).Probability:=Result.Entries(K).Probability/
            Number_Of_Response_Time;
      end loop;


   end Compute_Response_Time_Distribution ;


   procedure Initialize (
         D : in out Density_Item ) is 
   begin
      D.Response_Time:=0;
      D.Probability:=0.0;
   end Initialize;


   procedure Put (
         D : in     Density_Item ) is 
   begin
      Put(D.Response_Time'Img & " " & D.Probability'Img);
   end Put;


   procedure Compute_Response_Time_Distribution (
         My_Tasks  : in     Tasks_Set;           
         Sched     : in     Scheduling_Sequence; 
         Processor :        Unbounded_String;    
         Result    : in out Densities_Table      ) is 


      A_Task          : Task_Ptr;  
      My_Iterator     : Iterator;  
      Working_Density : Density_Table;  

   begin

      Initialize(Result);

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (Processor=A_Task.Location)
               then
            if A_Task.Task_Type = Periodic_Type
                  then
               Initialize(Working_Density);
               Compute_Response_Time_Distribution(Task_Ptr(A_Task),
                  Sched,
                  Working_Density);
               Add(Result, A_Task, Working_Density);
            else
               if  A_Task.Task_Type = Poisson_Type -- check this
                     then
                  Initialize(Working_Density);
                  Compute_Response_Time_Distribution(Task_Ptr(A_Task),
                     Sched,
                     Working_Density);
                  Add(Result, A_Task, Working_Density);
               else
                  raise Task_Model_Error;
               end if;
            end if;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

   end Compute_Response_Time_Distribution;


   procedure Response_Time_By_Simulation (
         My_Tasks  : in     Tasks_Set;           
         Sched     : in     Scheduling_Sequence; 
         Processor :        Unbounded_String;    
         Result    : in out Unbounded_String     ) is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      Result:=Lb_Task_Response_Time_From_Simulation(
         Current_Language) &
         Unbounded_Lf;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (Processor=A_Task.Location) then
            case A_Task.Task_Type is
               when Periodic_Type =>
                  Response_Time_By_Simulation(Periodic_Task_Ptr(A_Task),
                     Sched,
                     Result);
               when Aperiodic_Type =>
                  Response_Time_By_Simulation(Aperiodic_Task_Ptr(A_Task),
                     Sched,
                     Result);
               when Poisson_Type =>
                  Response_Time_By_Simulation(Poisson_Task_Ptr(A_Task),
                     Sched,
                     Result);
               when Parametric_Type =>
                  raise Constraint_Error;


            end case;
         end if;



         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

   end Response_Time_By_Simulation;



   procedure Have_Same_Period_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) is 

      Iterator1    : Tasks_Iterator;  
      Task1        : Task_Ptr;  
      Period_Value : Natural        := 0;  
      First        : Boolean        := True;  

   begin

      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, Task1, Iterator1);

         if  not ((Task1.Task_Type = Periodic_Type) or
               (Task1.Task_Type = Aperiodic_Type))
               then
            raise Task_Model_Error;
         end if;

         if (Task1.Location = Processor_Name)
               or (Processor_Name = Empty_String )
               then
            if First then
               First:=False;
               if Task1.Task_Type = Periodic_Type
                     then
                  Period_Value:=Periodic_Task_Ptr(Task1).Period;
               end if;
            else
               if Task1.Task_Type= Aperiodic_Type and
                     Period_Value /= 0
                     then
                  raise Task_Must_Have_The_Same_Period_Value;
               end if;
               if Task1.Task_Type= Periodic_Type and
                     Period_Value /= Periodic_Task_Ptr(Task1).Period
                     then
                  raise Task_Must_Have_The_Same_Period_Value;
               end if;
            end if;
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;

   end Have_Same_Period_Control;



end Task_Set;
