------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Fixed_Strings;
use Fixed_Strings;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;
with Unbounded_Strings;
use Unbounded_Strings;

with Config;
use Config;

with Tasks;
use Tasks;
with Parameters;
use Parameters;
use Parameters.Parameters_Table_Package;
with Task_Offsets;
use Task_Offsets;
use Task_Offsets.Offsets_Table_Package;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

with Scheduling_Sequences;
use Scheduling_Sequences;

with Text_Io;
use Text_Io;

with Sets;
with Tables;
with Indexed_Tables;


package Task_Set is



   ----------------------------------------------------------
   -- Types defined for I/O on task set
   ----------------------------------------------------------


   type Io_Parameter is 
      record 
         Name          : Fixed_String   := To_Bounded_String ("");  
         Ptype         : Parameter_Type;  
         Boolean_Value : Boolean;  
         Integer_Value : Integer;  
         Double_Value  : Double;  
         String_Value  : Fixed_String;  
      end record; 

   type Io_Parameters_Table is array (Parameters_Range) of Io_Parameter; 


   type Task_Descriptor is 
      record 
         Name                 : Fixed_String;  
         Location             : Fixed_String;  
         Capacity             : Natural;  
         Start_Time           : Natural;  
         Priority             : Priority_Range;  
         Task_Type            : Tasks_Type;  
         Policy               : Policies;  
         Period               : Natural;  
         Blocking_Time        : Natural;  
         Parametric_File_Name : Fixed_String;  
         Deadline             : Natural;  
         Jitter               : Natural;  
         Seed                 : Integer;  
         Offset               : Offsets_Table;  
         Param                : Io_Parameters_Table;  
      end record; 


   ----------------------------------------------------------
   -- Definition of a task set
   ----------------------------------------------------------


   package Generic_Task_Set is new Sets (
      Max_Element     => Config.Max_Tasks, 
      File_Descriptor => Task_Descriptor,  
      Element         => Task_Ptr,         
      Copy            => Copy,             
      Free            => Free,             
      Put             => Put);


   use Generic_Task_Set;

   type Tasks_Set is new Generic_Task_Set.Set with private; 


   subtype Tasks_Range is Generic_Task_Set.Element_Range;
   subtype Tasks_Iterator is   Generic_Task_Set.Iterator;



   ----------------------------------------------------------
   -- Exceptions defined in the task set package
   ----------------------------------------------------------


   -- Raised if periodic tasks are expected
   --
   Task_Must_Be_Periodic : exception;  

   -- Raised if a task has a wrong type
   --
   Task_Model_Error : exception;  

   -- Raised when a given task is not in a task set
   --
   Task_Not_Found : exception;  

   -- Raised when 2 tasks with the same name are added
   -- to a task set
   Task_Already_Defined : exception;  

   -- Raised when the name of a task parameter is not known
   --
   Invalid_Parameter : exception;  

   -- Raised when a treatement required that the processor
   -- utilization is less than 1 for a given task set
   --
   Processor_Utilization_Exceeded : exception;  

   -- Exceptions raised during control procedures on task
   -- set
   Task_Must_Have_Period_Equal_To_Deadline : exception;  
   Priority_Error                          : exception;  
   Start_Time_Error                        : exception;  
   Offset_Error                            : exception;  
   Task_Must_Have_The_Same_Period_Value    : exception;  


   ----------------------------------------------------------
   -- Data types and procedure used to compute a 
   --       response time density
   ----------------------------------------------------------
   type Density_Item is 
      record 
         Response_Time : Natural;  
         Probability   : Double;  
      end record; 


   procedure Initialize (
         D : in out Density_Item ); 
   procedure Put (
         D : in     Density_Item ); 


   package Density_Table_Package is new Tables(
      Density_Item, Max_Density_Size, Put);
   use Density_Table_Package;
   subtype Density_Table is Density_Table_Package.Table;
   subtype Density_Range is Density_Table_Package.Table_Range;


   package Densities_Table_Package is new Indexed_Tables(Density_Table,
      Task_Ptr, Max_Tasks, 0, Put, Initialize, Tasks.Put_Name);
   use  Densities_Table_Package;
   subtype Densities_Table is Densities_Table_Package.Indexed_Table;
   subtype Densities_Range is Densities_Table_Package.Indexed_Table_Range;



   ----------------------------------------------------------
   -- List of task parameters the user can set of get directly
   -- throught the task set
   ----------------------------------------------------------


   type Task_Parameters is 
         (Location,      
          Capacity,      
          Start_Time,    
          Priority,      
          Period,        
          Blocking_Time, 
          Deadline,      
          Jitter,        
          Seed); 


   ----------------------------------------------------------
   -- I/O operations on a task set
   ----------------------------------------------------------


   procedure Export_Xml (
         My_Tasks : in     Tasks_Set; 
         Into     : in     File_Type  ); 


   procedure Fill_Descriptor (
         My_Tasks : in     Tasks_Set;       
         Desc     :    out Task_Descriptor; 
         A_Task   : in     Task_Ptr         ); 


   procedure Extract_Descriptor (
         My_Tasks : in     Tasks_Set;       
         Desc     : in     Task_Descriptor; 
         A_Task   :    out Task_Ptr         ); 

   ----------------------------------------------------------
   -- Procedure to proceed modification on a task set
   ----------------------------------------------------------

   procedure Add_Task (
         My_Tasks             : in out Tasks_Set;                        
         Name                 : in     Unbounded_String;                 
         Location             : in     Unbounded_String;                 
         Task_Type            : in     Tasks_Type;                       
         Start_Time           : in     Natural;                          
         Capacity             : in     Natural;                          
         Period               : in     Natural;                          
         Deadline             : in     Natural;                          
         Jitter               : in     Natural;                          
         Blocking_Time        : in     Natural;                          
         Priority             : in     Priority_Range;                   
         Policy               : in     Policies;                         
         Offset               : in     Offsets_Table;                    
         Param                : in     Parameters_Table := No_Parameter; 
         Parametric_File_Name : in     Unbounded_String := Empty_String; 
         Seed_Value           : in     Integer          := 0             ); 


   procedure Add_Task (
         My_Tasks             : in out Tasks_Set;                        
         A_Task               : in out Task_Ptr;                         
         Name                 : in     Unbounded_String;                 
         Location             : in     Unbounded_String;                 
         Task_Type            : in     Tasks_Type;                       
         Start_Time           : in     Natural;                          
         Capacity             : in     Natural;                          
         Period               : in     Natural;                          
         Deadline             : in     Natural;                          
         Jitter               : in     Natural;                          
         Blocking_Time        : in     Natural;                          
         Priority             : in     Priority_Range;                   
         Policy               : in     Policies;                         
         Offset               : in     Offsets_Table;                    
         Param                : in     Parameters_Table := No_Parameter; 
         Parametric_File_Name : in     Unbounded_String := Empty_String; 
         Seed_Value           : in     Integer          := 0             ); 


   -- Retrieve task hosted on the processor "A_Processor"
   -- from the task set
   --
   procedure Delete_Processor (
         My_Tasks    : in out Tasks_Set;       
         A_Processor : in     Unbounded_String ); 


   -- Set the value of a parameter of a given task
   --
   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Unbounded_String  ); 


   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Priority_Range    ); 


   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Natural           ); 


   ----------------------------------------------------------
   -- Consult information from a task set
   ----------------------------------------------------------


   -- Get the value of a parameter of a given task
   --
   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Unbounded_String; 


   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Priority_Range; 


   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Natural; 


   -- Get the number of tasks hosted by the 
   -- processor "Processor_Name"
   --
   function Get_Number_Of_Task_From_Processor (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Tasks_Range; 


   -- Get the task pointer of the task named "Name"
   --
   function Search_Task (
         My_Tasks : in     Tasks_Set;       
         Name     : in     Unbounded_String ) 
     return Task_Ptr; 



   --------------------------------------------------------
   -- Functions to proceed sort on a task set
   --------------------------------------------------------


   function Increasing_Name (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 


   function Increasing_Deadline (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 


   function Decreasing_Deadline (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 


   function Increasing_Period (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 


   function Decreasing_Period (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 


   function Increasing_Priority (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 


   function Decreasing_Priority (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 


   ----------------------------------------------------
   -- These functions test and return TRUE if all the
   -- tasks of a task set have a common property
   ----------------------------------------------------

   function Deadline_Inferior_To_Period (
         My_Tasks : in     Tasks_Set ) 
     return Boolean; 


   function Is_Harmonic (
         My_Tasks : in     Tasks_Set ) 
     return Boolean; 

   ----------------------------------------------------
   -- These functions test and return TRUE if all the
   -- tasks of a task set have a common property
   --
   -- If "Processor_Name" is a null string, the control
   -- is done on all the task of the set. In the other
   -- case, the control is done on tasks located on the
   -- processor "Processor_Name"
   ----------------------------------------------------

   procedure Have_Deadlines_Equal_Than_Periods_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 


   procedure Have_Same_Period_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 


   procedure Periodic_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 


   procedure Priority_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 


   procedure Start_Time_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 


   procedure Offset_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 



   ----------------------------------------------------
   -- Compute some performance properties on 
   -- a task set
   ----------------------------------------------------

   function Scheduling_Period (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Natural; 


   function Processor_Utilization_Over_Deadline (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Double; 


   function Processor_Utilization_Over_Period (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Double; 


   procedure Response_Time_By_Simulation (
         My_Tasks  : in     Tasks_Set;           
         Sched     : in     Scheduling_Sequence; 
         Processor :        Unbounded_String;    
         Result    : in out Unbounded_String     ); 


   procedure Compute_Response_Time_Distribution (
         My_Tasks  : in     Tasks_Set;           
         Sched     : in     Scheduling_Sequence; 
         Processor :        Unbounded_String;    
         Result    : in out Densities_Table      ); 


   ----------------------------------------------------

   -- To compute the response time for each awoken occurence of a task

   type Task_Occurence_Table is array (Natural range <>) of Task_Occurence; 

   procedure Put (
         My_Task_Occurence_Table : Task_Occurence_Table ); 




private

   type Tasks_Set is new Generic_Task_Set.Set with null record; 






end Task_Set;








