------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------



with Task_Offsets;
use Task_Offsets;
use Task_Offsets.Offsets_Table_Package;
with Parameters;
use Parameters;
use Parameters.Parameters_Table_Package;

with Ada.Finalization;
with Convert_Strings;
with Convert_Unbounded_Strings;

with Config;
use  Config;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;



package Tasks is

   Compute_Aborded : exception;  

   type Tasks_Type is 
         (Periodic_Type,  
          Aperiodic_Type, 
          Poisson_Type,   
          Parametric_Type); 


   procedure To_Task_Type is 
   new Convert_Strings(Tasks_Type,Aperiodic_Type);
   procedure To_Task_Type is 
   new Convert_Unbounded_Strings(Tasks_Type,Aperiodic_Type);


   type Policies is 
         (Sched_Fifo,  
          Sched_Rr,    
          Sched_Others); 


   procedure To_Policy is 
   new Convert_Strings(Policies ,Sched_Fifo);
   procedure To_Policy is 
   new Convert_Unbounded_Strings(Policies ,Sched_Fifo);


   --------------------------------------------------

   type Generic_Task is abstract new Ada.Finalization.Controlled with 
      record 
         Name          : Unbounded_String := To_Unbounded_String ("");  
         Task_Type     : Tasks_Type;  
         Location      : Unbounded_String := To_Unbounded_String ("");  
         Capacity      : Natural          := 0;  
         Deadline      : Natural          := 0;  
         Start_Time    : Natural          := 0;  
         Priority      : Priority_Range   := Priority_Range'First;  
         Blocking_Time : Natural          := 0;  
         Policy        : Policies;  
         X             : Natural          := 0;  
         Y             : Natural          := 0;  
         Offset        : Offsets_Table;  
         Param         : Parameters_Table;  
      end record; 

   type Task_Ptr is access all Generic_Task'Class; 



   procedure Initialize (
         A_Task : in out Generic_Task ); 
   procedure Finalize (
         A_Task : in out Generic_Task ); 
   procedure Adjust (
         A_Task : in out Generic_Task ); 


   procedure Free (
         My_Task : in out Task_Ptr ); 
   pragma Import (C, Free, "__gnat_free");


   procedure Put (
         My_Task : in     Generic_Task ); 

   procedure Put (
         My_Task : in     Task_Ptr ); 


   procedure Put_Name (
         A_Task : in     Task_Ptr ); 


   function Copy (
         My_Task : in     Task_Ptr ) 
     return Task_Ptr; 



   --------------------------------------------------

   type Periodic_Task is new Generic_Task with 
      record 
         Period : Natural := 0;  
         Jitter : Natural := 0;  
      end record; 

   type Periodic_Task_Ptr is access all Periodic_Task'Class; 


   procedure Initialize (
         A_Task : in out Periodic_Task ); 
   procedure Finalize (
         A_Task : in out Periodic_Task ); 
   procedure Adjust (
         A_Task : in out Periodic_Task ); 
   procedure Put (
         My_Task : in     Periodic_Task ); 


   --------------------------------------------------


   type Aperiodic_Task is new Generic_Task with null record; 

   type Aperiodic_Task_Ptr is access all Aperiodic_Task'Class; 


   procedure Initialize (
         A_Task : in out Aperiodic_Task ); 
   procedure Finalize (
         A_Task : in out Aperiodic_Task ); 
   procedure Adjust (
         A_Task : in out Aperiodic_Task ); 

   --------------------------------------------------

   type Poisson_Task is new Generic_Task with 
      record 
         Period      : Natural := 0;    -- Average occurance period 
         Jitter      : Natural := 0;  
         Seed        : Integer := 0;  
         Predictable : Boolean := True;  
      end record; 

   type Poisson_Task_Ptr is access all Poisson_Task'Class; 


   procedure Initialize (
         A_Task : in out Poisson_Task ); 
   procedure Finalize (
         A_Task : in out Poisson_Task ); 
   procedure Adjust (
         A_Task : in out Poisson_Task ); 
   procedure Put (
         My_Task : in     Poisson_Task ); 

   --------------------------------------------------


   type Parametric_Task is new Generic_Task with 
      record 
         File_Name : Unbounded_String := To_Unbounded_String ("");  
      end record; 

   type Parametric_Task_Ptr is access all Parametric_Task'Class; 


   procedure Initialize (
         A_Task : in out Parametric_Task ); 
   procedure Finalize (
         A_Task : in out Parametric_Task ); 
   procedure Adjust (
         A_Task : in out Parametric_Task ); 
   procedure Put (
         My_Task : in     Parametric_Task ); 

   --------------------------------------------------
   -- Laurent : modelise tasks occurences

   type Task_Occurence is 
      record 
         Parent_Task       : Task_Ptr;  
         Arrival_Date      : Natural  := 0;  
         Residual_Capacity : Natural  := 0;  
         Response_Time     : Natural  := 0;  
         Absolute_Deadline : Natural  := 0;  
         Is_Active         : Boolean  := False;  
      end record; 

   procedure Put (
         My_Task_Occurence : in     Task_Occurence ); 

   procedure Initialize (
         My_Task_Occurence : in out Task_Occurence;        
         A_Task            : in     Periodic_Task_Ptr;     
         Arrival_Date      : in     Natural           := 0 ); 

   procedure Initialize (
         My_Task_Occurence : in out Task_Occurence;    
         A_Task            : in     Aperiodic_Task_Ptr ); 

   -- /Laurent

   procedure Initialize (
         My_Task_Occurence : in out Task_Occurence;       
         A_Task            : in     Poisson_Task_Ptr;     
         Arrival_Date      : in     Natural          := 0 ); 

   procedure Initialize (
         My_Task_Occurence : in out Task_Occurence;     
         A_Task            : in     Parametric_Task_Ptr ); 

end Tasks;
