------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;


with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

with Tasks;
use Tasks;
with Task_Set;
use Task_Set;

with Scheduler;
use Scheduler;
with Scheduler.Fixed_Priority.Rm;
use Scheduler.Fixed_Priority.Rm;

with Scheduling_Sequences;
use Scheduling_Sequences;

with Text_Io;
use Text_Io;

with Task_Offsets;
use Task_Offsets;
use Task_Offsets.Offsets_Table_Package;

with Buffers;
use Buffers;
use Buffers.Buffer_Roles;
use Buffers.Buffer_Size_Package;

with Tasks;
use Tasks;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Ada.Numerics.Float_Random;
use Ada.Numerics.Float_Random;



procedure Generation_Example is 

   -- functions
   function Get_Rand_Period (
         Min  : in     Natural;  
         Max  : in     Natural;  
         Seed : in     Generator ) 
     return Natural is 


      Rand_Number : Uniformly_Distributed;  
      Period_Aux  : Double;  
      Period_Prod : Natural;  

   begin
      Rand_Number := Random(Seed); -- between 0 and 1
      Period_Aux := Double(Rand_Number) * (Double(Max) - Double(Min));
      Period_Aux := Period_Aux + Double(Min);
      Period_Prod := Natural(Period_Aux);
      return Period_Prod;
   end Get_Rand_Period;

   -- constant
   Period_Max        : constant Natural := 50;  
   Max_Nb_Sched      : constant Natural := 10;  
   Nb_Ordonnancement : constant Natural := 10 * Max_Nb_Sched;  

   -- types
   type Set_Range is new Integer range 1 .. Nb_Ordonnancement; 
   type Tasks_Set_Set is array (Set_Range) of Tasks_Set; 
   type Roles_Set is array (Set_Range) of Buffer_Roles_Table; 
   type Size_Set is array (Set_Range) of Buffer_Size_Table; 
   type Bound_Set is array (Set_Range) of Natural; 
   type Scheduling_Sequence_Set is array (Set_Range) of Scheduling_Sequence; 

   -- variables
   My_Scheduler : Rm_Scheduler;  
   Systems      : Tasks_Set_Set;  
   Offset       : Offsets_Table;  
   Scheds       : Scheduling_Sequence_Set;  
   Roles        : Roles_Set;  
   My_Buff      : Buffer_Ptr;  
   Sizes        : Size_Set;  
   Tptr         : Task_Ptr;  
   Cpu          : Unbounded_String        := To_Unbounded_String ("cpu");
      

   Nb_Prod   : Integer;  
   Per_Cons,  
   Per_Prod,  
   Nb_Sched  : Natural;  
   Bounds    : Bound_Set;  
   Seed      : Generator;  
   Aux_Prod,  
   Flow_Aux,  
   Flow_Prod : Double;  
   Min,  
   Max       : Natural;  
begin

   -- intialization
   Initialize(Offset);
   My_Buff:=new Buffer;
   for I in Set_Range loop
      Initialize(Roles(I));
      Initialize(Sizes(I));
      Bounds(I) := 0;
   end loop;

   Nb_Prod := 1;
   Per_Cons := 4; -- consumer period

   -- generation of the set of tasks
   Nb_Sched := 1;
   for I in Set_Range loop

      Add_Task(Systems(I),Tptr, To_Unbounded_String("Cons"),
         Cpu, Periodic_Type, 0, 1, Per_Cons, Per_Cons, 0, 0, 1, Sched_Fifo,
         Offset);
      Roles(I).Entries(0).Item:=Tptr.Name;
      Roles(I).Entries(0).Data.Size:=1;
      Roles(I).Entries(0).Data.The_Role:=Consumer;

      -- variables initialization
      Reset(Seed);
      Max := Period_Max;
      Aux_Prod := 1.0/Double(Per_Cons);
      Per_Prod := Per_Cons;
      Flow_Aux := 1.0/Double(Per_Cons) - Double(Nb_Prod - 1)/Double(Max);
      Flow_Prod := 0.0;


      -- maximum test
      if (Flow_Aux < 0.0)
            then

         -- new maximum calculation
         Max := Nb_Prod * Per_Cons;
         Flow_Aux := 1.0/Double(Per_Cons) - Double(Nb_Prod - 1)/Double(
            Max);
      end if;



      -- producers generation
      for J in 1 .. Buffer_Roles_Range(Nb_Prod) loop
         -- period of producers is based on the respect of the conservative data flow : the data produce is egal to the data consum
         if (J = Buffer_Roles_Range(Nb_Prod)) -- last producer
               then
            Per_Prod := Natural(Double'Ceiling(1.0/Aux_Prod));
         else
            Min := Natural(Double'Ceiling(1.0/Flow_Aux));
            Per_Prod := Get_Rand_Period(Min,Max,Seed);
         end if;

         -- last producer period calculation
         Aux_Prod := Aux_Prod - 1.0/Double(Per_Prod);

         -- bandswith for next producers
         Flow_Aux := Flow_Aux - 1.0 / Double(Per_Prod) + 1.0/Double(Max);

         -- intermediary calculation of total data flow of the producers
         Flow_Prod := Flow_Prod + 1.0 / Double(Per_Prod);

         -- add the tasks to the system
         Add_Task(Systems(I),Tptr, (To_Unbounded_String("Prod") &
               To_Unbounded_String(J'Img)),
            Cpu , Periodic_Type, 0, 1, Per_Prod, Per_Prod , 0, 0, 1,
            Sched_Fifo, Offset);

         -- array of maximum generation
         Roles(I).Entries(J).Item:=Tptr.Name;
         Roles(I).Entries(J).Data.Size:=1;
         Roles(I).Entries(J).Data.The_Role:=Producer;
      end loop;

      -- display the total data of producers : checking



      -- next system with one more producer
      if Nb_Sched > Max_Nb_Sched
            then
         Nb_Prod := Nb_Prod + 1;
         Nb_Sched:=1;
      else
         Nb_Sched:=Nb_Sched+1;
      end if;
   end loop;


   -- display the schedulings
   for I in Set_Range  loop
      Build_Scheduling_Sequence(My_Scheduler, Systems(I), Cpu, Scheds(I));


   end loop;


   -- display the maximum
   for I in Set_Range  loop
      My_Buff.Roles:=Roles(I);
      Get_Buffer_Size( Scheds(I), My_Buff, Systems(I), Sizes(I));
      Bounds(I):=Get_Max_Size(Sizes(I));

      if Bounds(I) > (2*Natural(Get_Number_Of_Elements(Systems(I)))) -1
            then
         Put_Line("Error : ");
         Put(Systems(I));
      end if;

   end loop;

end Generation_Example ;
