------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Unbounded_Strings;
use  Unbounded_Strings;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

with Tasks;
use Tasks;

with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Processor_Set;
use Processor_Set;
use Processor_Set.Generic_Processor_Set;

with Text_Io;
use Text_Io;

with Scheduler;
use Scheduler;
with Scheduler.Dynamic_Priority.Edf;
use Scheduler.Dynamic_Priority.Edf;

with Task_Offsets;
use Task_Offsets;
use Task_Offsets.Offsets_Table_Package;

with Natural_Util;
use Natural_Util;

with Text_Io;
use Text_Io;

with Buffer_Set;
use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;

with Systems;
use Systems;

with Processors;
use Processors;

with  Gnat.Os_Lib;
use Gnat.Os_Lib;

with Network;
use Network;
with Network.Uni_Processor;
use Network.Uni_Processor;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Natural_Util;
use Natural_Util;

with Buffers;
use Buffers;
use Buffers.Buffer_Roles;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Ada.Numerics.Float_Random;
use Ada.Numerics.Float_Random;

procedure Test_Generation is 

   function Get_Rand_Parameter (
         Min  : in     Natural;  
         Max  : in     Natural;  
         Seed : in     Generator ) 
     return Natural is 

      Rand_Number : Uniformly_Distributed;  
      Aux         : Double;  
      Res         : Natural;  

   begin
      Rand_Number := Random(Seed); -- between 0 and 1
      Aux := Double(Rand_Number) * (Double(Max) - Double(Min));
      Aux := Aux + Double(Min);
      Res := Natural(Aux);
      return Res;
   end Get_Rand_Parameter;

   -- system 1 consumer/ N producers generation
   procedure Generation_1_N (
         Nb_Sched     : Natural; 
         Period_Max   : Natural; 
         Per_Cons_Par : Natural; 
         Save_Project : Boolean  ) is 

      -- types
      type Set_Range is new Integer range 1 .. Nb_Sched; 

      -- variables
      Offset    : Offsets_Table;  
      Roles     : Buffer_Roles_Table;  
      Tptr      : Task_Ptr;  
      Cpu       : Unbounded_String   := To_Unbounded_String ("cpu");  
      Nb_Prod   : Integer;  
      Per_Cons,  
      Per_Prod  : Natural;  
      Seed      : Generator;  
      Aux_Prod,  
      Flow_Aux,  
      Flow_Prod : Double;  
      Min,  
      Max       : Natural;  
      Edf       : Scheduler_Ptr;  
      Systems   : System;  

   begin

      -- intialization
      Nb_Prod := 1;
      Per_Cons := Per_Cons_Par; -- Periode of the consumer

      -- tasks sets generation
      for I in Set_Range loop
         -- initialization
         Initialize(Roles);
         Initialize(Offset);

         Initialize(Systems);
         Systems.Network:= new Monoprocessor_Network;

         Add_Task(Systems.Tasks,Tptr, To_Unbounded_String("Cons"),
            Cpu, Periodic_Type, 0, 1, Per_Cons, Per_Cons, 0, 0, 1,
            Sched_Fifo, Offset);
         Roles.Entries(0).Item:=Tptr.Name;
         Roles.Entries(0).Data.Size:=1;
         Roles.Entries(0).Data.The_Role:=Consumer;

         Roles.Nb_Entries := Buffer_Roles_Range(I) + 1;

         -- variables initialization
         Reset(Seed);
         Max := Period_Max;
         Aux_Prod := 1.0/Double(Per_Cons);
         Per_Prod := Per_Cons;
         Flow_Aux := 1.0/Double(Per_Cons) - Double(Nb_Prod - 1)/Double(
            Max);
         Flow_Prod := 0.0;

         New_Line;
         New_Line;

         -- test of the maximum
         if (Flow_Aux < 0.0)
               then

            -- new maximum calculation
            Max := Nb_Prod * Per_Cons;
            Flow_Aux := 1.0/Double(Per_Cons) - Double(Nb_Prod - 1)/Double(
               Max);
            Put_Line ("No solution with period maximum :" &
               Period_Max'Img);
            Put_Line ("new  maximum :" & Max'Img);

         end if;

         -- Producers generation
         for J in 1 .. Buffer_Roles_Range(Nb_Prod) loop
            -- period of producers is based on the respect of the conservative data flow : the data produce is egal to the data consum
            if (J = Buffer_Roles_Range(Nb_Prod)) -- last producer
                  then
               Per_Prod := Natural(Double'Ceiling(1.0/Aux_Prod));
               Put_Line ("Period = " & Per_Prod'Img);
            else
               Min := Natural(Double'Ceiling(1.0/Flow_Aux));
               Per_Prod := Get_Rand_Parameter(Min,Max,Seed);
               Put_Line ("Period = " & Per_Prod'Img);

            end if;

            -- last producer period calculation
            Aux_Prod := Aux_Prod - 1.0/Double(Per_Prod);

            -- bandswith for next producers
            Flow_Aux := Flow_Aux - 1.0 / Double(Per_Prod) + 1.0/Double(
               Max);

            -- intermediary calculation of total data flow of the producers
            Flow_Prod := Flow_Prod + 1.0 / Double(Per_Prod);

            -- add the tasks to the system
            Add_Task(Systems.Tasks,Tptr, (To_Unbounded_String("Prod") &
                  To_Unbounded_String(J'Img)),
               Cpu , Periodic_Type, 0, 1, Per_Prod, Per_Prod , 0, 0, 1,
               Sched_Fifo,
               Offset);

            -- array of maximum generation
            Roles.Entries(J).Item:=Tptr.Name;
            Roles.Entries(J).Data.Size:=1;
            Roles.Entries(J).Data.The_Role:=Producer;

            -- display the task set
            -- Put(Systems);

         end loop;

         -- add the processor
         Edf:= new Edf_Scheduler;
         Add_Processor(Systems.Processors,Cpu,Edf);

         -- add the buffer
         Add_Buffer(Systems.Buffers, To_Unbounded_String("Buffer"), 1, Cpu,
            Roles);

         -- display the total data of producers : checking
         Put_Line ("Producers data flow = " & Flow_Prod'Img);

         -- next system with one more producer
         Nb_Prod := Nb_Prod + 1;

         if (Save_Project = True) then
            Put_Line("fichier enregistre");
            Write_To_Xml_File(Systems, To_Unbounded_String("ProjGen") &
               Format(
                  Natural(I)));
         end if;
      end loop;
   end Generation_1_N;



   function Rand_Sched_Buff_Size_1_N (
         Nb_Rand_Scheduling : Natural;         
         Nb_Stud_Period     : Natural;         
         File_Name          : Unbounded_String ) 
     return Natural is 

      -- constant
      Nb_Tasks : constant Natural := 20;  

      -- types
      type Res_Time_Buffer_Set is array (Natural range 0 .. Nb_Tasks - 1) of Natural; 
      type Cnt_Resp_Time is array (Natural range 0 .. Nb_Tasks - 1) of Natural; 

      -- variables
      Sys              : System;  
      A_Buffer         : Buffer_Ptr;  
      A_Task           : Task_Ptr;  
      A_Processor      : Processor_Ptr;  
      My_Iterator_Task : Tasks_Iterator;  
      My_Iterator_Buf  : Buffers_Iterator;  
      My_Iterator_Proc : Processors_Iterator;  
      Res_Time         : Res_Time_Buffer_Set;  
      -- array of random response time 
      Buff_Size        : Natural;   -- size of the buffer   
      Seed             : Generator;  
      Minimum_Res_Time : Natural;  
      -- minimum of response time contain in res_time 
      Cnt : Cnt_Resp_Time;  
      -- array of counter of random response time generation 
      Cnt_Final : Cnt_Resp_Time;  
      -- array of the maximum number of response time generation 
      Continu       : Boolean;  
      Max_Buff_Size : Natural;  
      -- maximum size of the buffer 
      Lcm      : Natural; -- study period   
      L        : Natural; -- counter   
      Min_Resp : Natural;  
      Max_Resp : Natural;  

      -- return the minimum of the array of response time
      function Min_Res_Time (
            Min : in     Res_Time_Buffer_Set ) 
        return Natural is 

         Minimum : Natural;  
      begin

         Minimum := Min(0);
         for I in 1 .. Buffer_Roles_Range(Nb_Tasks-1) loop
            -- Put_Line("I = " & I'Img & ", r = " & Min(Natural(I))'Img );
            if (Min(Natural(I)) < Minimum)
                  then
               Minimum := Min(Natural(I));
            end if;
         end loop;

         return Minimum;
      end Min_Res_Time;

   begin
      if(File_Name = Empty_String)
            then
         Os_Exit(0);
      else
         Initialize(Sys);
         Read_From_Xml_File(Sys, File_Name);
         -- put(Sys);
      end if;

      -- one buffer and one processor (perhaps change this)
      Reset_Iterator(My_Iterator_Proc);
      Current_Element(Sys.Processors, A_Processor, My_Iterator_Proc);
      if (A_Processor = null) -- no buffer in the project
            then
         Put_Line("a processor must be specified");
         Os_Exit(0);
      end if;
      Reset_Iterator(My_Iterator_Buf);
      Current_Element(Sys.Buffers, A_Buffer, My_Iterator_Buf);
      if (A_Buffer = null) -- no buffer in the project
            then
         Put_Line("a buffer must be specified");
         Os_Exit(0);
      end if;
      Put(A_Buffer);

      Lcm := Scheduling_Period(Sys.Tasks, A_Processor.Name);
      L:=0;
      Reset_Iterator(My_Iterator_Task);
      loop
         -- take a task
         Current_Element(Sys.Tasks, A_Task, My_Iterator_Task);

         -- number of response time generation for each task
         Cnt_Final(L) := Lcm / Periodic_Task_Ptr(A_Task).Period *
            Nb_Stud_Period;

         Put_Line("NB Generation = " & Cnt_Final(Natural(L))'Img);

         L:=L+1; -- dangerous range of cnt must be >= number of task
         -- next task
         exit when Is_Last_Element(Sys.Tasks, My_Iterator_Task) = True;
         Next_Element(Sys.Tasks, My_Iterator_Task);
      end loop;
      Max_Buff_Size := 0;
      for T in 0 .. Nb_Rand_Scheduling loop

         -- !!! initialization !!!
         Buff_Size := 0;

         for J in 0 .. Buffer_Roles_Range(Nb_Tasks-1) loop
            Res_Time(Natural(J)) := Natural'Last;
         end loop;

         for J in 0 .. Cnt_Resp_Time'Last loop
            Cnt(J) := 1;
         end loop;

         -- start random response time generation
         Continu := True;
         while (Continu) loop
            Reset(Seed);

            Continu := False;
            -- if the buffer size doesn't change, stop the program
            Reset_Iterator(My_Iterator_Task);
            loop
               -- take a task
               Current_Element(Sys.Tasks, A_Task, My_Iterator_Task);

               -- generate a response time if needed
               for J in 0 .. Buffer_Roles_Range(Nb_Tasks-1) loop
                  if (A_Buffer.Roles.Entries(J).Item = A_Task.Name)
                        then
                     if (A_Buffer.Roles.Entries(J).Data.The_Role =
                           Consumer or
                           A_Buffer.Roles.Entries(J).Data.The_Role =
                           Producer)
                           then
                        if (Res_Time(Natural(J)) = Natural'Last and Cnt(
                                 Natural(J))<=Cnt_Final(Natural(J)))
                              then
                           Min_Resp := (Cnt(Natural(J))-1)*
                              Periodic_Task_Ptr(A_Task).Period;
                           Max_Resp := Min_Resp + A_Task.Deadline;
                           Res_Time(Natural(J)) :=
                              Get_Rand_Parameter(Min_Resp,Max_Resp, Seed);
                           Cnt(Natural(J)) := Cnt(Natural(J))+1;
                           -- Put_Line("Response time " & A_Task.Name & " = " & Res_Time(Natural(J))'img & " ms");
                        end if;
                     end if;
                  end if;
               end loop;

               -- next task
               exit when Is_Last_Element(Sys.Tasks, My_Iterator_Task) =
                  True;
               Next_Element(Sys.Tasks, My_Iterator_Task);
            end loop;

            -- find the minimum response time
            Minimum_Res_Time := Min_Res_Time(Res_Time);
            -- Put_Line("Min Res time, " & Cnt(0)'Img & " = " & Minimum_Res_Time'img);

            Reset_Iterator(My_Iterator_Task);
            loop

               -- take a task
               Current_Element(Sys.Tasks, A_Task, My_Iterator_Task);

               -- search for the type of the task which has the minimum response time
               -- (if equalitie, worst case if consumption before production)
               for J in 0 .. Buffer_Roles_Range(Nb_Tasks-1) loop
                  if (A_Buffer.Roles.Entries(J).Item = A_Task.Name)
                        then
                     if (A_Buffer.Roles.Entries(J).Data.The_Role =
                           Consumer)
                           then
                        if (Res_Time(Natural(J)) = Minimum_Res_Time and
                              Minimum_Res_Time < Natural'Last )
                              -- consumer has the smallest response time
                              then
                           if (Buff_Size > 0)
                                 then
                              Buff_Size := Buff_Size - 1;
                           end if;
                           Res_Time(Natural(J)) := Natural'Last;
                           Continu := True;
                           -- Put_Line("Buffer size = " & Buff_Size'img);

                        end if;
                     else
                        if (A_Buffer.Roles.Entries(J).Data.The_Role =
                              Producer)
                              then
                           if (Res_Time(Natural(J)) = Minimum_Res_Time and
                                 Minimum_Res_Time < Natural'Last )
                                 -- producer has the smallest response time
                                 then
                              Buff_Size := Buff_Size + 1;

                              if (Buff_Size > Max_Buff_Size)
                                    then
                                 Max_Buff_Size := Buff_Size;
                              end if;

                              Res_Time(Natural(J)) := Natural'Last;
                              Continu := True;
                              -- Put_Line("Buffer size = " & Buff_Size'img);
                           end if;
                        end if;
                     end if;
                  end if;

               end loop; -- end for

               -- next task
               exit when Is_Last_Element(Sys.Tasks, My_Iterator_Task) =
                  True;
               Next_Element(Sys.Tasks, My_Iterator_Task);
            end loop;

         end loop; -- end while
         -- Put_Line("Maximum buffer size = " & Max_Buff_Size'img);

      end loop; -- end for nb tasks set generation

      -- Put_Line("Maximumest buffer size = " & Max_Buff_Size'img);

      return Max_Buff_Size;

   end Rand_Sched_Buff_Size_1_N;

   -- variables
   Buff_Size      : Natural;  
   Nb_Project     : Natural;  
   Nb_Stud_Period : Natural;  
   Nb_Scheduling  : Natural;  
begin

   Nb_Project := 3;
   Nb_Stud_Period := 2;
   Nb_Scheduling := 10000;
   Generation_1_N (Nb_Project,100,10,True);

   for I in 1 .. Nb_Project loop

      Buff_Size :=
         Rand_Sched_Buff_Size_1_N (Nb_Scheduling,
         Nb_Stud_Period,
         To_Unbounded_String("ProjGen") & Format(Natural(I)));

      Put_Line("");
      Put_Line("MAXIMUM BUFFER SIZE = " & Buff_Size'Img);
      Put_Line("");
      Put_Line("----------------------------------------------------");

   end loop;
end Test_Generation;
