------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_Io;
use Text_Io;



with Ada.Sequential_Io;


package body Sets is




   procedure Reset (
         My_Set      : in out Set;             
         Free_Object : in     Boolean := False ) is 
   begin
      if(Free_Object = True)
            then
         for I in 0 .. My_Set.Number_Of_Elements-1  loop
            Free(My_Set.Container(I));
         end loop;
      end if;
      My_Set.Number_Of_Elements:=0;
   end Reset;



   procedure Write_To_File (
         My_Set     : in     Set'Class;         
         File_Name  : in     Unbounded_String;  
         Write_Mode : in     Io_Tools.File_Mode ) is 

      Desc : File_Descriptor;  

      package Set_File is new Ada.Sequential_Io( File_Descriptor );
      use Set_File;

      A_Set_File : Set_File.File_Type;  
   begin

      if(Write_Mode = Save_In_New_File)
            then
         Create(A_Set_File,
            Mode => Out_File,
            Name => To_String(File_Name));
      else
         Open(A_Set_File,
            Mode => Append_File,
            Name => To_String(File_Name));
      end if;

      if not Is_Open(A_Set_File)
            then
         raise Can_Not_Open_File_For_Writing;
      end if;

      for I in 0 .. (My_Set.Number_Of_Elements - 1) loop

         Fill_Descriptor(My_Set, Desc, My_Set.Container(I));

         Write(A_Set_File,Desc);
      end loop;

      Close(A_Set_File);
   end Write_To_File;


   procedure Read_From_File (
         My_Set    : in out Set'Class;       
         File_Name : in     Unbounded_String ) is 


      Desc : File_Descriptor;  

      package Set_File is new Ada.Sequential_Io(File_Descriptor);
      use Set_File;

      A_Set_File : Set_File.File_Type;  

   begin

      Reset(My_Set);

      Open(A_Set_File,
         Mode => In_File,
         Name => To_String(File_Name));


      if not Is_Open(A_Set_File)
            then
         raise Can_Not_Open_File_For_Reading;
      end if;


      while not End_Of_File(A_Set_File)
            loop

         Read(A_Set_File,Desc);
         Extract_Descriptor(My_Set, Desc, My_Set.Container(My_Set.
               Number_Of_Elements));
         My_Set.Number_Of_Elements:=My_Set.Number_Of_Elements + 1;

      end loop;

      Close(A_Set_File);
   end Read_From_File;





   procedure Add (
         My_Set    : in out Set;    
         A_Element : in     Element ) is 
   begin

      if (My_Set.Number_Of_Elements = Element_Range(Max_Element))
            then
         raise Full_Set;
      end if;

      My_Set.Container(My_Set.Number_Of_Elements):=A_Element;
      My_Set.Number_Of_Elements:=My_Set.Number_Of_Elements + 1;
   end Add;


   procedure Delete (
         My_Set      : in out Set;             
         A_Element   : in     Element;         
         Free_Object : in     Boolean := False ) is 

      Found : Boolean := False;  
   begin

      if (My_Set.Number_Of_Elements = 0)
            then
         raise Empty_Set;
      end if;

      for I in  0 .. (My_Set.Number_Of_Elements - 1) loop
         if (A_Element = My_Set.Container(I))
               then
            if (Free_Object = True)
                  then
               Free(My_Set.Container(I));
            end if;
            My_Set.Container(I):=My_Set.Container(My_Set.
               Number_Of_Elements - 1);
            My_Set.Number_Of_Elements:=My_Set.
               Number_Of_Elements - 1;
            Found:=True;
            exit;
         end if;
      end loop;
      if (not Found)
            then
         raise Element_Not_Found;
      end if;
   end Delete;


   procedure Next_Element (
         My_Set      : in     Set;     
         My_Iterator : in out Iterator ) is 
   begin
      if (My_Set.Number_Of_Elements = 0)
            then
         raise Empty_Set;
      end if;

      My_Iterator.Current:=My_Iterator.Current + 1;

   end Next_Element;


   procedure Current_Element (
         My_Set         : in     Set;     
         Return_Element :    out Element; 
         My_Iterator    : in     Iterator ) is 
   begin
      if (My_Set.Number_Of_Elements = 0)
            then
         raise Empty_Set;
      end if;

      Return_Element:=My_Set.Container(My_Iterator.Current);
   end Current_Element;


   procedure Get_Element_Number (
         My_Set         : in     Set;          
         Return_Element :    out Element;      
         Position       : in     Element_Range ) is 
   begin
      if (My_Set.Number_Of_Elements < Position)
            then
         raise Invalid_Argument;
      end if;

      Return_Element:=My_Set.Container(Position);

   end Get_Element_Number;



   function Get_Number_Of_Elements (
         My_Set : in     Set ) 
     return Element_Range is 
   begin
      return (My_Set.Number_Of_Elements);
   end Get_Number_Of_Elements;


   function Is_Empty (
         My_Set : in     Set ) 
     return Boolean is 
   begin
      return (My_Set.Number_Of_Elements = 0);
   end Is_Empty;


   function Is_Last_Element (
         My_Set      : in     Set;     
         My_Iterator : in     Iterator ) 
     return Boolean is 
   begin
      return (My_Iterator.Current = (My_Set.Number_Of_Elements - 1));
   end Is_Last_Element;


   function Is_First_Element (
         My_Set      : in     Set;     
         My_Iterator : in     Iterator ) 
     return Boolean is 
   begin
      return (My_Iterator.Current = 0);
   end Is_First_Element;


   procedure Reset_Iterator (
         My_Iterator : in out Iterator ) is 
   begin
      My_Iterator.Current:=0;
   end Reset_Iterator;


   procedure Sort (
         My_Set : in out Set;           
         Order  :        Order_Function ) is 

      Temp : Element;  

   begin
      for I in 0 .. My_Set.Number_Of_Elements-1  loop
         for J in 0 .. (My_Set.Number_Of_Elements-2) loop
            if not Order(My_Set.Container(J), My_Set.Container(J+1))
                  then
               Temp:=My_Set.Container(J);
               My_Set.Container(J):=My_Set.Container(J+1);
               My_Set.Container(J+1):=Temp;
            end if;
         end loop;
      end loop;
   end Sort;


   procedure Select_And_Copy (
         Src         : in     Set;            
         Dest        : in out Set;            
         Must_Select :        Select_Function ) is 

      Tmp : Element;  
   begin
      for I in 0 .. Src.Number_Of_Elements-1  loop
         if Must_Select(Src.Container(I))
               then
            Tmp:=Copy(Src.Container(I));
            Add(Dest,Tmp);
         end if;
      end loop;
   end Select_And_Copy;



   procedure Duplicate (
         Src  : in     Set; 
         Dest : in out Set  ) is 

      Tmp : Element;  
   begin
      Dest.Number_Of_Elements:=Src.Number_Of_Elements;
      for I in 0 .. Src.Number_Of_Elements-1  loop
         Tmp:=Copy(Src.Container(I));
         Dest.Container(I):=Tmp;
      end loop;
   end Duplicate;


   procedure Put (
         My_Set : in     Set ) is 
   begin
      Put_Line("Number of elements : "
         & My_Set.Number_Of_Elements'Img);
      New_Line;
      for I in 0 .. My_Set.Number_Of_Elements-1  loop
         Put(My_Set.Container(I));
         New_Line;
      end loop;
   end Put;



   pragma Warnings (Off);

   procedure Extract_Descriptor (
         My_Set    : in     Set;             
         Desc      : in     File_Descriptor; 
         A_Element :    out Element          ) is 
   begin
      null;
   end Extract_Descriptor;


   procedure Fill_Descriptor (
         My_Set    : in     Set;             
         Desc      :    out File_Descriptor; 
         A_Element : in     Element          ) is 
   begin
      null;
   end Fill_Descriptor;


   pragma Warnings (On);


end Sets;
