------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2005   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the EA 2215 Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Ada.Strings.Unbounded;
use  Ada.Strings.Unbounded;

with Io_Tools;
use Io_Tools;


generic

   Max_Element : Natural := 100;  

   type Element is private; 
   type File_Descriptor is private; 

   with procedure Put (
         A_Element : in     Element ); 
   with procedure Free (
         A_Element : in out Element ); 
   with function Copy (
         E_Lement : in     Element ) 
     return Element; 

package Sets is

   type Set is abstract tagged private; 
   type Iterator is private; 

   type Element_Range is new Natural range 0 .. Max_Element; 
   type Set_Ptr is access all Set'Class; 


   Invalid_Argument : exception;
   Element_Not_Found : exception;
   Full_Set : exception;
   Empty_Set : exception;




   procedure Fill_Descriptor (
         My_Set    : in     Set;             
         Desc      :    out File_Descriptor; 
         A_Element : in     Element          ); 


   procedure Extract_Descriptor (
         My_Set    : in     Set;             
         Desc      : in     File_Descriptor; 
         A_Element :    out Element          ); 


   procedure Write_To_File (
         My_Set     : in     Set'Class;         
         File_Name  : in     Unbounded_String;  
         Write_Mode : in     Io_Tools.File_Mode ); 



   procedure Reset (
         My_Set      : in out Set;             
         Free_Object : in     Boolean := False ); 


   procedure Put (
         My_Set : in     Set ); 


   procedure Duplicate (
         Src  : in     Set; 
         Dest : in out Set  ); 


   type Select_Function is access function 
   (Op1 : in Element) return Boolean;

   procedure Select_And_Copy (
         Src         : in     Set;            
         Dest        : in out Set;            
         Must_Select :        Select_Function ); 


   procedure Read_From_File (
         My_Set    : in out Set'Class;       
         File_Name : in     Unbounded_String ); 


   procedure Add (
         My_Set    : in out Set;    
         A_Element : in     Element ); 


   procedure Delete (
         My_Set      : in out Set;             
         A_Element   : in     Element;         
         Free_Object : in     Boolean := False ); 


   -----------------------------------------------
   -- Theses functions/procedures can be used to
   -- 
   -----------------------------------------------

   -- Lets go to the next element of the set
   --
   procedure Next_Element (
         My_Set      : in     Set;     
         My_Iterator : in out Iterator ); 

   -- Get the element currently pointed by
   -- the iterator
   --
   procedure Current_Element (
         My_Set         : in     Set;     
         Return_Element :    out Element; 
         My_Iterator    : in     Iterator ); 


   procedure Get_Element_Number (
         My_Set         : in     Set;          
         Return_Element :    out Element;      
         Position       : in     Element_Range ); 


   procedure Reset_Iterator (
         My_Iterator : in out Iterator ); 


   function Get_Number_Of_Elements (
         My_Set : in     Set ) 
     return Element_Range; 


   function Is_Empty (
         My_Set : in     Set ) 
     return Boolean; 


   function Is_Last_Element (
         My_Set      : in     Set;     
         My_Iterator : in     Iterator ) 
     return Boolean; 


   function Is_First_Element (
         My_Set      : in     Set;     
         My_Iterator : in     Iterator ) 
     return Boolean; 



   -- Sort utilitiy. The "order_function" return
   -- if op1 and op2 are not n the right order
   --

   type Order_Function is access function 
   (Op1 : in Element;  Op2 : in Element) return Boolean;

   procedure Sort (
         My_Set : in out Set;           
         Order  :        Order_Function ); 


private



   type Element_Table is array (Element_Range) of Element; 


   type Set is tagged 
      record 
         Number_Of_Elements : Element_Range := 0;  
         Container          : Element_Table;  
      end record; 


   type Iterator is 
      record 
         Current : Element_Range := Element_Range'First;  
      end record; 


   pragma Inline (Current_Element);
   pragma Inline (Next_Element);
   pragma Inline (Is_Last_Element);



end Sets;




